#include <stdio.h>
#include <iostream>
#include <string>
#include <sicstus/sicstus.h>
#include "bliss-0.73/graph.cc"
#include "bliss_interface_glue.h"

bliss::AbstractGraph* g = new bliss::Graph(0);
std::list<SP_term_ref> log_list = {};

SP_term_ref term_ref_list_to_prolog_list(std::list<SP_term_ref> list) {
    SP_term_ref prolog_list = SP_new_term_ref();
    for (SP_term_ref term_ref : list) {
        SP_cons_list(prolog_list, term_ref, prolog_list);
    }
    return prolog_list;
}

SP_term_ref store_permutation(const unsigned int N,
                              const unsigned int* perm) {
    std::list<SP_term_ref> local_generators = {};
    for(unsigned int i = 0; i < N; i++) {
        unsigned int j = perm[i];
        if(j == i)
            continue;
        bool is_first = true;
        while(j != i) {
            if(j < i) {
                is_first = false;
                break;
            }
            j = perm[j];
        }
        if(!is_first)
            continue;
        std::list<SP_term_ref> cycle_list = {};
        SP_term_ref first = SP_new_term_ref();
        SP_put_integer(first, i);
        cycle_list.push_front(first);
        j = perm[i];
        while(j != i) {
            SP_term_ref int_id = SP_new_term_ref();
            SP_put_integer(int_id, j);
            cycle_list.push_front(int_id);
            j = perm[j];
        }
        SP_term_ref cycle = term_ref_list_to_prolog_list(cycle_list);
        local_generators.push_front(cycle);
    }
    return term_ref_list_to_prolog_list(local_generators);
}

// Hook registered in find_automorphisms() and canonical_form() is triggered each time a new generator for the automorphism group is found.
static void log_automorphism(void* param, const unsigned int n, const unsigned int* aut){
    SP_term_ref perm = store_permutation(n, aut);
    log_list.push_front(perm);
}

void clean_state() {
    delete g;
    log_list = {};
}

extern "C" {
    void init_directed_graph_wrapper() {
        clean_state();
        g = new bliss::Digraph(0);
    }

    void init_undirected_graph_wrapper() {
        clean_state();
        g = new bliss::Graph(0);
    }

    SP_integer add_node_wrapper(SP_integer color) {
        return g->add_vertex(color);
    }

    void add_edge_wrapper(SP_integer id1, SP_integer id2) {
        g->add_edge(id1, id2);
    }

    SP_term_ref get_canonical_form() {
        log_list = {};
        bliss::Stats stats = bliss::Stats();
        g->canonical_form(stats, &log_automorphism, nullptr);
        return term_ref_list_to_prolog_list(log_list);
    }

    SP_term_ref find_automorphisms() {
        log_list = {};
        bliss::Stats stats = bliss::Stats();
        g->find_automorphisms(stats, &log_automorphism, nullptr);
        return term_ref_list_to_prolog_list(log_list);
    }
}
