:- module(assert_profiler,[add_assert_time_tracking/2, ts/0]).

% :- use_module(extension('timer/assert_profiler.pl')).

:- use_module(extension('timer/timer.pl'),[timer_det_call/2,timer_call/2,timer_init/0,timer_statistics/0]).

:- timer:timer_init.
:- print('Timer initialised. Use debug:timer_statistics. to obtain statistics.'),nl.


my_functor(V,F,N) :- var(V),!,F='VAR',N=0.
my_functor(_:P,F,N) :- !, my_functor(P,F,N).

my_functor(':-'(_,_),':-'(_,_),_) :- !, fail. %my_functor(H,F,N), my_functor(B,F2,_).
my_functor(P,F,N) :- functor(P,F,N).

add_module(M:P,_,R) :- !, R=M:P.
add_module(P,M,M:P).

get_prolog_module(Module) :- (prolog_load_context(module,Module) -> true ; Module=user).

is_assert(assert(T),T,MT,assert(MT)).
is_assert(asserta(T),T,MT,asserta(MT)).
is_assert(assertz(T),T,MT,assertz(MT)).
is_retract(retract(T),T,MT,retract(MT)).

%:- multifile user:term_expansion/6.
%user:goal_expansion(Term1, Layout1, Module, Term2, Layout2) :- print(Term1),nl,
%user:add_timer(Term1,Module,Term2), Layout1=Layout2.


add_assert_time_tracking((H :- Body),(H :- Body2)) :- add_timer_body(Body,Body2).

add_timer_body(A,B) :- var(A),!, A=B.
add_timer_body((A , B),(A2 , B2)) :- !, add_timer_body(A,A2), add_timer_body(B,B2).
add_timer_body((A ; B),(A2 ; B2)) :- !, add_timer_body(A,A2), add_timer_body(B,B2).
add_timer_body((A -> B),(A2 -> B2)) :- !, add_timer_body(A,A2), add_timer_body(B,B2).
add_timer_body(\+(A),\+(A2)) :- !, add_timer_body(A,A2).
add_timer_body(when(G,A),when(G,A2)) :- !, add_timer_body(A,A2).

% -------------

add_timer_body(A,timer:timer_det_call(F,A2)) :-
    is_assert(A,T,MT,A2),
    nonvar(T),
    my_functor(T,F,N), !,
    get_prolog_module(Module),
    add_module(T,Module,MT),
    print(assert(F,N,Module,A,A2)),nl.
add_timer_body(A,timer:timer_call(F,A2)) :-
    is_retract(A,T,MT,A2),
    nonvar(T),
    my_functor(T,F,N), !,
    get_prolog_module(Module),
    add_module(T,Module,MT),
    print(retract(F,N,Module,A,A2)),nl.
%add_timer_body(A,(print(A),nl,A)) :-
%    A = call_cleanup(T,_TT),
%    my_functor(T,F,_), !,
%    print(call_cleanup(F)),nl.

% -------------

add_timer_body(B,B).



ts :- timer:timer_statistics.



:- multifile user:term_expansion/6.
user:term_expansion(Term1,Layout1,Tokens1,Term2,Layout1,[assert_profiler|Tokens1]) :-
    nonmember(assert_profiler,Tokens1),
    add_assert_time_tracking(Term1,Term2).


