#ifndef QUEUE_H_INCLUDED
#define QUEUE_H_INCLUDED

#include "hash.h"

typedef struct Cell {
    char *term;
    char digest[HASH_LEN];
    size_t term_length;
    struct Cell* next; 
} tCell;

typedef struct workQueue {
    tCell *first;
    tCell *last;
    size_t c;
    int strategy;
} wQueue;


/* initialises a new queue */
wQueue *init_queue(const int strategy);

/* enqueue a term with length len and hash code digest into the given queue;
 * function takes ownership of the term */
void enqueue(wQueue *q, char * const term, const size_t len, const char digest[HASH_LEN]);

/* dequeues a cell from given queue;
 * returns NULL, if queue was empty.
 * note that the next pointer will be NULL. */
tCell *dequeue(wQueue * const q);

/* returns 1 if the queue is empty,
 * 0 otherwise */
int is_empty(wQueue const * const q);

/* returns the current queue size of the given queue */
size_t q_size(wQueue const * const q);

/* tears down a queue and all its queued items */
void teardown_queue(wQueue **q);

/* tears down a cell and its term */
void teardown_cell(tCell **c);

#endif
