#ifndef SHM_COMMON_H_INCLUDED
#define SHM_COMMON_H_INCLUDED

#include <semaphore.h>
#include "hash.h"


#include "tnode.h"
#define INITIAL_SIZE    10000
#define SEMAP_DIGESTS_VAL   10
#define SEMAP_NODES_VAL     5

#define PRE_INIT_NODES (32*32*32*32)
#define DEPTH_START 4

#ifdef BLOCKS
unsigned int blocks_read;
#endif

struct hash_meta_infos {
    volatile size_t pos_hashes; // equals known states
    volatile size_t pos_nodes;
    volatile size_t checked_states;
    volatile unsigned int max_depth;
};

struct shm_digest {
    char *ptr;
    int64_t index;
};

struct shm_node {
    tNode *nodeptr;
    int64_t index;
};


struct semap_env {
    int names_initialised;
    char *semap_meta_s;
    char *semap_hashes_s;
    char *semap_hashwrite_s;
    char *semap_nodes_s[BRANCH_FACTOR];
    char *semap_multiwrite_s[BRANCH_FACTOR];

    sem_t *semap_meta;
    sem_t *semap_hashes;
    sem_t *semap_hashwrite;
    sem_t *semap_nodes[BRANCH_FACTOR];
    sem_t *semap_multiwrite[BRANCH_FACTOR];
};

struct shm_block {
    int shmid;
    void *ptr;
};

struct shm_env {
    struct semap_env semaphores;

    size_t elements_per_nodes_block;
    size_t elements_per_digests_block;

    key_t meta_infos;

    key_t digests_start;
    key_t next_digests_key;

    key_t nodes_start;
    key_t next_nodes_key;

    struct shm_block shm_meta_infos;
    
    struct shm_block *shm_digests;
    size_t shm_digests_len;

    struct shm_block *shm_nodes;
    size_t shm_nodes_len;
};

void init_semaphore_names(struct semap_env * const env, const int id);
void free_semaphore_names(struct semap_env * const env);

void get_pos_tuple(size_t elements_per_block, int64_t index, size_t *block, size_t *offset);

struct shm_env *init_shm_common(struct semap_env sem_env,
                                size_t elements_per_nodes_block,
                                size_t elements_per_digests_block,
                                key_t meta_info,
                                key_t digests_start,
                                key_t nodes_start);

// these operations may or have to update the environment
char *get_shm_digests(struct shm_env * const env,
                     size_t amount);

tNode *lookup_node(struct shm_env * const env,
                   int64_t index);
char *lookup_digest(struct shm_env * const env,
                    int64_t index);

struct shm_node alloc_node(struct shm_env * const env);
struct shm_digest alloc_digest(struct shm_env * const env);

void shm_cleanup(struct shm_env * const env, int server);

// these operations must not update the environment
void inc_checked_states(struct shm_env const * const env);

struct hash_meta_infos get_meta_infos(struct shm_env const * const env);

void lock_digests_read(struct shm_env const * const env);
void release_digests_read(struct shm_env const * const env);

void lock_nodes_read(struct shm_env const * const env,
                     const int pos);
void release_nodes_read(struct shm_env const * const env,
                        const int pos);

void update_max_depth(struct shm_env const * const env, int depth);

void lock_nodes_write(struct shm_env const * const env, const int pos);
void lock_digests_write(struct shm_env const * const env);

void release_nodes_write(struct shm_env const * const env, const int pos);
void release_digests_write(struct shm_env const * const env);

#endif
