#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <semaphore.h>
#include <fcntl.h>
#include "shm_common.h"
#include "shm_srv.h"

#include "assert_helper.h"

void init_trie_meta(const struct shm_env * const env) {
    *((struct hash_meta_infos*) (env->shm_meta_infos.ptr)) = (struct hash_meta_infos) {0};
}

static void rmsem(struct semap_env * const env, const int id) {
    if (!env->names_initialised) {
        init_semaphore_names(env, id);
    }
    sem_unlink(env->semap_meta_s);
    sem_unlink(env->semap_hashes_s);
    sem_unlink(env->semap_hashwrite_s);
    int i;
    for (i = 0; i < BRANCH_FACTOR; i++) {
        sem_unlink(env->semap_nodes_s[i]);
        sem_unlink(env->semap_multiwrite_s[i]);
    }
}

struct semap_env create_semaphores(const int id) {
    struct semap_env r = (struct semap_env) {0};
    init_semaphore_names(&r, id);
    rmsem(&r, id);

    r.semap_meta      = sem_open(r.semap_meta_s,      O_CREAT | O_EXCL, 0666, 1);
    r.semap_hashes    = sem_open(r.semap_hashes_s,    O_CREAT | O_EXCL, 0666, SEMAP_DIGESTS_VAL);
    r.semap_hashwrite = sem_open(r.semap_hashwrite_s, O_CREAT | O_EXCL, 0666, 1);
    assert_es(r.semap_meta      != SEM_FAILED, "Creating semaphores failed. Maybe they still exist from a prior run or the system limit for sempahores is already reached?");
    assert_es(r.semap_hashes    != SEM_FAILED, "Creating semaphores failed. Maybe they still exist from a prior run or the system limit for sempahores is already reached?");
    assert_es(r.semap_hashwrite != SEM_FAILED, "Creating semaphores failed. Maybe they still exist from a prior run or the system limit for sempahores is already reached?");
    int i;
    for (i = 0; i < BRANCH_FACTOR; i++) {
        r.semap_nodes[i] = sem_open(r.semap_nodes_s[i], O_CREAT | O_EXCL, 0666, SEMAP_NODES_VAL);
        assert_es(r.semap_nodes[i] != SEM_FAILED, "Creating semaphores failed. Maybe they still exist from a prior run or the system limit for sempahores is already reached?");
        r.semap_multiwrite[i] = sem_open(r.semap_multiwrite_s[i], O_CREAT | O_EXCL, 0666, 1);
        assert_es(r.semap_multiwrite[i] != SEM_FAILED, "Creating semaphores failed. Maybe they still exist from a prior run or the system limit for sempahores is already reached?");
    }

    return r;
}


void shm_srv_cleanup(struct semap_env * const semaphores, const int id) {
    if (!semaphores->names_initialised) {
        init_semaphore_names(semaphores, id);
    }
    int rc = sem_close(semaphores->semap_meta);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashes);
    assert(rc == 0);
    rc = sem_close(semaphores->semap_hashwrite);
    assert(rc == 0);
    int i;
    for (i = 0; i < BRANCH_FACTOR; i++) {
        rc = sem_close(semaphores->semap_nodes[i]);
        assert(rc == 0);
        rc = sem_close(semaphores->semap_multiwrite[i]);
        assert(rc == 0);
    }
    rc = sem_unlink(semaphores->semap_meta_s);
    assert(rc == 0);
    rc = sem_unlink(semaphores->semap_hashes_s);
    assert(rc == 0);
    rc = sem_unlink(semaphores->semap_hashwrite_s);
    assert(rc == 0);
    for (i = 0; i < BRANCH_FACTOR; i++) {
        rc = sem_unlink(semaphores->semap_nodes_s[i]);
        rc = sem_unlink(semaphores->semap_multiwrite_s[i]);
        assert(rc == 0);
    }
    free_semaphore_names(semaphores);
    *semaphores = (struct semap_env) {0};
}
