#ifndef PSTATS_H_INCLUDED
#define PSTATS_H_INCLUDED

/* invariant:
 * for any consistent pStats, the following expression must evaluate to true:
 * empty_workers + balanced_workers + full_workers + in_transfer == len
 */
typedef struct worker_queues {
    size_t *queuesizes;
    size_t len;
    int empty_workers;
    int balanced_workers;
    int full_workers;
    int in_transfer;
} pStats;


int update_qs(pStats * const queuesizes,
              const int id,
              const size_t queuesize,
              const size_t thresh,
              int fd_logfile);


// TODO: documentation

/* returns the index of the greatest entry in wqs
 * which is not marked as in transfer,
 * or -1 if all are in transfer or empty */
int get_max_not_transfer_queue(pStats const * const wqs);

/* returns the index of the greatest entry in wqs,
 * or -1 if all are empty */
int get_max_queue(pStats const * const wqs);

/* returns the amount of possible work transfers based on wqs and thresh */
int proxy_can_balance(pStats const * const wqs, const size_t thresh);

/* requires at least one index in wqs where the value in zero.
 * returns the first index of a zero in wqs */
int get_empty_queue(pStats const * const wqs);

zmsg_t *create_stats_msg(const int id,
                         pStats const * const wqs,
                         const int port_start,
                         const int proxynumber,
                         const int fd_logfile);
int mark_transfer_begin(pStats * const wqs, const int id, const size_t thresh, int fd_logfile);
int mark_transfer_end(pStats * const wqs, const int id, const size_t thresh, int fd_logfile);
size_t is_in_transfer(pStats const * const wqs, const int id);

int handle_stats(zmsg_t * const msg,
                 const int header,
                 const size_t thresh,
                 pStats * const queuesizes,
                 const int fd_logfile);
#endif
