#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "../common/hashtrie.h"
#include "../common/shm_common.h"
#include "../common/shm_cln.h"
#include "../common/shm_srv.h"

#include "../common/hash.h"
#include "../common/transition_trie.h"


char const * const hash1 = "hash1xxxxxxxxxxxxxxx";
char const * const hash2 = "hash2xxxxxxxxxxxxxxx";
char const * const hash3 = "hash3xxxxxxxxxxxxxxx";
char const * const hash4 = "hash4xxxxxxxxxxxxxxx";
char const * const hash5 = "hash5xxxxxxxxxxxxxxx";

static void setup_trie(hashtrie t) {
    char infos[INFO_LEN] = {0};
    trie_put(t, hash1, infos);
    trie_put(t, hash2, infos);
    trie_put(t, hash3, infos);
    trie_put(t, hash4, infos);
    trie_put(t, hash5, infos);
}

/* 1 ---> 2 ------> 4
 *         \       /
 *          \->3->/
 *              \
 *               \---->5
 */
void setup_transitions(const hashtrie t, transition_chunks *transitions) {
    put_predecessor(transitions, t, hash1, hash2);
    put_predecessor(transitions, t, hash2, hash4);
    put_predecessor(transitions, t, hash2, hash3);
    put_predecessor(transitions, t, hash3, hash4);
    put_predecessor(transitions, t, hash3, hash5);
}

void trace_hash1(const hashtrie t, transition_chunks const * const transitions) {
    hash_list *trace1 = generate_trace(t, transitions, hash1);
    assert(0 == memcmp(trace1->hash, hash1, HASH_LEN));
    assert(NULL == trace1->next);
    teardown_hash_list(&trace1);
    assert(trace1 == NULL);
    printf(".");
}

void trace_hash2(const hashtrie t, transition_chunks const * const transitions) {
    hash_list *trace2 = generate_trace(t, transitions, hash2);
    assert(0 == memcmp(trace2->hash, hash1, HASH_LEN));
    assert(trace2->next);

    assert(0 == memcmp(trace2->next->hash, hash2, HASH_LEN));
    assert(trace2->next->next == NULL);

    teardown_hash_list(&trace2);
    assert(trace2 == NULL);
    printf(".");
}

void trace_hash3(const hashtrie t, transition_chunks const * const transitions) {
    hash_list *trace3 = generate_trace(t, transitions, hash3);
    assert(0 == memcmp(trace3->hash, hash1, HASH_LEN));
    assert(trace3->next);

    assert(0 == memcmp(trace3->next->hash, hash2, HASH_LEN));
    assert(trace3->next->next);

    assert(0 == memcmp(trace3->next->next->hash, hash3, HASH_LEN));
    assert(trace3->next->next->next == NULL);

    teardown_hash_list(&trace3);
    assert(trace3 == NULL);
    printf(".");
}

void trace_hash4(const hashtrie t, transition_chunks const * const transitions) {
    hash_list *trace4 = generate_trace(t, transitions, hash4);
    assert(0 == memcmp(trace4->hash, hash1, HASH_LEN));
    assert(trace4->next);

    assert(0 == memcmp(trace4->next->hash, hash2, HASH_LEN));
    assert(trace4->next->next);

    assert(0 == memcmp(trace4->next->next->hash, hash4, HASH_LEN));
    assert(trace4->next->next->next == NULL);

    teardown_hash_list(&trace4);
    assert(trace4 == NULL);
    printf(".");
}

void trace_hash5(const hashtrie t, transition_chunks const * const transitions) {
    hash_list *trace5 = generate_trace(t, transitions, hash5);
    assert(0 == memcmp(trace5->hash, hash1, HASH_LEN));
    assert(trace5->next);

    assert(0 == memcmp(trace5->next->hash, hash2, HASH_LEN));
    assert(trace5->next->next);

    assert(0 == memcmp(trace5->next->next->hash, hash3, HASH_LEN));
    assert(trace5->next->next->next);

    assert(0 == memcmp(trace5->next->next->next->hash, hash5, HASH_LEN));
    assert(trace5->next->next->next->next == NULL);

    teardown_hash_list(&trace5);
    assert(trace5 == NULL);
    printf(".");
}

int main(int argc, char *argv[]) {
    struct semap_env senv = create_semaphores(0);
    hashtrie t = get_trie(senv, 10000);

    setup_trie(t);

    assert(trie_count_elements(t) == 5);
    printf(".");
    assert(trie_get_checked_states_count(t) == 0);
    printf(".");


    transition_chunks transitions = {0};
    setup_transitions(t, &transitions);


    trace_hash1(t, &transitions);
    trace_hash2(t, &transitions);
    trace_hash3(t, &transitions);
    trace_hash4(t, &transitions);
    trace_hash5(t, &transitions);
    
    teardown_hashtrie(&t, 1, 0);
    puts("");
    return 0;
}
