#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <assert.h>
#include <stdio.h>

#include <zmq.h>
#include <czmq.h>

#include "../common/queue.h"
#include "../common/hash.h"
#include "../common/msg_headers.h"

#include "../worker/wctrl.h"

#include "test_helpers.h"


static void test_add_wps(void) {
    zmsg_t *res = zmsg_new();
    wQueue *q = get_queue(5);

    const int EXPECTED_SIZE = 3;
    add_wps(res, q, 100);
    assert(zmsg_size(res) == 2 * EXPECTED_SIZE);

    char rdigest[HASH_LEN];

    char *digest0 = zmsg_popstr(res);
    char *term0 = zmsg_popstr(res);
    assert(0 == strcmp(term0, "0"));
    hash("0", 2, rdigest);
    assert(0 == memcmp(rdigest, digest0, HASH_LEN));
    free(digest0);
    free(term0);

    char *digest2 = zmsg_popstr(res);
    char *term2 = zmsg_popstr(res);
    assert(0 == strcmp(term2, "2"));
    hash("2", 2, rdigest);
    assert(0 == memcmp(rdigest, digest2, HASH_LEN));
    free(digest2);
    free(term2);

    char *digest4 = zmsg_popstr(res);
    char *term4 = zmsg_popstr(res);
    assert(0 == strcmp(term4, "4"));
    hash("4", 2, rdigest);
    assert(0 == memcmp(rdigest, digest4, HASH_LEN));
    free(digest4);
    free(term4);

    teardown_queue(&q);
    zmsg_destroy(&res);

    printf(".");
}

/* TODO: this is broken for now */
/*
static void test_send_work_local(zsock_t * const req) {
    wQueue *q = get_queue(5);
    send_work("inproc://foo-pull", q, 100, L_TRANSFER_TO);
    zmsg_t *res;
    int rc = zsock_recv(pull, "m", &res);
    assert(rc == 0);

    const int EXPECTED_SIZE = 3;
    assert(zmsg_size(res) == 2 * EXPECTED_SIZE + 1);

    char *header_s = zmsg_popstr(res);
    int header;
    rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    assert(header == LOCAL_TRANSFER);
    free(header_s);

    char rdigest[HASH_LEN];

    char *digest0 = zmsg_popstr(res);
    char *term0 = zmsg_popstr(res);
    assert(0 == strcmp(term0, "0"));
    hash("0", 2, rdigest);
    assert(0 == memcmp(rdigest, digest0, HASH_LEN));
    free(digest0);
    free(term0);

    char *digest2 = zmsg_popstr(res);
    char *term2 = zmsg_popstr(res);
    assert(0 == strcmp(term2, "2"));
    hash("2", 2, rdigest);
    assert(0 == memcmp(rdigest, digest2, HASH_LEN));
    free(digest2);
    free(term2);

    char *digest4 = zmsg_popstr(res);
    char *term4 = zmsg_popstr(res);
    assert(0 == strcmp(term4, "4"));
    hash("4", 2, rdigest);
    assert(0 == memcmp(rdigest, digest4, HASH_LEN));
    free(digest4);
    free(term4);

    teardown_queue(&q);
    zmsg_destroy(&res);
    printf(".");
}

static void test_send_work_global(zsock_t * const pull) {
    wQueue *q = get_queue(1);
    send_work("inproc://foo-pull", q, 100, G_TRANSFER_TO);
    zmsg_t *res;
    int rc = zsock_recv(pull, "m", &res);
    assert(rc == 0);

    const int EXPECTED_SIZE = 1;
    assert(zmsg_size(res) == 2 * EXPECTED_SIZE + 1);

    char *header_s = zmsg_popstr(res);
    int header;
    rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    assert(header == GLOBAL_TRANSFER);
    free(header_s);

    char rdigest[HASH_LEN];

    char *digest0 = zmsg_popstr(res);
    char *term0 = zmsg_popstr(res);
    assert(0 == strcmp(term0, "0"));
    hash("0", 2, rdigest);
    assert(0 == memcmp(rdigest, digest0, HASH_LEN));
    free(digest0);
    free(term0);

    teardown_queue(&q);
    zmsg_destroy(&res);
    printf(".");
}

static void test_add_entire_queue(zsock_t * const pull) {
    wQueue *q = get_queue(2);
    send_work("inproc://foo-pull", q, 100, P_TRANSFER_ALL);
    zmsg_t *res;
    int rc = zsock_recv(pull, "m", &res);
    assert(rc == 0);

    const int EXPECTED_SIZE = 2;
    assert(zmsg_size(res) == 2 * EXPECTED_SIZE + 1);

    char *header_s = zmsg_popstr(res);
    int header;
    rc = sscanf(header_s, "%d", &header);
    assert(rc == 1);
    assert(header == GLOBAL_TRANSFER);
    free(header_s);

    char rdigest[HASH_LEN];

    char *digest0 = zmsg_popstr(res);
    char *term0 = zmsg_popstr(res);
    assert(0 == strcmp(term0, "0"));
    hash("0", 2, rdigest);
    assert(0 == memcmp(rdigest, digest0, HASH_LEN));
    free(digest0);
    free(term0);

    char *digest1 = zmsg_popstr(res);
    char *term1 = zmsg_popstr(res);
    assert(0 == strcmp(term1, "1"));
    hash("1", 2, rdigest);
    assert(0 == memcmp(rdigest, digest1, HASH_LEN));
    free(digest1);
    free(term1);

    teardown_queue(&q);
    zmsg_destroy(&res);
    printf(".");
}*/


int main(int argc, char *argv[]) {
    test_add_wps();

    zsock_t *pull = zsock_new_pull("inproc://foo-pull");
    zsock_t *push = zsock_new_push("inproc://foo-pull");
    assert(pull);

   // test_send_work_local(pull);
   // test_send_work_global(pull);
   // test_add_entire_queue(pull);

    zsock_destroy(&pull);
    zsock_destroy(&push);

    puts("");
    return 0;
}
