#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <zmq.h>
#include <czmq.h>

#include <stdio.h>
#include <sys/time.h>

#include "../common/hash.h"
#include "../common/msg_headers.h"
#include "../common/queue.h"
#include "../common/hashtrie.h"

#include "../common/verbose.h"

#include "wpush.h"
#include "wcallbacks.h"

extern wQueue *g_queue;
extern int g_worker_id;
extern zsock_t *g_zocket_push;
extern hashtrie g_trie;
extern int g_generate_successors;

// for testing only
zmsg_t *new_message(void) {
    zmsg_t *msg = zmsg_new();
    return msg;
}

zmsg_t *new_results_message(void) {
    zmsg_t *msg = zmsg_new();
    zmsg_addstrf(msg, "%d", IS_RESULT_MSG);
    return msg;
}

static struct timeval pl_start;
static struct timeval pl_stop;

void start_timer(void) {
    gettimeofday(&pl_start, NULL);
}

void stop_timer(void) {
    gettimeofday(&pl_stop, NULL);
}

size_t timeval_to_size(const struct timeval val) {
    time_t secs = val.tv_sec;
    suseconds_t musecs = val.tv_usec;
    return (size_t) secs * 1000000 + musecs;
}

struct timeval delta(struct timeval start, struct timeval stop) {
    struct timeval delta;
    timersub(&stop, &start, &delta);
    return delta;
}

void put_succ(zmsg_t *msg, char *term, int term_length, int fd_logfile) {
    if (g_generate_successors) {
        char buf[HASH_INFO];
        hash(term, term_length, buf);

        char infos[] = {0}; // TODO: for now

        if (!trie_contains(g_trie, buf)) {
            char * const termi = (char*) malloc(term_length);
            memcpy(termi, term, term_length);
            if (VERBOSE_WORKER_STATES) {
                char buffi[READABLE_HASH_LEN];
                print_readable_digest((unsigned char*) buf, buffi);
                dprintf(fd_logfile, "{:type :enqueued-state, :hash :%s}\n", buffi);
            }
            enqueue(g_queue, termi, term_length, buf);
        }


        memcpy(buf + HASH_LEN, infos, INFO_LEN);
        zmsg_addmem(msg, buf, HASH_INFO);
    }
}

void msg_put_str(zmsg_t *msg, char *termstr) {
    zmsg_addstr(msg, termstr);
}


static size_t local_state_nmbr = 0;

void add_stats(zmsg_t *msg) {
    size_t tProlog   = timeval_to_size(delta(pl_start, pl_stop));
    size_t timestamp = timeval_to_size(pl_stop); // close enough
    size_t qs = q_size(g_queue);


    zmsg_pushstrf(msg, "%lu", local_state_nmbr);
    zmsg_pushstrf(msg, "%lu", timestamp);
    zmsg_pushstrf(msg, "%lu", tProlog);
    zmsg_pushmem (msg, (char*) &qs, sizeof(size_t));
    zmsg_pushstrf(msg, "%i",  g_worker_id);
    zmsg_pushstrf(msg, "%d",  IS_HASH_MSG);
    local_state_nmbr++;
}

void send_and_destroy_msg(zmsg_t *msg) {
    int rc = zsock_send(g_zocket_push, "m", msg);
    assert(rc == 0);
    zmsg_destroy(&msg);
}
