:- module(b_cogen,[reset_cogen/0, compile_closure/4, eval_compiled_closure/4]).


:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides a compiler to Prolog for some B expressions and predicates.').

% -----------------------------------------

% COGEN Code Generator (to be moved)

:- dynamic eval_compiled_closure/4.

% test call: b_cogen:eval_compiled_closure(3, [[],[],no_wf_available], [int(10)], E).
% | ?- statistics(walltime,_),b_cogen:eval_compiled_closure(2, [[],[],no_wf_available], [int(25)], E),statistics(walltime,W).
%E = int(121393),
%W = [1125434,608] ?
% Fibonacci_NewSyntax.mch takes 3888 ms for fib(26) and 2472 ms for fib(25)
% | ?- statistics(walltime,_),set_prolog_flag(profiling,on),b_cogen:eval_compiled_closure(1, [[],[],no_wf_available], [int(25)], E),statistics(walltime,W), print_profile.

% not faster if we compile the Prolog code:
%eval_compiled_closure(11, [A,B,C], [D], E) :-
%        kernel_mappings:kernel_call_predicate_check_element_of_wf(D, global_set('NATURAL'), C, [nodeid(pos(13,1,7,12,7,21))]),
%        if(kernel_mappings:kernel_call_predicate_check_element_of_wf(D,avl_set(node(int(0),true,1,empty,node(int(1),true,0,empty,empty))),C,[nodeid(pos(17,1,8,16,8,26))]), (F=int(1),true), (F=G,((kernel_objects:int_minus(D,int(1),H),apply_compiled_closure(11,[A,B,C],H,I)),kernel_objects:int_minus(D,int(2),J),apply_compiled_closure(11,[A,B,C],J,K)),kernel_objects:int_plus(I,K,G))),
%        E=F.

reset_cogen :- retractall(eval_compiled_closure(_,_,_,_)).

compile_closure(Paras,_,Body,InnerMemoID) :-
     cogen_pred(Body,Paras,ParaValues,Dyn,Code),!,
     decompose_parameters(ParaValues,ArgVals,ReturnVal),
     Dyn = [_LS,_S,_WF],
     Clause = (eval_compiled_closure(InnerMemoID,Dyn,ArgVals,ReturnVal) :- Code),
     portray_clause(Clause),nl,nl,
     assert(Clause).

decompose_parameters(Paras,Args,ReturnVal) :- append(Args,[ReturnVal],Paras).

:- use_module(bsyntaxtree, [get_texpr_type/2]).

% cogen_pred(TPred,Paras,ParameterValues,DynEnvironment,GeneratedCode)
cogen_pred(b(Pred,pred,Infos),Paras,ParaValues,Dyn,Code) :-
   cogen_pred_aux(Pred,Infos,Paras,ParaValues,Dyn,Code).

cogen_pred_aux(conjunct(A,B),_Infos,Paras,ParaValues,Dyn,Code) :- !,
   cogen_pred(A,Paras,ParaValues,Dyn,CA),
   cogen_pred(B,Paras,ParaValues,Dyn,CB),
   conjoin(CA,CB,Code).
cogen_pred_aux(member(A,B),Infos,Paras,ParaValues,Dyn,Code) :- !,
   cogen_expr(A,Paras,ParaValues,Dyn,CA,AVal),
   cogen_expr(B,Paras,ParaValues,Dyn,CB,BVal),
   conjoin(CA,CB,CAB),
   Dyn = [_,_,WF|_],
   %conjoin(CAB,kernel_objects:check_element_of_wf(AVal,BVal,WF),Code).
   conjoin(CAB,kernel_mappings:kernel_call_predicate_check_element_of_wf(AVal,BVal,WF,Infos),Code).
cogen_pred_aux(equal(A,B),_Infos,Paras,ParaValues,Dyn,Code) :- !,
   cogen_expr(A,Paras,ParaValues,Dyn,CA,AVal),
   cogen_expr(B,Paras,ParaValues,Dyn,CB,BVal),
   conjoin(CA,CB,CAB),
   Dyn = [_,_,WF|_],
   get_texpr_type(A,Type),
   create_equal_object(Type,AVal,BVal,WF,EqCode) ,
   conjoin(CAB,EqCode,Code).
cogen_pred_aux(Comp,_Infos,Paras,ParaValues,Dyn,Code) :-
   arith_comparison(Comp,A,B,CompCode,AVal,BVal),!,
   cogen_expr(A,Paras,ParaValues,Dyn,CA,AVal),
   cogen_expr(B,Paras,ParaValues,Dyn,CB,BVal),
   conjoin(CA,CB,CAB),
   conjoin(CAB,CompCode,Code).
cogen_pred_aux(P,Infos,_Paras,_ParaValues,Dyn,Code) :-
   Dyn = [LS,S,WF|_],
   Code = b_interpreter:b_test_boolean_expression(b(P,pred,Infos),LS,S,WF).


create_equal_object(Type,A,B,_WF,Code) :- atomic_type(Type),!,
   Code = '='(A,B).
create_equal_object(_,A,B,WF,Code) :-
   Code = kernel_mappings:kernel_call_predicate_equal_object_optimized(A,B,WF).

% atomic types were equality is Prolog equality
atomic_type(integer).
atomic_type(string).
atomic_type(boolean).
atomic_type(global(_)).
atomic_type(couple(TA,TB)) :- atomic_type(TA),atomic_type(TB).

arith_comparison(greater_equal(A,B),A,B,kernel_objects:greater_than_equal(AVal,BVal),AVal,BVal).
arith_comparison(greater_than(A,B),A,B,kernel_objects:greater_than(AVal,BVal),AVal,BVal).
arith_comparison(less_than_equal(A,B),A,B,kernel_objects:less_than_equal(AVal,BVal),AVal,BVal).
arith_comparison(less_than(A,B),A,B,kernel_objects:less_than(AVal,BVal),AVal,BVal).


conjoin(true,A,R) :- !, R=A.
conjoin(A,true,R) :- !, R=A.
conjoin(A,B,(A,B)).

cogen_expr(b(Pred,Type,Infos),Paras,ParaValues,Dyn,Code,Val) :-
   cogen_expr_aux(Pred,Type,Infos,Paras,ParaValues,Dyn,Code,Val).

is_memoization_closure_val(V,_) :- var(V),!,fail.
is_memoization_closure_val(closure(P,T,B),MemoID) :- memoization:is_memoization_closure(P,T,B,MemoID).

:- use_module(library(lists),[nth1/3]).
cogen_expr_aux(value(V),_,_,_,_,_,Code,Val) :- !, Code=true, Val=V.
cogen_expr_aux(identifier(X),_,_,Paras,ParaValues,_,Code,Val) :- nth1(Nr,Paras,X), !,
     nth1(Nr,ParaValues,Val),
     Code=true.
cogen_expr_aux(function(A,B),Type,Infos,Paras,ParaValues,Dyn,Code,Val) :- !,
    cogen_expr(A,Paras,ParaValues,Dyn,CA,AVal),
    cogen_expr(B,Paras,ParaValues,Dyn,CB,BVal),
    conjoin(CA,CB,CAB),
    (is_memoization_closure_val(AVal,MemoID)
      ->  conjoin(CAB, apply_compiled_closure(MemoID,Dyn,BVal,Val), Code)
      ;   bsyntaxtree:get_texpr_type(A,FunctionType),
          Dyn = [LS,S,WF|_],
          Span = span_predicate(b(function(A,B),Type,Infos),LS,S),
          conjoin(CAB,kernel_call_apply_to(AVal,BVal,Val,FunctionType,Span,WF),Code)
    ).
cogen_expr_aux(BinOP,_Type,_Infos,Paras,ParaValues,Dyn,Code,Val) :-
    binop(BinOP,A,B,PrologPred),!,
    cogen_expr(A,Paras,ParaValues,Dyn,CA,AVal),
    cogen_expr(B,Paras,ParaValues,Dyn,CB,BVal),
    conjoin(CA,CB,CAB),
    PrologCall =.. [PrologPred,AVal,BVal,Val],
    conjoin(CAB,kernel_objects:PrologCall,Code).
cogen_expr_aux(if_then_else(A,B,C),_Type,_Infos,Paras,ParaValues,Dyn,Code,Val) :- !,
    cogen_pred(A,Paras,ParaValues,Dyn,CA), % Possibly reify A ?
    cogen_expr(B,Paras,ParaValues,Dyn,CB,BVal),
    cogen_expr(C,Paras,ParaValues,Dyn,CC,CVal),
    Code = if(CA,(Val=BVal,CB),(Val=CVal,CC)). % possibly call reified version ?
cogen_expr_aux(E,Type,Infos,_Paras,_ParaValues,Dyn,Code,Val) :-
     Dyn = [LS,S,WF|_],
     Code = b_compute_expression(b(E,Type,Infos),LS,S,Val,WF).

binop(add(A,B),A,B,int_plus).
binop(minus(A,B),A,B,int_minus).
binop(multiplication(A,B),A,B,times).

:- public apply_compiled_closure/4.
apply_compiled_closure(MemoID,Dyn,BVal,Val) :-
   couples_to_list(BVal,BValL),
   % TO DO: memoize
   eval_compiled_closure(MemoID,Dyn,BValL,Val).

couples_to_list(NC,L) :- couple_to_list_dcg(NC,L,[]).
couple_to_list_dcg((A,B)) --> !,
   couple_to_list_dcg(A), [B].
couple_to_list_dcg(E) --> [E].
