% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(bsyntaxtree,
          [is_texpr/1, % checks if the given argument is a typed expression

           get_texpr_expr/2, % get the expression part of a typed expression
           get_texpr_type/2, % get the type of a typed expression
           get_texpr_info/2, % get the list of information of a typed expression
           get_texpr_id/2, % get the id of a typed identifier, fails if it's not an identifier
           def_get_texpr_id/2, % same as above, but raises error if arg1 is not an identifier
           def_get_texpr_ids/2, % maplist of the above, i.e., translate a typed id list into a plain list of id names
           create_typed_id/3, % create a typed identifier
           same_id/3, % true if two typed identifiers have the same id
           get_texpr_exprs/2, % list variant of the above
           get_texpr_types/2, % list variant of the above
           get_texpr_infos/2, % list variant of the above
           get_texpr_ids/2,   % list variant of the above
           get_texpr_pos/2,   % get the position of a typed expression
           extract_pos_infos/2, % get the position info sub list of an info list
           copy_pos_infos/3, % copy position infos from one typed expression to another
           merge_info/3, % merge two information lists as well as possible
           update_infos/3, % provide updates to an existing info list
           get_info_pos/2,    % get the position directly from info field
           contains_info_pos/1, % true if info field contains position infos

           same_texpr/2, % check if two typed expressions are equal modulo Info fields
           different_texpr_values/2, % check if two type expressions are WD and definitely denote different values

           create_texpr/4, % creates a typed expression by giving expression, type and infos
           add_texpr_infos/3, % adds some more information to a typed expression at front
           add_texpr_info_if_new/3, % adds one info field to typed B expression info list at front if it is new
           add_texpr_infos_if_new/3, % ditto but list of infos can be added
           add_info_if_new/3, % adds to info list
           safe_create_texpr/3, % a version of create_texpr which extracts wd-info from sub-expressionssafe_create_texpr
           safe_create_texpr/4,
           sub_expression_contains_wd_condition/1, % utility to check if sub-expression contains wd condition

           get_rodin_name/2, get_rodin_model_name/2,
           get_texpr_label/2, get_texpr_labels/2, get_info_labels/2,
           add_labels_to_texpr/3,
           get_texpr_description/2, % get @desc pragma for typed expression
           info_has_ignore_pragma/1, % check if an info list has an ignore pragma
           always_well_defined/1, always_well_defined_or_disprover_mode/1,
           is_truth/1, is_falsity/1,
           conjunct_predicates/2,
           conjunct_predicates_with_pos_info/3, conjunct_predicates_with_pos_info/2,
           is_a_conjunct/3, is_a_conjunct_without_label/3, decompose_conjunct/3,
           is_a_disjunct/3, is_an_implication/3, is_an_equivalence/3, is_a_negation/2,
           conjunction_to_list/2, member_in_conjunction/2, select_member_in_conjunction/3,
           member_in_conjunction_cse/3,
           disjunct_predicates/2,
           disjunct_predicates_with_pos_info/3,
           disjunction_to_list/2,
           is_a_disjunct_or_implication/4,

           predicate_components/2, % split a predicate into components which use distinct identifiers
           predicate_components_in_scope/3, % ditto with an optional list of local variables
           predicate_components_with_restriction/4,
           predicate_identifiers/2, predicate_identifiers_in_scope/3,

           project_predicate_on_identifiers/5,

           find_identifier_uses/3, find_identifier_uses_if_necessary/3,
           find_identifier_uses_l/3,
           find_typed_identifier_uses/3, find_typed_identifier_uses/2,
           find_typed_identifier_uses_l/3,
           get_global_identifiers/1, get_global_identifiers/2,
           occurs_in_expr/2, some_id_occurs_in_expr/2,
           single_usage_identifier/3,

           create_exists/3, create_exists_or_let_predicate/3,
           create_exists_opt_disj/3,
           create_exists_opt/3, create_exists_opt/4, create_exists_opt/5,
           create_forall/3,
           create_negation/2, is_negation_of/2,
           create_implication/3,
           create_equivalence/3,
           create_equality/3,
           create_couple/3, create_couple/2,
           create_comprehension_set/4,

           detect_global_predicates/4,

           definitely_not_empty_set/1,

           replace_id_by_expr/4,
           replace_id_by_expr_with_count/5, % also count number of replacements
           replace_ids_by_exprs/4,
           remove_used_id_from_info/3, % remove an id from used_id info field if it exists
           remove_used_ids_from_info/3, % remove list of ids (order of args different !)

           rename_bt/3, % a simplified version of replace_ids_by_exprs, which assumes target of renamings are variables
           rename_bt_l/3,
           remove_bt/4,
           syntaxtransformation/5,

           map_over_bexpr/2, map_over_typed_bexpr/2, map_over_typed_bexpr/3,
           map_over_typed_bexpr_with_names/2,
           map_over_bexpr_top_down_acc/3, map_over_type_bexpr_top_down_acc/3,
           reduce_over_bexpr/4,
           transform_bexpr/3, % transform a typed B expression bottom-up
           transform_bexpr_with_scoping/3, % ditto, but we also provide info about local ids
           transform_bexpr_with_bup_accs/5, transform_bexpr_with_acc/5,
           uses_implementable_integers/1,
           min_max_integer_value_used/3, min_max_integer_value_used/5,
           syntaxtraversion/6,
           safe_syntaxelement/5,

           remove_all_infos/2, extract_info/2, extract_info/3,
           bsyntax_pattern/2,

           remove_all_infos_and_ground/2,

           check_if_typed_predicate/1,
           check_if_typed_expression/1,
           check_if_typed_substitution/1,

           strip_and_norm_ast/2,
           same_norm_texpr/2,
           get_texpr_functor/3,

           is_set_type/2, get_set_type/2, get_texpr_set_type/2,
           is_just_type/1, is_just_type/2,

           create_recursive_compset/6,
           unique_typed_id/3,
           mark_bexpr_as_symbolic/2,

           identifier_sub_ast/3, exchange_ast_position/5,

           has_declared_identifier/2, add_declaration_for_identifier/3,
           check_ast/1, check_ast/2,
           print_ast/1
          ]).

% meta_predicate annotations should appear before loading any code:

:- meta_predicate map_over_full_bexpr_no_fail(1,?).

:- meta_predicate map_over_bexpr(1,?).
:- meta_predicate map_over_typed_bexpr(1,?).
:- meta_predicate map_over_typed_bexpr(2,?,?).
:- meta_predicate map_over_bexpr_top_down_acc(3,?,?).
:- meta_predicate map_over_type_bexpr_top_down_acc(3,?,?).

:- meta_predicate reduce_over_bexpr(3,?,?,?).
:- meta_predicate transform_bexpr(2,?,?).
:- meta_predicate l_transform_bexpr(?,2,?).
:- meta_predicate transform_bexpr_with_scoping(3,?,?).
:- meta_predicate transform_bexpr_with_scoping2(3,?,?,?).
:- meta_predicate l_transform_bexpr_with_scoping(?,3,?,?).
:- meta_predicate transform_bexpr_with_bup_accs(4,?,?,?,?).
:- meta_predicate l_transform_bexpr_with_bup_accs(?,4,?,?,?).
:- meta_predicate transform_bexpr_with_acc(4,?,?,?,?).
:- meta_predicate l_transform_bexpr_with_acc(?,4,?,?,?).

% -----------


:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module provides operations on the type-checked AST.').

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(library(terms)).

:- use_module(self_check).
:- use_module(error_manager).
:- use_module(translate,[print_bexpr/1,translate_bexpression/2]).
:- use_module(gensym,[gensym/2]).
:- use_module(preferences,[get_preference/2,preference/2]).
:- use_module(debug,[debug_format/3]).

:- use_module(typing_tools,[is_infinite_type/1,normalize_type/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% basic access to enriched syntax tree

is_texpr(b(_,_,_)).

get_texpr_expr(b(Name,_,_),R) :- !,R=Name.
get_texpr_expr(E,_) :- add_error_fail(get_texpr_expr,'B Expression not properly wrapped: ',E).

get_texpr_type(b(_,Type,_),R) :- !, R=Type.
get_texpr_type(E,_) :- add_error_fail(get_texpr_type,'B Expression not properly wrapped: ',E).
get_texpr_info(b(_,_,Info),R) :- !, R=Info.
get_texpr_info(E,_) :- add_error_fail(get_texpr_info,'B Expression not properly wrapped: ',E).

:- mode get_texpr_id(i,o).
get_texpr_id(b(N,_,_),Id) :- !, N=identifier(Id).
get_texpr_id(E,_) :- add_error_fail(get_texpr_id,'B Expression not properly wrapped: ',E).
/* same as above: but must succeed */
def_get_texpr_id(b(identifier(Id),_,_),R) :- !,R=Id.
def_get_texpr_id(b(E,_,_),_) :- !,
   add_error_fail(def_get_texpr_id,'Could not extract identifier from typed expression: ',E).
def_get_texpr_id(ID,_) :- add_error_fail(def_get_texpr_id,'Could not extract identifier, not properly wrapped: ',ID).

% translate a list of ids with the b/3 wrapper into a plain id list; throw error if not an identifier
% used to also be called get_texpr_id_names
def_get_texpr_ids(L,Ids) :- maplist(def_get_texpr_id,L,Ids).

create_typed_id(IDName,Type,b(identifier(IDName),Type,[])).

same_id(X,Y,ID) :- get_texpr_id(X,ID), get_texpr_id(Y,ID).


% check if two wrapped expressions are equal (modulo associated Info, e.g. source loc info)
same_texpr(b(E1,Type,_),b(E2,Type,_)) :- % should we do a == check first ?? probably not as it may traverse the entire structure (see email exchange with SICStus 28.1.2015)
  same_functor(E1,E2),
  safe_syntaxelement(E1,Subs1,_Names1,_List1,Constant),
  safe_syntaxelement(E2,Subs2,_Names2,_List2,Constant2),
  Constant==Constant2, % in case we have values with variables inside !
  same_sub_expressions(Subs1,Subs2).

same_sub_expressions([],[]).
same_sub_expressions([H1|T1],[H2|T2]) :- same_texpr(H1,H2), same_sub_expressions(T1,T2).

% check if two wrapped expressions definitely denote a different value (and are well-defined)
% only detects certain cases !
different_texpr_values(b(E1,Type,I1),b(E2,Type,I2)) :-
 different_value(E1,E2,CheckWD),
 (CheckWD=false -> true
  ; always_well_defined_or_disprover_mode(b(E1,Type,I1)),
    always_well_defined_or_disprover_mode(b(E2,Type,I2))).
different_value(boolean_false,boolean_true,false) :- !.
different_value(boolean_true,boolean_false,false) :- !.
different_value(integer(X),integer(Y),false) :- !, X\=Y.
different_value(string(X),string(Y),false) :- !, X\=Y.
different_value(empty_set,S,check_wd) :- !, non_empty_set(S).
different_value(S,empty_set,check_wd) :- !, non_empty_set(S).
% should we compare two set_extensions ? couples ? records ?

non_empty_set(set_extension([_|_])).
non_empty_set(sequence_extension([_|_])).
%non_empty_set(value(avl_set(node....


get_texpr_exprs([],[]).
get_texpr_exprs([E|Rest],[N|NRest]) :- get_texpr_expr(E,N),get_texpr_exprs(Rest,NRest).
get_texpr_types([],[]).
get_texpr_types([E|Rest],[T|TRest]) :- get_texpr_type(E,T),get_texpr_types(Rest,TRest).
get_texpr_infos([],[]).
get_texpr_infos([E|Rest],[I|IRest]) :- get_texpr_info(E,I),get_texpr_infos(Rest,IRest).
get_texpr_ids([],[]).
get_texpr_ids([E|Rest],[I|IRest]) :- get_texpr_id(E,I),get_texpr_ids(Rest,IRest).

get_texpr_pos(TExpr,Pos) :-
    get_texpr_info(TExpr,Infos),
    ( member(nodeid(Pos1),Infos) -> !,Pos=Pos1
    ; Pos = none).

% similar to get_texpr_pos, but returns sub-list of infos relating to position
extract_pos_infos(Infos,InfoRes) :-
    ( member(nodeid(Pos1),Infos) -> !,InfoRes=[nodeid(Pos1)]
    ; InfoRes = []).

% copy position info from first arg to second argument
copy_pos_infos(b(_,_,Infos1),Arg2,Res) :-
    member(nodeid(Pos1),Infos1),!,
    Arg2 = b(E,T,Infos2), Res = b(E,T,Infos3),
    delete_pos_info(Infos2,Infos2d),
    Infos3 = [nodeid(Pos1)|Infos2d].
copy_pos_infos(_,TE,TE).

delete_pos_info(Infos1,Infos2) :-
   select(nodeid(_),Infos1,D),!, Infos2=D.
   %delete(Infos1,nodeid(_),Infos2).
delete_pos_info(I,I).



get_info_pos(Infos,Pos) :- (member(nodeid(Pos1),Infos) -> Pos=Pos1 ; Pos=none).
contains_info_pos(Infos) :- (member(nodeid(_),Infos) -> true).

create_texpr(Expr,Type,Info,b(Expr,Type,Info)).

add_texpr_infos(b(Expr,Type,Old),Infos,b(Expr,Type,New)) :- !,
    append(Infos,Old,New).
add_texpr_infos(Other,Infos,Res) :-
    add_internal_error('Illegal call, not BExpr:',add_texpr_infos(Other,Infos,Res)),fail.

add_texpr_info_if_new(b(Expr,Type,Old),Info,b(Expr,Type,New)) :- !,
    add_info_if_new(Old,Info,New).
add_texpr_info_if_new(Other,Infos,Res) :-
    add_internal_error('Illegal call, not BExpr:',add_texpr_info_if_new(Other,Infos,Res)),fail.
% add multiple infos:
add_texpr_infos_if_new(b(Expr,Type,Old),Infos,b(Expr,Type,New)) :-
   add_infos(Infos,Old,New).

% add to info list:
add_info_if_new(Old,Info,New) :-
    (member(Info,Old) -> New=Old ; New = [Info|Old]).

add_infos([]) --> [].
add_infos([Info|T]) --> add_info(Info), add_infos(T).
add_info(Info,Old,New) :-
    (member(Info,Old) -> New=Old ; New = [Info|Old]).


% try and extract the Rodin name of a predicate or expression
get_rodin_name(Expression,Name) :- get_texpr_pos(Expression,rodinpos(Name,_)), Name \= [].
get_rodin_name(Expression,Name) :- get_texpr_pos(Expression,rodinpos(_Model,Name,_)), Name \= []. % new rodinpos

get_rodin_model_name(Expression,Model) :-
   get_texpr_pos(Expression,rodinpos(Model,Name,_)), Name \= []. % new rodinpos

% try and extract the Rodin name or label pragma of a predicate or expression
get_texpr_labels(TExpr,Label) :-
    get_texpr_info(TExpr,Infos),
    member(I,Infos), info_label(I,Label).
get_info_labels(Infos,Label) :-
    member(I,Infos), info_label(I,Label).
info_label(nodeid(rodinpos(Name,_)),[Name]) :- Name \= [].
info_label(nodeid(rodinpos(Model,Name,_)),[FullName]) :- Name \= [], % new rodinpos
    ajoin([Model,':',Name], FullName). % TO DO: if only one level; don't do this
info_label(label(LabelList),LabelList).

add_labels_to_texpr(E,[],R) :- !, R=E. % no labels to add
add_labels_to_texpr(b(E,T,I),Labels,b(E,T,NewI)) :-
    (select(label(OldList),I,Rest)
      -> append(Labels,OldList,NewList), NewI=[label(NewList)|Rest]
      % TO DO: what if we have Rodin labels
      ; NewI = [label(Labels)|I]
    ).

get_texpr_label(TExpr,Label) :- get_texpr_labels(TExpr,Labels), member(Label,Labels).

get_texpr_description(TExpr,Description) :-
    get_texpr_info(TExpr,Infos),
    memberchk(description(Description),Infos).

% check if an info list has an ignore pragma
info_has_ignore_pragma(Infos) :-
  member(description(D),Infos), D='prob-ignore'.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert_must_succeed( (E = "1/2",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "3 mod 2",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "1/0",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2*(3 mod card({}))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "1/(2+1-3)",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "max({1,2,3,0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "min({1,2,3,0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "min({1,2,3,0}-(0..4))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2+max({1,2,3, 1/0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none,\+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2-card({1,2,3})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2-card({1,2,3, 1/0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none,\+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "bool(2-size([1,2,3, 1/0])=3)",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==boolean, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "bool(2-size([1,2,3, 1/2, 3 mod 2, 3**2])=size([]))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==boolean, Err==none, always_well_defined(ET) ) ).

always_well_defined_or_disprover_mode(BE) :-
  (preferences:preference(disprover_mode,true) -> true % we can assume all calls are well-defined
    ; always_well_defined(BE)).

always_well_defined(b(E,_,Infos)) :- !,
  (nonmember(contains_wd_condition,Infos) -> true
   ; always_wd(E) -> true  % some special rules
   %; nl, functor(E,F,N), F \= function, print(non_wd(F,N)),nl,nl,fail
   ).
always_well_defined(E) :- add_error_fail(always_well_defined,'Illegal call: ',always_well_defined(E)).


:- assert_must_succeed( always_wd(div( b(integer(2),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_fail( always_wd(div( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_succeed( always_wd(modulo( b(integer(2),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_fail( always_wd(modulo( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_fail( always_wd(modulo( b(integer(2),integer,[]), b(integer(-2),integer,[]) )) ).
:- assert_must_fail( always_wd(modulo( b(integer(-1),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_succeed( always_wd(power_of( b(integer(2),integer,[]), b(integer(3),integer,[]) )) ).
:- assert_must_succeed( always_wd(power_of( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_fail( always_wd(power_of( b(integer(2),integer,[]), b(integer(-1),integer,[]) )) ).

% catch cases where the construct is currently so instantiated that we can determine that it is well-defined
% can happen e.g. during closure compilation
:- use_module(custom_explicit_sets,[check_in_domain_of_avlset/2]).
:- use_module(kernel_tools,[ground_value/1]).
always_wd(power_of(X,Y)) :- integer_value(Y,Val), number(Val),Val>=0, always_well_defined(X).
always_wd(div(X,Y)) :-  integer_value(Y,Val), number(Val),Val\=0, always_well_defined(X).
always_wd(modulo(X,Y)) :- integer_value(Y,Val), number(Val),Val>0,
   integer_value(X,ValX), number(ValX), ValX>=0. % TO DO: maybe treat this differently in TLA / Event-B minor mode ?
always_wd(function(X,Y)) :- nonvar(X),
   X= b(value(AVLSET),_,_), nonvar(AVLSET), AVLSET=avl_set(AVL),
   always_wd_avl_function(AVL,Y).
always_wd(min(X)) :- non_empty_wd_set_value(X).
always_wd(max(X)) :- non_empty_wd_set_value(X).
always_wd(card(X)) :- finite_wd_set_value(X).
always_wd(size(X)) :- finite_wd_seq_value(X).
always_wd(first(X)) :- non_empty_wd_seq_value(X).
always_wd(front(X)) :- non_empty_wd_seq_value(X).
always_wd(last(X)) :- non_empty_wd_seq_value(X).
always_wd(tail(X)) :- non_empty_wd_seq_value(X). % TO DO: add restrict_front, restrict_tail ?
always_wd(general_intersection(X)) :- non_empty_wd_set_value(X).
always_wd(value(_)). % we have already computed the value and will raise WD error; to be 100 % safe we could restrict this to ground values
% other candidates: size(_), first(_) last(_) tail(_) front(_) restrict_front(_,_) restrict_tail(_,_) rel_iterate(_,_)
always_wd(typeset).
% operators that are always wd on their own:
always_wd(record_field(RecEx,FieldName)) :- ground(FieldName),
   always_well_defined(RecEx).
always_wd(unary_minus(A)) :- always_well_defined(A).
always_wd(unary_minus_real(A)) :- always_well_defined(A).
always_wd(first_of_pair(A)) :- always_well_defined(A).
always_wd(add(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(add_real(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(minus(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(minus_real(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(multiplication(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(multiplication_real(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(equal(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(not_equal(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(less_equal(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(less_equal_real(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(greater_equal(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(less(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(less_real(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(greater(A,B)) :- always_well_defined(A),always_well_defined(B).
always_wd(couple(A,B)) :- always_well_defined(A),always_well_defined(B).
% TO DO: add more/all other operators ?

always_wd_avl_function(AVL,Y) :- nonvar(Y), Y= b(value(Val),_,_),
   ground_value(Val), !,
   check_in_domain_of_avlset(Val,AVL). % ,print(ok),nl.
always_wd_avl_function(AVL,Y) :- always_well_defined(Y),
   custom_explicit_sets:quick_definitely_maximal_total_function_avl(AVL).

integer_value(b(E,integer,_),Val) :- integer_value2(E,Val).
integer_value2(integer(X),R) :- !, R=X.
integer_value2(value(int(X)),R) :- !, R=X.
%integer_value2(X,_) :- print(not_integer_value2(X)),nl,fail.

non_empty_wd_set_value(b(E,_,_)) :- non_empty_set2(E).
non_empty_set2(value(X)) :- definitely_not_empty(X).
non_empty_set2(set_extension(S)) :- l_always_well_defined(S). % the set_extension could contain wd_errors !!
non_empty_set2(sequence_extension(S)) :- l_always_well_defined(S). % ditto
definitely_not_empty(X) :- var(X),!,fail.
definitely_not_empty(avl_set(_)).
definitely_not_empty([_|_]).

non_empty_wd_seq_value(b(E,_,_)) :- non_empty_seq2(E).
non_empty_seq2(value(X)) :- definitely_not_empty_seq(X).
non_empty_seq2(sequence_extension(S)) :- l_always_well_defined(S), S=[_|_].

definitely_not_empty_seq(X) :- var(X),!,fail.
definitely_not_empty_seq(avl_set(A)) :- custom_explicit_sets:is_avl_sequence(A).
definitely_not_empty_seq([El1|T]) :- T==[],nonvar(El1), El1=(IDX,_), IDX==int(1). % TO DO: add more cases ?


finite_wd_set_value(b(E,_,_)) :- finite_set2(E).
finite_set2(empty_set).
finite_set2(empty_sequence).
finite_set2(value(X)) :- X==[] -> true ; definitely_not_empty(X).
finite_set2(set_extension(S)) :- l_always_well_defined(S). % the set_extension could contain wd_errors !!
finite_set2(sequence_extension(S)) :- l_always_well_defined(S). % ditto

finite_wd_seq_value(b(E,_,_)) :- finite_seq2(E).
finite_seq2(empty_set).
finite_seq2(empty_sequence).
finite_seq2(value(X)) :- finite_seq_value(X).
finite_seq2(sequence_extension(S)) :- l_always_well_defined(S).

finite_seq_value(X) :- var(X),!,fail.
finite_seq_value([]).
finite_seq_value(avl_set(A)) :- custom_explicit_sets:is_avl_sequence(A).
% it could be expensive to check if non empty list is a B sequence ??

l_always_well_defined([]).
l_always_well_defined([H|T]) :- always_well_defined(H), l_always_well_defined(T).

is_truth(b(F,pred,_)) :- is_truth_aux(F).
is_truth_aux(truth).
is_truth_aux(value(V)) :- V==pred_true. % can occur in CSE mode

is_falsity(b(F,pred,_)) :- is_falsity_aux(F).
is_falsity_aux(falsity).
is_falsity_aux(value(V)) :- V==pred_false. % can occur in CSE mode

% conjunction of a list of predicates
% NOTE: bsyntaxtree:conjunct_predicates([P1,P2,P3],R).  -->  generates R = b(conjunct(b(conjunct(P1,P2),pred,[]),P3),pred,[])
conjunct_predicates(V,R) :- var(V),!, add_internal_error('Variable conjunction list: ',conjunct_predicates(V,R)),fail.
conjunct_predicates([],b(truth,pred,[])).
conjunct_predicates([P|Rest],Result) :- conjunct2(Rest,P,Result).
conjunct2([],P,P).
conjunct2([Q|Rest],P,Result) :- conjunct3(P,Q,R), conjunct2(Rest,R,Result).
conjunct3(b(truth,_,_),P,P) :- !.
conjunct3(P,b(truth,_,_),P) :- !.
conjunct3(A,B,b(conjunct(A,B),pred,NewInfo)) :- extract_info(A,B,NewInfo).

% disjunction of a list of predicates
disjunct_predicates([],b(falsity,pred,[])).
disjunct_predicates([P|Rest],Result) :- disjunct2(Rest,P,Result).
disjunct2([],P,P).
disjunct2([Q|Rest],P,Result) :- disjunct3(P,Q,R), disjunct2(Rest,R,Result).
disjunct3(b(falsity,_,_),P,R) :- !, R=P.
disjunct3(b(truth,T,I),_,R) :- !, R=b(truth,T,I).
disjunct3(P,b(falsity,_,_),R) :- !, R=P.
disjunct3(_,b(truth,T,I),R) :- !, R=b(truth,T,I).
disjunct3(A,B,b(disjunct(A,B),pred,NewInfo)) :- extract_info(A,B,NewInfo).

% conjunct two predicates and try and construct position information
conjunct_predicates_with_pos_info(A,B,AB) :-
   conjunct_predicates([A,B],AB0),
   (try_get_merged_position_info(A,B,ABI)
    -> add_texpr_infos(AB0,[ABI],AB)  %,print(abi(ABI)),nl
     ; AB=AB0).

% disjunct two predicates and try and construct position information
disjunct_predicates_with_pos_info(A,B,AB) :-
   disjunct_predicates([A,B],AB0),
   (try_get_merged_position_info(A,B,ABI)
    -> add_texpr_infos(AB0,[ABI],AB)  %,print(abi(ABI)),nl
     ; AB=AB0).

try_get_merged_position_info(b(_,_,I1),b(_,_,I2),PosInfo) :-
   (try_get_merged_position_info_aux(I1,I2,MergedInfo) -> PosInfo = MergedInfo
    ; get_non_label_posinfo(Pos1,I1) -> PosInfo=nodeid(Pos1)
    ; get_non_label_posinfo(Pos2,I2) -> PosInfo=nodeid(Pos2)
   ).
try_get_merged_position_info_aux(I1,I2,nodeid(pos(C,Filenumber,Srow,Scol,Erow,Ecol))) :-
    member(nodeid(pos(C1,Filenumber,Srow1,Scol1,Erow1,Ecol1)),I1),
    (number(C1),number(Srow1),number(Scol1),number(Erow1),number(Ecol1)
     -> true ; add_internal_error('Info field 1 not yet instantiated: ',try_get_merged_position_info(I1)),fail),
    !,
    member(nodeid(pos(C2,Filenumber,Srow2,Scol2,Erow2,Ecol2)),I2),
    (number(C2),number(Srow2),number(Scol2),number(Erow2),number(Ecol2)
     -> true ; add_internal_error('Info field 2 not yet instantiated: ',try_get_merged_position_info(I2)),fail),
    !,
    % merge position info if in same file
    (C1 =< C2 -> C=C1, Srow=Srow1,Scol=Scol1 ; C=C2, Srow=Srow2,Scol=Scol2),
    ((Erow1 > Erow2 ; Erow1=Erow2, Ecol1 >= Ecol2)
       -> Erow =Erow1, Ecol=Ecol1
       ;  Erow =Erow2, Ecol=Ecol2).

% get position info which is not a label; label info should not be propagated to outer conjuncts/disjuncts/...
get_non_label_posinfo(Pos,Infos) :- member(nodeid(Pos),Infos),
    \+ functor(Pos,rodinpos,_).

get_texpr_non_label_posinfo(b(_,_,Infos),nodeid(Pos)) :- get_non_label_posinfo(Pos,Infos).

% conjunct list of predicates and try and construct position information
conjunct_predicates_with_pos_info([H|T],Res) :-
   last(T,Last),
   try_get_merged_position_info(H,Last,MergedPosInfo),
   !,
   conjunct_predicates([H|T],Res0),
   add_texpr_infos(Res0,[MergedPosInfo],Res).
conjunct_predicates_with_pos_info(L,Res) :- conjunct_predicates(L,Res).

% a version of create_texpr which collects automatically important infos from sub-expressions
safe_create_texpr(Expr,Type,b(Expr,Type,Info)) :- %
   (sub_expression_contains_wd_condition(Expr) -> Info = [contains_wd_condition] ; Info=[]).

safe_create_texpr(Expr,Type,Info,b(Expr,Type,FullInfo)) :- %
   (sub_expression_contains_wd_condition(Expr) -> FullInfo = [contains_wd_condition|Info] ; FullInfo=Info).

sub_expression_contains_wd_condition(Expr) :- sub_expression_contains_wd_condition(Expr,_).
sub_expression_contains_wd_condition(Expr,Sub) :-
   safe_syntaxelement(Expr,Subs,_Names,_L,_C),
   member(b(Sub,_,Infos),Subs),
   (var(Infos)
    -> add_internal_error('Typed expression not sufficiently instantiated (variable Infos): ',sub_expression_contains_wd_condition(Expr)),
       fail
    ; memberchk(contains_wd_condition,Infos)).

% provide updated infos (e.g., reads(...)) and remove any old info fields with same functor
update_infos([],Infos,Infos).
update_infos([Update|T],OldInfos,NewInfos) :-
   functor(Update,F,N),
   functor(Old,F,N),
   delete(OldInfos,Old,OldInfos1),
   update_infos(T,[Update|OldInfos1],NewInfos).

% merge two info lists and try to reconcile position information:
merge_info(Info1,Info2,Res) :-
   merge_imp_info2(Info2,Info1,NewInfo),
   (try_get_merged_position_info_aux(Info1,Info2,NewPos)
    -> delete_pos_info(NewInfo,NI2),
       Res = [NewPos|NI2]
     ; Res = NewInfo).

% extract important info but without used_ids:
extract_info_wo_used_ids(b(_,_,Info1),Info) :-
  extract_pos_infos(Info1,PosInfos),
  extract_just_important_info_aux(Info1,[],I1),
  append(PosInfos,I1,Info).
extract_info_wo_used_ids_and_pos(b(_,_,Info1),b(_,_,Info2),Info) :-
   extract_just_important_info_aux(Info1,[],I1),
   extract_just_important_info_aux(Info2,I1,Info).

% extract important info from one sub-expression:
extract_info(b(_,_,Info1),NewInfo) :-
   extract_imp_info1(Info1,I1),!, NewInfo = I1.
extract_info(A,R) :-
   add_internal_error('Could not extract info: ',extract_info(A,R)),
   R=[].
% extract imortant info fields from two (sub-)expressions
extract_info(b(_,_,Info1),b(_,_,Info2),NewInfo) :-
   merge_imp_info2(Info1,Info2,II),!, NewInfo = II.
%:- use_module(library(ordsets),[ord_intersection/3]).
%   ord_intersection(Info1,Info2,NewInfo),!.
% TO DO: should we merge nodeid position information !?
extract_info(A,B,R) :- add_internal_error('Could not extract info: ',extract_info(A,B,R)),
 R=[].

% extract important info and used_ids
extract_imp_info1([],[]).
extract_imp_info1([H|T],Res) :-
  ((important_info(H); H=used_ids(_)) -> Res=[H|ET], extract_imp_info1(T,ET)
   ; extract_imp_info1(T,Res)
   ).

% extract and merge important info and try to merge used_ids from two info lists
merge_imp_info2(Info1,Info2,ResInfos) :-
   extract_imp_info1(Info1,I1),
   extract_imp_info1(Info2,I2),
   (select(used_ids(Ids1),I1,II1)
     -> (select(used_ids(Ids2),I2,II2)
          ->  ord_union(Ids1,Ids2,Ids3),
              append(II1,[used_ids(Ids3)|II2],II)
           ;  append(II1,I2,II)
        )
     ;  delete(I2,used_ids(_),II2),
        append(I1,II2,II)
   ),
   sort(II,ResInfos). % TO DO: sort sublists ? and use ord_union?

important_info(contains_wd_condition).
important_info(prob_annotation(_)).
important_info(allow_to_lift_exists).
important_info(removed_typing).
%important_info(lambda_result(_)). % should probably not be copied

% variation of above which does not extract and merge used_ids:
extract_just_important_info_aux([],Acc,Acc).
extract_just_important_info_aux([H|T],Acc,Res) :-
  (important_info(H), \+ member(H,Acc) -> extract_just_important_info_aux(T,[H|Acc],Res)
    ; extract_just_important_info_aux(T,Acc,Res)).


is_a_conjunct(b(conjunct(A,B),pred,_),A,B).
% use is_a_conjunct_without_label if you want to avoid decomposing conjunction associated with a single label
is_a_conjunct_without_label(b(conjunct(A,B),pred,I),A,B) :- \+ get_info_labels(I,_).
% use decompose conjunct if you want to propagate labels down to the conjuncts
decompose_conjunct(b(conjunct(A,B),pred,I),ResA,ResB) :-
   (get_info_labels(I,Labels)
    -> add_labels_to_texpr(A,Labels,ResA), add_labels_to_texpr(B,Labels,ResB)
    ; ResA=A, ResB=B).

conjunction_to_list(C,L) :- var(C),!,
   add_error_fail(conjunction_to_list,'Internal error: var :',conjunction_to_list(C,L)).
conjunction_to_list(C,List) :-
    conjunction_to_list2(C,List,[]).
conjunction_to_list2(X,I,O) :- X=b(E,_,_),
   (E=conjunct(LHS,RHS) -> conjunction_to_list2(LHS,I,Inter),
                           conjunction_to_list2(RHS,Inter,O)
    ; E = truth -> I=O
    ; I = [X|O]).

member_in_conjunction(M,Conj) :- is_a_conjunct(Conj,LHS,RHS),!,
   (member_in_conjunction(M,LHS) ; member_in_conjunction(M,RHS)).
member_in_conjunction(M,M).

% a version of member_in_conjunction which can deal with lazy_let_pred :
% member_in_conjunction_cse(FullConjunctWithLets,InnerConjunctNotALet,  Conjunction)
member_in_conjunction_cse(M,InnerConj,Conj) :- is_a_conjunct(Conj,LHS,RHS),!,
   (member_in_conjunction_cse(M,InnerConj,LHS) ; member_in_conjunction_cse(M,InnerConj,RHS)).
member_in_conjunction_cse(Conj,InnerConj,b(lazy_let_pred(ID,Share,Body),pred,Info)) :-
   Conj = b(lazy_let_pred(ID,Share,BConj),pred,Info),!,
   member_in_conjunction_cse(BConj,InnerConj,Body).
member_in_conjunction_cse(M,M,M).

% not terribly efficient way to select and remove a conjunct from a predicate
select_member_in_conjunction(M,Conj,Rest) :- is_a_conjunct(Conj,LHS,RHS),!,
   (select_member_in_conjunction(M,LHS,RL), conjunct_predicates([RL,RHS],Rest)
    ; select_member_in_conjunction(M,RHS,RR), conjunct_predicates([LHS,RR],Rest)).
select_member_in_conjunction(M,M,b(truth,pred,[])).

is_a_disjunct(b(disjunct(A,B),pred,_),A,B).
is_a_negation(b(negation(A),pred,_),A).

disjunction_to_list(C,L) :- var(C),!,
   add_error_fail(disjunction_to_list,'Internal error: var :',disjunction_to_list(C,L)).
disjunction_to_list(C,List) :-
    disjunction_to_list2(C,List,[]).
disjunction_to_list2(C,I,O) :- is_a_disjunct(C,LHS,RHS),!,
   disjunction_to_list2(LHS,I,Inter),
   disjunction_to_list2(RHS,Inter,O).
disjunction_to_list2(X,[X|R],R).

is_an_implication(b(implication(A,B),pred,_),A,B).

is_an_equivalence(b(equivalence(A,B),pred,_),A,B).


is_a_disjunct_or_implication(b(DI,pred,_),Type,A,B) :- is_a_disj_or_impl_aux(DI,Type,A,B).
is_a_disj_or_impl_aux(disjunct(A,B),'disjunction',A,B).
is_a_disj_or_impl_aux(implication(A,B),'implication',NA,B) :-
    create_negation(A,NA).
is_a_disj_or_impl_aux(negation(b(conjunct(A,B),pred,_)),'negated conjunction',NA,NB) :-
    create_negation(A,NA),create_negation(B,NB).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove all info fields
remove_all_infos(TExpr,TNExpr) :-
    var(TExpr),!,TExpr=TNExpr.
remove_all_infos(TExpr,TNExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),
    create_texpr(NExpr,Type,_,TNExpr),
    maplist(remove_all_infos,Subs,NSubs).

% replace all info fields
remove_all_infos_and_ground(TExpr,TNExpr) :-
    var(TExpr),!,TExpr=TNExpr.
remove_all_infos_and_ground(TExpr,TNExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),
    create_texpr(NExpr,Type,[],TNExpr),
    maplist(remove_all_infos_and_ground,Subs,NSubs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a version of create_exists_opt which also detects top-level disjunctions
create_exists_opt_disj(Ids,b(disjunct(A,B),pred,Info1),NewPred) :- !,
    create_exists_opt_disj(Ids,A,PA), create_exists_opt_disj(Ids,B,PB),
    disjunct_predicates_with_pos_info(PA,PB,b(NP,pred,Info2)),
    NewPred=b(NP,pred,NewInfo),
    extract_just_important_info_aux(Info1,Info2,NewInfo). % copy symbolic or other relevant info
create_exists_opt_disj(Ids,Pred,NewPred) :- conjunction_to_list(Pred,L),
   create_exists_opt(Ids,L,NewPred).

% create_exists_opt(TIds,Preds,NewPred)
%   creating an existential quantification with some optimisations.
% TIds: The typed identifiers that are quantified
% Preds: A list of predicates
% NewPred: The (optimised) quantified expression
% Basically two optimisations are performed:
%  - identifiers that are not used at all are removed from the quantifier
%  - predicates that do not use one of the quantified identifiers are moved
%    outside the quantification, resulting in a predicate of the form "P & #x.Q(x)"

create_exists_opt(TIds,Preds,NewPred) :-
    create_exists_opt(TIds,Preds,[],NewPred,_).

create_exists_opt(TIds,Preds,NewPred,Modified) :-
    create_exists_opt(TIds,Preds,[],NewPred,Modified).
create_exists_opt([],Preds,_,NewPred,Modified) :- !, Modified = false,
    conjunct_predicates(Preds,NewPred).
%create_exists_opt(TIds,Preds,AdditionalInfos,NewPred,Modified) :-
%   preference(data_validation_mode,true),
%   % no not lift predicates outside of existential quantifiers and change order; see rule_avec_DV.mch ClearSy example (N_ITERa_avec_DV_sans_DV) and test 1945
%   !, Modified=false, % print(not_optimizing_exists(TIds)),nl,
%   conjunct_predicates(Preds,P),
%   create_exists_detect_tautology_aux(TIds,P,AdditionalInfos,NewPred).
create_exists_opt(TIds,Preds,AdditionalInfos,NewPred2,Modified) :-
    get_texpr_ids(TIds,UnsortedIds), sort(UnsortedIds,Ids),
    create_exists_opt2(Preds,first,Ids,[],Inner,Outer,UsedIds),
    % Inner: quantified part, Outer: part which does not have to be quantified
    ( Inner = []    -> AllPreds=Outer % no exists needed
    ; % Inner = [_|_] ->
        ceo_filter_used_ids(TIds,UsedIds,UsedTIds), % filter out unused variables
        conjunct_predicates_with_pos_info(Inner,I),
        create_exists_detect_tautology_aux(UsedTIds,I,AdditionalInfos,Exists),
        % remove_typing also added in create_exists_detect_tautology_aux
        append(Outer,[Exists],AllPreds)
    ),
    (Outer=[_|_] -> Modified=true % TO DO: check if TIds=UsedTIds
      ; UsedIds == Ids -> Modified= false
      ; Modified=true),
    conjunct_predicates(AllPreds,NewPred2).
create_exists_opt2([],_,_Ids,UsedIds,[],[],UsedIds) :- !.
create_exists_opt2([Pred|Prest],First,Ids,UsedIdsIn,Inner,Outer,UsedIdsOut) :- !,
    find_identifier_uses_if_necessary(Pred, [], LocalUses),
    ord_intersection(Ids,LocalUses,LocalExUses),
    ( LocalExUses = [],
      (First==first -> true
       ; preference(data_validation_mode,false)) % see test 1945; step 39 in trace replay
       % always_well_defined_or_disprover_mode(Pred) % TODO: we should probably also check this !
       ->
        Inner = Irest,
        Outer = [Pred|Orest], % move Pred outside of existential quantifier as it does not depend on quantified variables
        % format('MOVED out of #(~w): ',[Ids]),translate:print_bexpr(Pred), format('  Uses ~w   [~w]~n',[LocalUses,LocalExUses]),
        UsedIdsIn = Urest, F1=first
    ; LocalExUses = [_|_] ->
        Inner = [Pred|Irest],
        Outer = Orest,
        ord_union(UsedIdsIn,LocalExUses,Urest),
        F1 = not_first_anymore
    ),
    create_exists_opt2(Prest,F1,Ids,Urest,Irest,Orest,UsedIdsOut).
create_exists_opt2(Pred,First,Ids,UsedIdsIn,Inner,Outer,UsedIdsOut) :-
   add_error(create_exists_opt,'Expecting predicate list: ',Pred),
   create_exists_opt2([Pred],First,Ids,UsedIdsIn,Inner,Outer,UsedIdsOut).

% construct exists optimized: keep only used TIds
ceo_filter_used_ids([],_UsedIds,[]).
ceo_filter_used_ids([TId|Irest],UsedIds,Filtered) :-
    get_texpr_id(TId,Id),
    ( ord_member(Id,UsedIds) -> Filtered = [TId|Frest] % id used: keep it
    ; otherwise              -> Filtered = Frest),
    ceo_filter_used_ids(Irest,UsedIds,Frest).

% a version of create_exists which also detects x=E and x:E and x>E tautologies
% we could replace this by a more generic prover
create_exists_detect_tautology_aux([TID],Pred,_,Truth) :-  % print(create(TID,Pred)),nl,
    is_equal(Pred,LHS,RHS),
    get_texpr_id(LHS,ID), get_texpr_id(TID,ID),
    \+ occurs_in_expr(ID,RHS),
    always_well_defined_or_disprover_mode(RHS),
    !, % replace #x.(x=E) by TRUTH
    debug_format(19,'Detected tautology exists equality over ~w~n',[ID]),
    Truth=b(truth,pred,[]).
create_exists_detect_tautology_aux([TID],b(member(LHS,RHS),pred,_),_,Truth) :-
    get_texpr_id(LHS,ID), get_texpr_id(TID,ID),
    definitely_not_empty_set(RHS),
    always_well_defined_or_disprover_mode(RHS),
    !, % replace #x.(x:RHS) by TRUTH
    debug_format(19,'Detected tautology exists membership over ~w~n',[ID]),
    Truth=b(truth,pred,[]).
create_exists_detect_tautology_aux([TID],b(Pred,_,_),_,Truth) :-
    is_comparison(Pred,L,R), get_texpr_id(TID,ID),
    ((LHS,RHS)=(L,R) ; (LHS,RHS)=(R,L)),
    get_texpr_id(LHS,ID),
    \+ occurs_in_expr(ID,RHS),
    always_well_defined_or_disprover_mode(RHS),
    !, % replace #x.(x COMP E) by TRUTH
    debug_format(19,'Detected tautology comparison over ~w~n',[ID]),
    Truth=b(truth,pred,[]).
%create_exists_detect_tautology_aux(Ids,b(disjunct(A,B),pred,Info1),AdditionalInfos,NewPred) :- !,
%    create_exists(Ids,A,PA), create_exists(Ids,B,PB),
%    disjunct_predicates_with_pos_info(PA,PB,b(NP,pred,Info2)),
%    NewPred=b(NP,pred,NewInfo),
%    print(disj_exists2), translate:print_bexpr(b(NP,pred,Info2)),nl, trace,
%    extract_just_important_info_aux(Info1,Info2,NewInfo). % copy symbolic or other relevant info
create_exists_detect_tautology_aux(Ids,Pred,AdditionalInfos,NewPred2) :-
    create_exists_or_let_predicate(Ids,Pred,NewPred),
    add_texpr_infos_if_new(NewPred,[removed_typing|AdditionalInfos],NewPred2).
    % removed_typing to avoid spurious exists_body_warning, see test 1681, 625

is_equal(b(equal(LHS,RHS),pred,_),A,B) :-
    ((A,B)=(LHS,RHS) ; (A,B)=(RHS,LHS)).

is_comparison(greater(A,B),A,B).
is_comparison(less(A,B),A,B).
is_comparison(greater_equal(A,B),A,B).
is_comparison(less_equal(A,B),A,B).

% is true if a predicate Pred can be split into two parts:
% Outer which does not depend on LocalIds (can be lifted out) and Inner which does
detect_global_predicates(LocalIds,Pred,Outer,Inner) :-
    get_texpr_ids(LocalIds,UnsortedIds), sort(UnsortedIds,Ids),
    conjunction_to_list(Pred,Preds),
    split_predicate_local_global(Preds,Ids,OuterL,InnerL),
    OuterL \= [],
    conjunct_predicates(OuterL,Outer),
    conjunct_predicates(InnerL,Inner).
split_predicate_local_global([],_Ids,[],[]).
split_predicate_local_global([P|Ps],Ids,Outer,Inner) :-
    (is_local_predicate(Ids,P)
      -> Inner = [P|Is], split_predicate_local_global(Ps,Ids,Outer,Is)
      ;  Outer = [P|Os], split_predicate_local_global(Ps,Ids,Os,Inner)).
is_local_predicate(Ids,Pred) :-
    find_identifier_uses_if_necessary(Pred, [], LocalUses),
    ord_intersect(Ids,LocalUses).

%:- use_module(b_global_sets,[b_global_set/1]).
definitely_not_empty_set(b(SET,T,_)) :- not_empty_set_aux(SET,T).
not_empty_set_aux(bool_set,_).
not_empty_set_aux(integer_set(_),_).
not_empty_set_aux(float_set,_).
not_empty_set_aux(real_set,_).
not_empty_set_aux(string_set,_).
not_empty_set_aux(set_extension(X),_) :- dif(X,[]).
not_empty_set_aux(sequence_extension(X),_) :- dif(X,[]).
not_empty_set_aux(value(S),_) :- nonvar(S), (S=avl_set(_) -> true ; S=[_|_]).
%not_empty_set_aux(identifier(X),set(global(X))) :- bmachine:b_get_machine_set(X). % what if we have a local variable ? ENSURE THAT WE DO NOT ALLOW identifier X to stand for global set X; see ExistentialGlobalSetIDTest in Tester % also: b_global_set not yet computed when ast_cleanup runs on startup !
% TO DO: determine which identifier(X) refer to global set names
not_empty_set_aux(interval(A,B),_) :- get_texpr_expr(A,integer(IA)), get_texpr_expr(B,integer(IB)), IA =< IB.


% a simple let-detection
create_exists_or_let_predicate([H|T],b(conjunct(LHS,RHS),pred,I),NewPred) :- get_texpr_id(H,ID),
   is_equal(LHS,TID,IDEXPR), % TO DO: should we do a more complicated check here ? exist technique useful for SLOT-24 codespeed test
   get_texpr_id(TID,ID), \+ occurs_in_expr(ID,IDEXPR),
   maplist(not_occurs_in_expr(IDEXPR),T),
   !,
   NewPred = b(let_predicate([TID],[IDEXPR],Body),pred,I),
   %print('LET: '),translate:print_bexpr(NewPred),nl,
   create_exists_or_let_predicate(T,RHS,Body).
create_exists_or_let_predicate(Ids,Pred,NewPred) :- create_exists(Ids,Pred,NewPred).

not_occurs_in_expr(IDEXPR,TID) :- get_texpr_id(TID,ID), \+ occurs_in_expr(ID,IDEXPR).

create_exists([],Pred,NewPred) :- !,Pred=NewPred.
create_exists(Ids,Pred,NewPred) :-
    find_identifier_uses_for_quantifier_body(Ids,Pred, Used),
    extract_info_wo_used_ids(Pred,NewImportantInfo),
    create_texpr(exists(Ids,Pred),pred,[used_ids(Used)|NewImportantInfo],NewPred).


create_forall([],Pred,NewPred) :- !,Pred=NewPred.
create_forall(Ids,Pred,NewPred) :-
    find_identifier_uses_for_quantifier_body(Ids,Pred, Used),
    split_forall_body(Pred,LHS,RHS),
    extract_info_wo_used_ids_and_pos(LHS,RHS,NewImportantInfo),
    create_texpr(forall(Ids,LHS,RHS),pred,[used_ids(Used)|NewImportantInfo],NewPred).
split_forall_body(b(implication(LHS,RHS),_,_),LHS,RHS) :- !.
split_forall_body(RHS,b(truth,pred,[]),RHS).

create_implication(b(truth,pred,_),P,Res) :- !, Res=P.
create_implication(b(falsity,pred,_),P,Res) :- !, create_negation(P,Res).
create_implication(Lhs,Rhs,b(implication(Lhs,Rhs),pred,NewInfo)) :-
   extract_info(Lhs,Rhs,NewInfo).

create_equivalence(Lhs,Rhs, b(equivalence(Lhs,Rhs),pred,NewInfo)) :-
   extract_info(Lhs,Rhs,NewInfo).

create_negation(b(P,pred,I),Res) :- create_negation_aux(P,I,R),!,Res=R.
create_negation(Pred,b(negation(Pred),pred,NewInfo)) :-
   extract_info(Pred,Infos),
   (get_texpr_non_label_posinfo(Pred,Pos) -> NewInfo = [Pos|Infos] ; NewInfo=Infos).

create_negation_aux(truth,I,R) :- !, R=b(falsity,pred,I).
create_negation_aux(falsity,I,R) :- !, R=b(truth,pred,I).
create_negation_aux(negation(Pred),_,R) :- R=Pred. % not(not(P)) = P
% we could add some rules about negating member <-> not_member, ... but be careful with effects on is_negation_of

% check if something is the negation of something else (quite stupid at the moment)
% used, e.g., to detect IF-THEN-ELSE constructs in b_ast_cleanup
is_negation_of(P,NP) :-
   create_negation(P,NotP), % works both with not(A),A  or A,not(A)
   same_texpr(NotP,NP).
is_negation_of(b(P,pred,_),b(NP,pred,_)) :- is_negation_aux(P,NP).

is_negation_aux(equal(A,B),NP) :- is_negation_of_equality(NP,A,B).
is_negation_aux(not_equal(A,B),NP) :- is_negation_of_disequality(NP,A,B).
is_negation_aux(subset(XA,SA),not_subset(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(not_subset(XA,SA),subset(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(subset_strict(XA,SA),not_subset_strict(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(not_subset_strict(XA,SA),subset_strict(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(member(XA,SA),not_member(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(not_member(XA,SA),member(XB,SB)) :- same_texpr(XA,XB), same_texpr(SA,SB).
is_negation_aux(less(A,B),greater_equal(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less(A,B),less_equal(BB,AA)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less_equal(B,A),less(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less_equal(A,B),greater(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(greater_equal(A,B),NP) :- is_negation_aux(less_equal(B,A),NP).
is_negation_aux(greater(A,B),NP) :- is_negation_aux(less(B,A),NP).
is_negation_aux(less_equal_real(B,A),less_real(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(less_real(B,A),less_equal_real(AA,BB)) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_aux(negation(A),negation(B)) :- is_negation_of(A,B).
is_negation_aux(truth,falsity).
is_negation_aux(falsity,truth).
% TO DO: detect more ?? x < y   <=>  not ( x > y-1 )

is_negation_of_equality(not_equal(AA,BB),A,B) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_of_equality(equal(AA,BB),A,B) :- same_texpr(A,AA), neg_value(BB,B).

neg_value(b(boolean_true,_,_),b(boolean_false,_,_)).
neg_value(b(boolean_false,_,_),b(boolean_true,_,_)).

is_negation_of_disequality(equal(AA,BB),A,B) :- same_texpr(A,AA), same_texpr(B,BB).
is_negation_of_disequality(not_equal(AA,BB),A,B) :- same_texpr(A,AA), neg_value(BB,B).


create_equality(b(_,TA,_),b(_,TB,_),_) :- TA \= TB, \+ unify_types_strict(TA,TB),!,
        add_internal_error('Creating equality with incompatible types:',equal(TA,TB)),fail.
create_equality(A,B,Equality) :-
        safe_create_texpr(equal(A,B),pred,Equality).

create_couple(A,B,b(couple(A,B),couple(TA,TB),Infos)) :-
      get_texpr_type(A,TA), get_texpr_type(B,TB),
      extract_info(A,B,Infos).

% couplise_list for typed expression list
create_couple([A],Couple) :- !,A=Couple.
create_couple([A,B|Rest],Couple) :-
    create_couple(A,B,Couple1),
    create_couple([Couple1|Rest],Couple).

% derive typing from Ids
create_comprehension_set(Ids,Pred,Info,CompSet) :-
    get_texpr_types(Ids,Types),
    couplise_list(Types,ElementType),
    create_texpr(comprehension_set(Ids,Pred),set(ElementType),Info,CompSet).

% --------------------------------------


% find uses of identifiers

occurs_in_expr(ID,TExpr) :- var(ID),!,
   add_internal_error('Variable id: ',occurs_in_expr(ID,TExpr)),fail.
occurs_in_expr(ID,TExpr) :- ID=b(_,_,_),!, add_internal_error('Non-atomic identifier: ',occurs_in_expr(ID,TExpr)),fail.
occurs_in_expr(ID,TExpr) :-
 (find_identifier_uses_if_necessary(TExpr,[],Used)
  -> %check_sorted(Used),
     ord_member(ID,Used)
   ; add_failed_call_error(find_identifier_uses(TExpr,[],_)),fail).
  % TO DO: optimize so that we only look for ID; we don't have to keep track of other IDs

%check_sorted(List) :- sort(List,SL), (SL \= List -> add_internal_error('Not sorted:',List) ; true).

some_id_occurs_in_expr([H|T],TExpr) :- (var(H) ; \+ atomic(H)),
    add_internal_error('Must be (sorted) atomic identifiers: ',some_id_occurs_in_expr([H|T],TExpr)),
    fail.
some_id_occurs_in_expr(SortedIDs,TExpr) :-
    find_identifier_uses_if_necessary(TExpr,[],Used),
    list_to_ord_set(Used,UsedSorted),
    ord_intersect(UsedSorted,SortedIDs).

% a special version for quantifier bodies, attempting to reuse used_ids info
find_identifier_uses_for_quantifier_body(TIds,Body,Res) :-
    get_texpr_ids(TIds,Ids),
    list_to_ord_set(Ids,Ignore),
    find_identifier_uses_if_necessary(Body,Ignore,Res).

% only run find_identifier_uses if necessary because no used_ids Info field present
% TO DO: we could consider using used_ids within find_typed_identifier_uses: drawback: we would need to store types
find_identifier_uses_if_necessary(b(_,_,I),Ignore,Res) :- member(used_ids(UIds),I),!, %print('+'),nl,
     %(is_ordset(Ignore) -> true ; add_internal_error('Not ordset: ',Ignore)),
     ord_subtract(UIds,Ignore,Res).
find_identifier_uses_if_necessary(Expr,Ignore,Res) :- find_identifier_uses(Expr,Ignore,Res).

:- use_module(tools,[safe_sort/3]).
find_identifier_uses(TExpr,Ignored,Ids) :-
    %tools_printing:print_term_summary(find_identifier_uses(TExpr,Ignored,Ids)),nl,
    check_is_texpr(TExpr,find_identifier_uses),
    (find_typed_identifier_uses(TExpr,Ignored,TIds)
     -> get_texpr_ids(TIds,RIds),
        % in case of type errors, or when type checking is not yet complete, we can have multiple entries of the same identifier with different types !
        remove_dups(RIds,Ids)
     ;  add_internal_error('Call failed:',find_typed_identifier_uses(TExpr,Ignored,_)),
        Ids=[]
    ).
find_identifier_uses_l(TExprs,Ignored,Ids) :-
    find_typed_identifier_uses_l(TExprs,Ignored,TIds),
    get_texpr_ids(TIds,RIds),
    remove_dups(RIds,Ids).  % see above

check_is_texpr(X,Context) :-
   (get_texpr_expr(X,_) -> true ; add_internal_error('Expected TExpr: ', check_is_texpr(X,Context))).

find_typed_identifier_uses(TExpr,Ids) :-
    get_global_identifiers(Ignored), % ignore all global sets and global constants; hence find_typed_identifier_uses/2 usually should only be called for top level expressions
    find_typed_identifier_uses(TExpr,Ignored,Ids).

find_typed_identifier_uses(TExpr,Ignored,Ids) :- var(TExpr),!,
    add_internal_error('Variable typed expression: ', find_typed_identifier_uses(TExpr,Ignored,Ids)),
    Ids = [].
find_typed_identifier_uses(TExpr,Ignored,Ids) :-
    find_typed_identifier_uses_l([TExpr],Ignored,Ids).

find_typed_identifier_uses2(TExpr,Ignored,Ids,Rest) :-
    %syntaxtraversion_fixed(TExpr,Expr,QSubs,Subs,TNames),
    get_texpr_expr(TExpr,Expr),
    safe_syntaxelement(Expr,Subs,TNames,_,_), % QSubs=[],
    ( uses_an_identifier(Expr,Id,TExpr,Ignored) ->
        ( ord_member(Id,Ignored) -> Ids=Rest
        ; otherwise             ->
            get_texpr_type(TExpr,Type),
            normalize_type(Type,NType), % replace seq by set;
            % warning: when type-check not yet complete we have variables here
            create_texpr(identifier(Id),NType,[],TId),
            %print(adding(Id,NType,Ignored)),nl,
            Ids = [TId|Rest])
    % ; (Expr = becomes_such(TIds,_)), nl,print(uses_primes(Expr)),nl,fail
    ; TNames = [] -> find_typed_identifier_uses2_l(Subs,Ignored,Ids,Rest)
    ; otherwise ->
        %find_typed_identifier_uses2_l(QSubs,Ignored,Ids,Ids2), % useless here as QSubs=[]
        get_texpr_ids(TNames,Names),
        list_to_ord_set(Names,SNames), ord_union(SNames,Ignored,Ignored2),
        find_typed_identifier_uses2_l(Subs,Ignored2,Ids,Rest)).
find_typed_identifier_uses2_l([],_) --> [].
find_typed_identifier_uses2_l([Expr|Rest],Ignored) -->
    find_typed_identifier_uses2(Expr,Ignored),
    find_typed_identifier_uses2_l(Rest,Ignored).

%Note: above we do not remap uses of Id$0 to Id in becomes_such;
%      this is done in find_read_vars_for_becomes_such in b_read_write_info


uses_an_identifier(Expr,Id) :- uses_an_identifier(Expr,Id,none,[]).

uses_an_identifier(identifier(Id),Id,_,_).
uses_an_identifier(lazy_lookup_pred(Id),Id,_,_).
uses_an_identifier(lazy_lookup_expr(Id),Id,_,_).
uses_an_identifier(value(_),Id,b(_,_,Info),Ignored) :- Ignored \= [],
    member(was_identifier(Id),Info),
    member('$examine_value_was_identifier_info',Ignored),!.

find_typed_identifier_uses_l(TExprs,Ignored,Ids) :-
    list_to_ord_set(Ignored,SIgnored),
    find_typed_identifier_uses2_l(TExprs,SIgnored,Unsorted,[]),
    safe_sort(find_typed_identifier_uses,Unsorted,Ids),
    check_typed_ids(Ids).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_safe_mode,true)).
check_typed_ids([]) :- !.
check_typed_ids([b(identifier(ID),T,_)|Tail]) :- !, check_ids3(Tail,ID,T).
check_typed_ids(Other) :- add_internal_error('Unexpected typed id list:',check_typed_ids(Other)).

check_ids3([],_,_) :- !.
check_ids3([b(identifier(ID),T,_)|Tail],ID1,T1) :- !,
   (ID=ID1 -> add_internal_error('Identifier appears multiple times with types:',id(ID,T1,T)) ; true),
   check_ids3(Tail,ID,T).
check_ids3(Other,ID1,T1) :- add_internal_error('Unexpected typed id list:',check_ids3(Other,ID1,T1)).
:- else.
check_typed_ids(_).
:- endif.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% break predicate into components with disjunct identifiers
%  Pred: a predicate
%  Components: a list of terms component/2 where the first argument
%    is a predicate, the second is the set of used identifiers
predicate_components(Pred,Res) :- predicate_components_in_scope(Pred,[],Res).
predicate_components_in_scope(Pred,LocalVars,Res) :-
   predicate_components_with_restriction(Pred,LocalVars,all,Res).

predicate_components_with_restriction(Pred,LocalVars,RestrictionList,Res) :-
    conjunction_to_list(Pred,Preds),
    l_predicate_identifiers(Preds,LocalVars,PredIds),
    ( RestrictionList=all -> R=all
    ; otherwise           -> list_to_ord_set(RestrictionList,R)),
    try_find_and_remove_equalities(PredIds,PredIds2),
    % print(find_components(R)),nl,nl,
    %%(member(pred(P,Ids,X),PredIds2), nl,print(pred(Ids,X)),nl,translate:print_bexpr(P),nl,fail ; true),
    find_components(PredIds2,R,Components),
    !,
    %maplist(print_component,Components),
    Components=Res.
predicate_components_with_restriction(Pred,_,_,[component(Pred,[])]) :-
    add_internal_error('predicate_components failed: ',predicate_components_with_restriction(Pred,_,_,_)).

% print_component(component(Pred,Ids)) :- format('Component over ~w :~n',[Ids]), translate:print_bexpr(Pred),nl.

% get the list of used identifiers for each predicate
l_predicate_identifiers([],_LocalVars,[]).
l_predicate_identifiers([Pred|PRest],LocalVars,[pred(Pred,Ids,_Selected)|IRest]) :-
    predicate_identifiers_in_scope(Pred,LocalVars,Ids),  % Do not ignore local variables; used instead of enumerate set elements
    l_predicate_identifiers(PRest,LocalVars,IRest).


try_find_and_remove_equalities(PredAndIds,PredAndIds2) :-
    preferences:get_preference(partition_predicates_inline_equalities,true),
    preferences:get_preference(try_kodkod_on_load,false),
    find_and_remove_equalities(PredAndIds,RR),
    !, PredAndIds2=RR.
try_find_and_remove_equalities(Ps,Ps).

:- use_module(debug,[debug_println/2]).
% find and apply obvious equalities so that they do not interfere with partitioning into components
% example:  c = 1 &  f: 1..c --> A  &  g: 1..c --> B
% TO DO: preprocess and do one pass to detect potential equalities
find_and_remove_equalities([],[]).
find_and_remove_equalities(List,[pred(P,PIds,Sel)|FT]) :-
    select(pred(P,PIds,Sel),List,Rest),
    %PIds = [Id],
    identifier_equality(P,Id,Value),
    %(value_which_can_be_replaced(Value) -> true ; nl,print('Not replaced: '),translate:print_bexpr(Value),nl,fail),
    (get_texpr_id(Value,Id2)
     -> Id2 \= Id %,print(inline_id(Id,Id2)),nl
        % Note: this inlining does *not* help with partitioning; but does help ProB detect common predicates/expressions
     ; PIds=[Id],value_which_can_be_replaced(Value)),
    debug_println(9,replace_simple_equality(Id,Value,PIds)),
    maplist(apply_to_pred(Id,Value),Rest,RT),
    !,
    find_and_remove_equalities(RT,FT).
% TO DO: detect equalityes x = EXPR, where EXPR does not contain x and where x occurs in no other predicate
% we can then annotate x as not to enumerate
find_and_remove_equalities(R,R).

identifier_equality(b(equal(LHS,RHS),_,_),Id,EqTerm) :-
   (get_texpr_id(LHS,Id) -> EqTerm = RHS
    ; get_texpr_id(RHS,Id) -> EqTerm = LHS).

value_which_can_be_replaced(b(E,T,_)) :- value_which_can_be_replaced2(E,T).
   %(value_which_can_be_replaced2(E,T) -> true ; print(not_val(E)),nl,fail).
value_which_can_be_replaced2(value(_),_).
value_which_can_be_replaced2(integer(_),_).
%value_which_can_be_replaced2(identifier(I),global(G)) :- b_global_constant(G), id_not_used_anywhere(I).
value_which_can_be_replaced2(integer_set(_),_).
value_which_can_be_replaced2(unary_minus(A),_) :- value_which_can_be_replaced(A).
value_which_can_be_replaced2(add(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
% we could compute the value
value_which_can_be_replaced2(minus(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
value_which_can_be_replaced2(multiplication(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
value_which_can_be_replaced2(div(A,B),_) :-
   integer_value(B,IB), number(IB), IB \= 0,
   value_which_can_be_replaced(A).
value_which_can_be_replaced2(floored_div(A,B),T) :- value_which_can_be_replaced2(div(A,B),T).
% should we add:  with WD check: division, modulo, .... ? see also simple2 in b_ast_cleanup
value_which_can_be_replaced2(max_int,_).
value_which_can_be_replaced2(min_int,_).
value_which_can_be_replaced2(float_set,_).
value_which_can_be_replaced2(real(_),_).
value_which_can_be_replaced2(real_set,_).
value_which_can_be_replaced2(string(_),_).
value_which_can_be_replaced2(string_set,_).
value_which_can_be_replaced2(boolean_true,_).
value_which_can_be_replaced2(boolean_false,_).
value_which_can_be_replaced2(bool_set,_).
value_which_can_be_replaced2(empty_set,_).
value_which_can_be_replaced2(empty_sequence,_).
value_which_can_be_replaced2(couple(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
value_which_can_be_replaced2(interval(A,B),_) :- value_which_can_be_replaced(A), value_which_can_be_replaced(B).
value_which_can_be_replaced2(set_extension(L),_) :- maplist(value_which_can_be_replaced,L).
value_which_can_be_replaced2(sequence_extension(L),_) :- maplist(value_which_can_be_replaced,L).
value_which_can_be_replaced2(cartesian_product(A,B),_) :- % typing equations, like t_float = INTEGER*NATURAL1
    simple_value_set(A), simple_value_set(B).
value_which_can_be_replaced2(pow_subset(A),_) :- simple_value_set(A).
%value_which_can_be_replaced2(identifier(_),_). % TO DO: check that this is not the LHS identifier which we replace !!
% identifiers can also be replaced: we check above that the only identifier in the predicate is the equality identifier
% TO DO: enable this but then we need to fix replace_id_by_expr2 to updated the used_ids info ! + we can have scoping issues !?; see test 1358

% TO DO: also allow inlining of prj1/prj2 of simple_value_set:
% not_val(comprehension_set([b(identifier(_zzzz_unary),integer,[generated(first)]),b(identifier(_zzzz_binary),integer,[generated(first)]),b(identifier(_lambda_result_),integer,[lambda_result])],b(equal(b(identifier(_lambda_result_),integer,[lambda_result]),b(identifier(_zzzz_unary),integer,[generated(first)])),pred,[prob_annotation(LAMBDA),lambda_result])))
% not_val(comprehension_set([b(identifier(_zzzz_unary),integer,[generated(second)]),b(identifier(_zzzz_binary),integer,[generated(second)]),b(identifier(_lambda_result_),integer,[lambda_result])],b(equal(b(identifier(_lambda_result_),integer,[lambda_result]),b(identifier(_zzzz_binary),integer,[generated(second)])),pred,[prob_annotation(LAMBDA),lambda_result])))

% TO DO: maybe check if it is an infinite type which cannot be evaluated anyway
simple_value_set(b(E,_,_)) :- simple_value_set2(E).
%simple_value_set2(bool_set). % Warning: we could have a finite construct which gets evaluated multiple times
simple_value_set2(string_set).
simple_value_set2(integer_set(_)).
simple_value_set2(float_set).
simple_value_set2(real_set).
simple_value_set2(cartesian_product(A,B)) :- simple_value_set(A), simple_value_set(B).
simple_value_set2(pow_subset(A)) :- simple_value_set(A).
% POW, records struct(_) set ?

% apply a substiution of ID/Expr on a pred(EXPR,VarList,X) term
% Expr must either be a variable or contain no variables at all
apply_to_pred(ID,Expr,pred(E1,PIds1,X),pred(E2,PIds3,X)) :-
   (select(ID,PIds1,PIds2)
   -> replace_id_by_expr(E1,ID,Expr,E2),  %,print(applied(ID)),nl,translate:print_bexpr(E2),nl
   % TO DO: replace_id_by_expr does not seem to update used_ids info !! check_used_ids_info fails for test 1358 if we allow identifiers inside Expr
      (get_texpr_id(Expr,NewID)
       -> ord_add_element(PIds2,NewID,PIds3)
        ; PIds3 = PIds2)
      %, format('Apply ~w -> ~w : ',[ID,NewID]),translate:print_bexpr(E2),nl
     ; E1=E2,PIds1=PIds3).


% project a predicate : keep only those Predicates that are directly or
%  indirectly relevant for Ids; FullIds are all identifiers used by ProjectedPredicate
project_predicate_on_identifiers(Pred,Ids,ProjectedPredicate,FullIds, RestList) :-
   (debug:debug_mode(on)
     -> print('project: '),print_bexpr(Pred),nl, print('  on : '), print(Ids),nl
     ;  true),
   conjunction_to_list(Pred,Preds),
    l_predicate_identifiers(Preds,[],PredIds), % TO DO: allow LocalVariables to be passed
   % print(predids(PredIds)),nl,
   try_find_and_remove_equalities(PredIds,PredIds2),
   extract_all_predicates(Ids,all,Ids,PredIds2,ProjectedPredicates, RestList,FullIds),
   conjunct_predicates(ProjectedPredicates,ProjectedPredicate),
   (debug:debug_mode(on)
     -> print('*result: '),print_bexpr(ProjectedPredicate),nl,
        print('  on : '),    print(FullIds),nl
     ; true).

%print_preds([]).
%print_preds([pred(P,_IDs,_)|T]) :- translate:print_bexpr(P), nl, print('   '), print_preds(T).

:- use_module(b_global_sets,[b_get_global_constants/1, b_get_enumerated_set_constants/1, b_get_global_sets/1]).
predicate_identifiers(Pred,Ids) :- predicate_identifiers_in_scope(Pred,[],Ids).
predicate_identifiers_in_scope(Pred,LocalVariables,Ids) :-
    get_global_identifiers(IS),
    list_to_ord_set(LocalVariables,LV),
    ord_subtract(IS,LV,Ignore2), % Do not ignore any local variable; it will be used instead of enumerate set element
    find_identifier_uses_if_necessary(Pred,Ignore2,Ids1),  % Ignore enumerated set names
    list_to_ord_set(Ids1,Ids).

get_global_identifiers(IDs) :- get_global_identifiers(IDs,all).
% get global set and constant identifiers which you usually want to exclude for find_identifier_uses
get_global_identifiers(IDs,Option) :-
    (Option=ignore_promoted_constants
       % do not include those constants that have been automatically promoted as enumerated set elements
     -> b_get_enumerated_set_constants(EnumeratedSetCsts)
      ; b_get_global_constants(EnumeratedSetCsts)
     ),
    % b_global_sets:b_get_global_enumerated_sets(GSets), % is there a reason to exclude deferred sets ??; cardinality inference,... are all done before partitioning ?
    b_get_global_sets(GSets),
    append(GSets,EnumeratedSetCsts,GE),
    list_to_ord_set(GE,IDs).

% find_components(ListOf_pred, Restrict, Out:ListOfComponents)
% role of Restrict: all if we do normal partition or List of VariableIDs on which we restrict our attention to (for Existential quantifier construction)
find_components([],_,[]).
find_components([pred(P,PIds,true)|PRest],Restrict,[component(Pred,Ids)|CRest]) :-
    % find all predicates which are using identifiers occuring in PIds
    % (and additionally those which use common identifiers )
    ord_restrict(Restrict,PIds,InterIDs),
    %format('Treating predicate with ids ~w; restr. intersect = ~w~n',[PIds,InterIDs]),
    ( InterIDs =[] ->
        % we simply copy this predicate into a single component; not with the scope of Restricted IDs
        % print(skip(PIds,P)),nl, %
        Pred=P, Ids=PIds, PRest=Rest
    ; otherwise ->
        extract_all_predicates(InterIDs,Restrict,PIds,PRest,Preds,Rest,Ids),
        conjunct_predicates_with_pos_info([P|Preds],Pred)
    ),
    find_components(Rest,Restrict,CRest).
extract_all_predicates([],_,_,Preds,Found,Rest,[]) :- !,  % selecting done at end: keep same order of conjuncts
   select_predicates(Preds,Found,Rest).
extract_all_predicates(Ids,Restrict,OldIds,Preds,Found,Rest,ResultIds) :-
    % search for all predicates that directly use one of the
    % identifiers in "Ids"
    select_all_using_preds(Preds,Ids,FoundIds),
    ord_subtract(FoundIds,OldIds,NewIds),
    ord_restrict(Restrict,NewIds,NewIdsToExtract),
    ord_union(OldIds,FoundIds,OldIds2),
    % now recursively do this which the new identifiers that we have found
    extract_all_predicates(NewIdsToExtract,Restrict,OldIds2,Preds,Found,Rest,RestIds),
    ord_union([Ids,NewIds,RestIds],ResultIds).

% mark all predicates which intersect with ComponentIds and compute new ids to be added to the component
select_all_using_preds([],_,[]).
select_all_using_preds([pred(_P,PIds,Selected)|PRest],ComponentIds,NewFoundIds) :-
    ( (Selected==true ; ord_disjoint(PIds,ComponentIds)) ->
        NewFoundIds1 = []
    ; otherwise ->         % we select the predicate for the component
        ord_subtract(PIds,ComponentIds,NewFoundIds1),
        Selected=true
    ),
    select_all_using_preds(PRest,ComponentIds,NewFoundIds2),
    ord_union(NewFoundIds1,NewFoundIds2,NewFoundIds).

% ord_intersection with special case for all term
ord_restrict(all,Ids,Res) :- !, Res=Ids.
ord_restrict(Restrict,Ids,Res) :- ord_intersection(Ids,Restrict,Res).

select_predicates(Predicates,FoundPreds,OtherPreds) :-
    split_list(is_selected_predicate,Predicates,FoundPredIds,OtherPreds),
    maplist(extract_found_predicate,FoundPredIds,FoundPreds).
is_selected_predicate(pred(_P,_PIds,Selected)) :- ground(Selected).
extract_found_predicate(pred(P,_PIds,_Selected),P).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace (simultaneously) multiple identifiers by expressions
% this could be used to replace the predicate replace_id_by_expr by mapping everything in a
% singleton list. however, that would involve list operations instead of unifications
parse_pred(Codes,TExpr) :- %format('Parsing ~s~n',[Codes]),
    bmachine:b_parse_machine_predicate_from_codes_open(no_quantifier,Codes,[],[],TExpr).
parse_expr(Codes,TExpr) :- bmachine:b_parse_machine_expression_from_codes(Codes,[],TExpr,_Type,true,_Error).
test_result(1,b(equal(b(comprehension_set([b(identifier(__FRESH____),integer,_)],
    b(greater(b(identifier(__FRESH____2),integer,_),b(identifier(x),integer,_)),pred,_)),set(integer),_),
    b(empty_set,set(integer),_)),pred,_)).

:- assert_must_succeed((bsyntaxtree:parse_pred("{x|x:INTEGER & x>y}={}",T1), replace_ids_by_exprs(T1,[],[],T1))).
:- assert_must_succeed((bsyntaxtree:parse_pred("{x|x:INTEGER & x>y}={}",T1), bsyntaxtree:parse_expr("100",T2),
                        replace_ids_by_exprs(T1,[y],[T2],R),bsyntaxtree:parse_pred("{x|x>100}={}",T3),same_texpr(R,T3) )).
:- assert_must_succeed((bsyntaxtree:parse_pred("{x|x:INTEGER & x>y}={}",T1),
                        replace_ids_by_exprs(T1,[y],[b(identifier(x),integer,[])],R),
                        bsyntaxtree:test_result(1,R),bsyntaxtree:parse_pred("{x|x:INTEGER & x>x}={}",T3),\+ same_texpr(R,T3) )).
:- assert_must_succeed((bsyntaxtree:parse_pred("{x,y|x:INTEGER & y:INTEGER & x>v & y>w}={}",T1),
                        replace_ids_by_exprs(T1,[v,w],[b(identifier(w),integer,[]),b(identifier(v),integer,[])],R),
                        bsyntaxtree:parse_pred("{x,y|x:INTEGER & y:INTEGER & x>w & y>v}={}",T3), same_texpr(R,T3) )).
:- assert_must_succeed((gensym:reset_gensym, bsyntaxtree:parse_pred("{x,y|x:INTEGER & y:INTEGER & x>v & y>w}={}",T1),
                        replace_ids_by_exprs(T1,[v,w],[b(identifier(w),integer,[]),b(identifier(x),integer,[])],R),
                        translate:translate_bexpression(R,TR),
                        TR = '{__FRESH____1,y|__FRESH____1 > w & y > x} = {}')). %'{__FRESH____1,y|(__FRESH____1 : INTEGER & y : INTEGER) & (__FRESH____1 > w & y > x)} = {}' )).

replace_ids_by_exprs(TExpr,[TId],[Inserted],Replaced) :- !, % better, more robust and efficient version
    check_ids([TId],[Id]),
    replace_id_by_expr(TExpr,Id,Inserted,Replaced).
replace_ids_by_exprs(TExpr,[],[],Replaced) :- !, Replaced=TExpr. % Nothing to do
replace_ids_by_exprs(TExpr,Ids,Exprs,Replaced) :-
    check_ids(Ids,Ids2), % convert to atomic identifiers
    find_identifier_uses_l(Exprs,[],ExprsUsedIds),
    sort(ExprsUsedIds,SExprsUsedIds),
    findall(rename(I,E),(nth0(N,Ids2,I),nth0(N,Exprs,E)),RenameList),
    %print(replace(RenameList,SExprsUsedIds)),nl,
    replace_ids_by_exprs2(RenameList,SExprsUsedIds,TExpr,Replaced,_). %, nl,print(done),nl.
% a version of replace_ids_by_exprs2 for maplist:
%replace_ids_by_exprs1(RenameList,SExprsUsedIds,TExpr,Replaced) :-
%   replace_ids_by_exprs2(RenameList,SExprsUsedIds,TExpr,Replaced,_). %, print('Rep: '), translate:print_bexpr(Replaced),nl.
check_ids([],[]).
check_ids([H|T],[ID|IT]) :- (atomic(H) -> ID=H ; def_get_texpr_id(H,ID)), check_ids(T,IT).
replace_ids_by_exprs2(RenameList,ExprsUsedIds,TExpr,Replaced,WDC) :-
    remove_bt(TExpr,Expr,NewExpr,TNewExpr),
    ( Expr = identifier(Id), member(rename(Id,Inserted),RenameList) ->
        Replaced = Inserted,
        get_texpr_info(Inserted,Infos),
        (memberchk(contains_wd_condition,Infos)
        -> WDC = true ; WDC = false) % WDC = true means we have added a wd-condition where previously there was none
    ; otherwise ->
        syntaxtransformation_fixed(Expr,QSubs,Subs1,Names,NQSubs,NSubs,NewExpr),
        find_variable_clashes(Names,ExprsUsedIds,RenameNames),    % check for variable caputure
        (RenameNames = []
         -> Subs = Subs1 % no variable capture occured
         ; %format('*** VARIABLE CAPTURE : ~w~n~n',[RenameNames]),
           %maplist(replace_ids_by_exprs1(RenameNames,[]),Subs1,Subs) % replace affected names by fresh ids in sub arguments
           rename_bt_l(Subs1,RenameNames,Subs) % replace affected names by fresh ids in sub arguments (will also change list of quantified variables itself)
         ),
        %print(e(SNameIds,QSubs,Subs)),nl,
        l_replace_ids_by_exprs2(QSubs,RenameList,ExprsUsedIds,NQSubs,WDC1), % QSubs are like RHS of let expression, where Names are not in scope
        remove_hidden_names(Names,RenameList,UpdatedRenameList),
        %print(rem_hidden(Names,UpdatedRenameList)),nl,
        ( UpdatedRenameList = [] -> % all Ids are now hidden for the inner expressions
            NSubs=Subs, WDC=WDC1
        ; otherwise -> %print(recur(Subs)),nl,
            l_replace_ids_by_exprs2(Subs,UpdatedRenameList,ExprsUsedIds,NSubs,WDC2),
            and_wdc(WDC1,WDC2,WDC)
        ),
        TNewExpr = b(E1,T1,Info1),
        rename_update_used_ids_info(RenameList,Info1,Info2),
        %(E1 = exists(P,_) -> print(exists(P,Id,Info1,Info2)),nl ; true),
        add_wd_if_needed(WDC,b(E1,T1,Info2),Replaced)
    ).

% check if we have to rename any quantified variable to avoid variable capture of RHS of renamings
% example, suppose we rename x/y+1  and we enter {y|y>x}  we have to generate {fresh|fresh>y+1} and *not* {y|y>y+1}
find_variable_clashes([],_,[]).
find_variable_clashes([Name|Names],ExprsUsedIds,[rename(ID,FRESHID)|Renaming] ) :-
   def_get_texpr_id(Name,ID),
   ord_member(ID,ExprsUsedIds), % the quantified name is also introduced by the renaming
   !,
   gensym('__FRESH__',FRESHID),
   find_variable_clashes(Names,ExprsUsedIds,Renaming).
find_variable_clashes([_|Names],ExprsUsedIds,Renaming) :-
   find_variable_clashes(Names,ExprsUsedIds,Renaming).


l_replace_ids_by_exprs2([],_,_,[],false).
l_replace_ids_by_exprs2([H|T],UpdatedRenameList,ExprsUsedIds,[IH|IT],WDC) :-
    replace_ids_by_exprs2(UpdatedRenameList,ExprsUsedIds,H,IH,WDC1),
    l_replace_ids_by_exprs2(T,UpdatedRenameList,ExprsUsedIds,IT,WDC2),
    and_wdc(WDC1,WDC2,WDC).

% remove any identifiers that are now "invisible" because they are masked by quantified names
% e.g.,  when we enter #x.(P)  then a renaming of x will be "hidden" inside P
remove_hidden_names([],RenameList,RenameList).
remove_hidden_names([Name|Names],RenameList,NewRenameList) :-
    def_get_texpr_id(Name,ID),
    delete(RenameList,rename(ID,_),RenameList1),
    !, % only one rename should exist
    %print(del(ID,RenameList1)),nl,
    remove_hidden_names(Names,RenameList1,NewRenameList).

find_rhs_ids(rename(Id,TExpr),rename_ids(Id,InsUsedIds)) :- find_identifier_uses(TExpr,[],InsUsedIds).

% apply a rename list to the used_ids,... information fields
% this is more tricky than applying a single identifier, as we first have to deleted all ids
% and remember which ones were deleted, and only then insert the corresponding ids
% e.g., we could have a RenameList = [rename(p,q),rename(q,p)] ; see test 1776 M1_Internal_v3.mch
rename_update_used_ids_info(RenameList,IIn,IOut) :-
    maplist(find_rhs_ids,RenameList,RenameList2),
    maplist(apply_rename_list(RenameList2),IIn,IOut).

apply_rename_list(RenameList,I,NI) :-
    apply_rename_list2(I,RenameList,NI).
apply_rename_list2(used_ids(IDS),RenameList,used_ids(NewIDS)) :- !, apply_rename_list_to_ids(RenameList,IDS,[],NewIDS).
apply_rename_list2(reads(IDS),RenameList,reads(NewIDS)) :- !, apply_rename_list_to_ids(RenameList,IDS,[],NewIDS).
apply_rename_list2(modifies(IDS),RenameList,modifies(NewIDS)) :- !, apply_rename_list_to_ids(RenameList,IDS,[],NewIDS).
apply_rename_list2(Info,_,Info).

% apply a rename list to a sorted list of ids
apply_rename_list_to_ids([],Acc,ToInsert,Res) :- ord_union(Acc,ToInsert,Res).
apply_rename_list_to_ids([rename_ids(Id,NewIds)|T],Acc,ToInsert,Res) :-
   (ord_delete_existing_element(Acc,Id,Acc2) % the Id occurs and is deleted
      ->  ord_union(NewIds,ToInsert,ToInsert2),
          apply_rename_list_to_ids(T,Acc2,ToInsert2,Res)
      ;   apply_rename_list_to_ids(T,Acc,ToInsert,Res)).

ord_delete_existing_element(List,El,ResList) :- % ord_del_element also succeeds if El is not in the list !
    ord_intersection([El],List,[El],ResList).
% -----------------------
% remove an Identifier from used_ids Info field if it exists
remove_used_id_from_info(I,ID_to_remove,NI) :-
   update_used_ids_info(I,ID_to_remove,[],NI).

remove_used_ids_from_info([],I,I).
remove_used_ids_from_info([ID_to_remove|T],I,NI) :- remove_used_id_from_info(I,ID_to_remove,I2),
  remove_used_ids_from_info(T,I2,NI).

% remove a single Identifier from used_ids Info field if it exists and insert sorted list of ids instead
% a simpler version of rename_update_used_ids_info for a single identifier
update_used_ids_info([],_,_,[]).
update_used_ids_info([InfoField|T],ID_to_remove,IDsInserted,[NewInfoField|NT]) :-
    (update_id_from_info_field(ID_to_remove,IDsInserted,InfoField,R)
     -> NewInfoField=R
      ; NewInfoField=InfoField),
    update_used_ids_info(T,ID_to_remove,IDsInserted,NT).

update_id_from_info_field(ID_to_remove,IDsInserted,I,NI) :-
    update_id_from_info_field2(I,ID_to_remove,IDsInserted,NI).
update_id_from_info_field2(used_ids(IDS),ID,IDsInserted,used_ids(NewIDS)) :- update_id(IDS,ID,IDsInserted,NewIDS).
update_id_from_info_field2(reads(IDS),ID,IDsInserted,reads(NewIDS)) :- update_id(IDS,ID,IDsInserted,NewIDS).
update_id_from_info_field2(modifies(IDS),ID,IDsInserted,modifies(NewIDS)) :- update_id(IDS,ID,IDsInserted,NewIDS).

update_id(IDS,ID_to_remove,IDsInserted,NewIDS) :-
    ord_delete_existing_element(IDS,ID_to_remove,IDS2), % the Id occurs and is deleted
    ord_union(IDsInserted,IDS2,NewIDS).
%if ord_del_element fails we do not add IDsInserted: we assume the used_ids info is correct and ID_to_remove does not occur !
% We could use this info to avoid traversing subtree !
% print(projecting_away_unknown_id(ID_to_remove,IDS)),nl,

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert_must_succeed((gensym:reset_gensym, bsyntaxtree:exists_ast(A), replace_id_by_expr(A,y,b(identifier(x),integer,[]),RA),translate:translate_bexpression(RA,TR),
 TR = 'r = {__FRESH____1|__FRESH____1 : 1 .. x & __FRESH____1 mod 2 = 1}' )).
:- assert_must_succeed((bsyntaxtree:parse_pred("{x|x:INTEGER & x>y}={}",T1),
                        replace_id_by_expr(T1,y,b(identifier(x),integer,[]),R),
                        bsyntaxtree:test_result(1,R),bsyntaxtree:parse_pred("{x|x:INTEGER & x>x}={}",T3),
                        \+ same_texpr(R,T3) )).
:- assert_must_succeed((bsyntaxtree:parse_pred("{x|x:INTEGER & x>y}={}",T1), bsyntaxtree:parse_expr("100",T2),
                        replace_id_by_expr(T1,y,T2,R),bsyntaxtree:parse_pred("{x|x>100}={}",T3),same_texpr(R,T3) )).

% replace an identifier Id by an expression Inserted
replace_id_by_expr(TExpr,Id,Inserted,Replaced) :-
   replace_id_by_expr_with_count(TExpr,Id,Inserted,Replaced,_).

replace_id_by_expr_with_count(TExpr,Id,Inserted,Replaced,NrReplacements) :- \+ atomic(Id),!,
   add_internal_error('Id not atomic: ',replace_id_by_expr(TExpr,Id,Inserted,Replaced)),
   Replaced = TExpr, NrReplacements=0.
replace_id_by_expr_with_count(TExpr,Id,Inserted,Replaced,NrReplacements) :-
    %find_all_relevant_quantified_vars(Id,TExpr,QVars),
    find_identifier_uses(Inserted,[],SInsUsedIds), % SInsUsedIds is sorted
    replace_id_by_expr2(Id,Inserted,SInsUsedIds,TExpr,Replaced,_WDC,0,NrReplacements).

replace_id_by_expr2(Id,Inserted,InsUsedIds,TExpr,Replaced,WDC,InR,OutR) :-
    remove_bt(TExpr,Expr,NewExpr,TNewExpr),
    ( Expr = identifier(Id) ->  % TODO: count number of replacements
        Replaced = Inserted,
        OutR is InR+1,
        get_texpr_info(Inserted,Infos),
        (memberchk(contains_wd_condition,Infos)
         -> WDC = true ; WDC = false) % WDC = true means we have added a wd-condition where previously there was none
    ; otherwise ->
        syntaxtransformation_fixed(Expr,QSubs,Subs,Names,NQSubs,NSubs,NewExpr),
        %maplist(replace_id_by_expr2(Id,Inserted),QSubs,NQSubs),
        l_replace_id_by_expr2(QSubs,Id,Inserted,InsUsedIds,NQSubs,WDC1,InR,InR2),
        get_texpr_id(TId,Id),
        ( memberchk(TId,Names) -> % the Id is now hidden for the inner expressions
               NSubs=Subs, WDC=WDC1, OutR = InR2
        ; (InsUsedIds \= [],
           get_texpr_ids(Names,Ns),sort(Ns,SNs),
           ord_intersection(SNs,InsUsedIds,Captured),
           Captured \= [] %, print(inter(SNs,InsUsedIds,Captured)),nl
           )
           % The Names introduced clash with variables used in the Inserted expression
           ->  findall(rename(X,FRESHID),(member(X,Captured),gensym:gensym('__FRESH__',FRESHID)),RenameList),
               %print(rename(RenameList)),nl,
               rename_bt_l(Subs,RenameList,RenSubs),
               l_replace_id_by_expr2(RenSubs,Id,Inserted,InsUsedIds,NSubs,WDC2,InR2,OutR),
               and_wdc(WDC1,WDC2,WDC)
        ; otherwise ->
               l_replace_id_by_expr2(Subs,Id,Inserted,InsUsedIds,NSubs,WDC2,InR2,OutR),
               and_wdc(WDC1,WDC2,WDC)
        ),
        TNewExpr = b(E1,T1,Info1),
        update_used_ids_info(Info1,Id,InsUsedIds,Info2),
        %(E1 = exists(P,_) -> print(exists(P,Id,Info1,Info2)),nl ; true),
        add_wd_if_needed(WDC,b(E1,T1,Info2),Replaced)
    ).

l_replace_id_by_expr2([],_,_,_,[],false,R,R).
l_replace_id_by_expr2([H|T],Id,Inserted,InsUsedIds,[IH|IT],WDC,InR,OutR) :-
   replace_id_by_expr2(Id,Inserted,InsUsedIds,H,IH,WDC1,InR,InR2),
   l_replace_id_by_expr2(T,Id,Inserted,InsUsedIds,IT,WDC2,InR2,OutR),
   and_wdc(WDC1,WDC2,WDC).

and_wdc(true,_,R) :- !,R=true.
and_wdc(_,true,R) :- !, R=true.
and_wdc(_,_,false).

% add contains_wd_condition if a change occured during replacement of id by expression
add_wd_if_needed(true,b(E,T,Infos),Replaced) :-
   nonmember(contains_wd_condition,Infos),
   !,
   Replaced = b(E,T,[contains_wd_condition|Infos]).
add_wd_if_needed(_,T,T).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% syntaxtransformation_fixed/7 is the same as syntaxtransformation/5 with the exception
% that we distinguish between subexpressions that have the newly introduced identifiers (Names)
% in scope (OSubs) and those who don't (OExprs). The only expressions where the latter case is relevant
% are let_expression and let_predicate.
% TODO: This is a quick'n dirty fix for only some cases.
% NO LONGER REQUIRED: let_expression and let_predicate now obey another semantic, not the Z semantics anymore
%syntaxtransformation_fixed(OExpr,OExprs,OSubs,Names,NExprs,NSubs,NExpr) :-
%    is_let_pred_or_expr(OExpr,OTIds,OExprs,OExprPred,LetType),
%    !,
%    Names=OTIds,
%    is_let_pred_or_expr(NExpr,NTIds,NExprs,NExprPred,LetType),
%    same_length(OExprs,NExprs),same_length(OTIds,NTIds),
%    last(OTIds,OExprPred,OSubs),last(NTIds,NExprPred,NSubs),
%    !.
syntaxtransformation_fixed(Expr,[],Subs,Names,[],NSubs,NExpr) :-
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr).

% note : LET substitution is dealt with differently as there we currently allow to use previously introduced IDs in RHS
%is_let_pred_or_expr(let_expression(TIds,Exprs,Expr),TIds,Exprs,Expr,let_expression).
%is_let_pred_or_expr(let_predicate(TIds,Exprs,Expr), TIds,Exprs,Expr,let_predicate).
%is_let_pred_or_expr(lazy_let_pred(ID,SharedExpr,Expr), [ID],[SharedExpr],Expr, lazy_let_pred).
%is_let_pred_or_expr(lazy_let_expr(ID,SharedExpr,Expr), [ID],[SharedExpr],Expr, lazy_let_expr).
%is_let_pred_or_expr(lazy_let_subst(ID,SharedExpr,Expr), [ID],[SharedExpr],Expr, lazy_let_subst).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rename identifier

% r = {x|x : 1..y & x mod 2 = 1}
exists_ast(AST) :- AST =
 b(equal(b(identifier(r),set(integer),[nodeid(none)]),
    b(comprehension_set([b(identifier(x),integer,[nodeid(none)])],
      b(conjunct(b(member(b(identifier(x),integer,[nodeid(none)]),
   b(interval(b(integer(1),integer,[nodeid(none)]),b(identifier(y),integer,[nodeid(none)])),
       set(integer),[nodeid(none)])),pred,[nodeid(none)]),
 b(equal(b(modulo(b(identifier(x),integer,[nodeid(none)]),
 b(integer(2),integer,[nodeid(none)])),integer,[contains_wd_condition,nodeid(none)]),
 b(integer(1),integer,[nodeid(none)])),pred,
 [contains_wd_condition,nodeid(none)])),pred,[contains_wd_condition,nodeid(none)])),
 set(integer),[contains_wd_condition,nodeid(none)])),
 pred,[contains_wd_condition,nodeid(none)]).

:- assert_must_succeed((bsyntaxtree:exists_ast(A), rename_bt(A,[rename(x,xx)],RA), RA==A )).
:- assert_must_succeed((bsyntaxtree:exists_ast(A), rename_bt(A,[rename(r,v)],RA),
                    translate:translate_bexpression(RA,TR), TR='v = {x|x : 1 .. y & x mod 2 = 1}' )).
:- assert_must_succeed((gensym:reset_gensym, bsyntaxtree:exists_ast(A), rename_bt(A,[rename(y,x)],RA), RA \= A,
                        translate:translate_bexpression(RA,TR),
                        TR = 'r = {__FRESH____1|__FRESH____1 : 1 .. x & __FRESH____1 mod 2 = 1}' )).

% a simplified version of replace_ids_by_exprs, which assumes target of renamings are variables
rename_bt(Expr,[],Res) :- !,Res=Expr.
rename_bt(OExpr,Renamings,NExpr) :-
    create_texpr(Old,Type,OInfo,OExpr),
    create_texpr(New,Type,NInfo,NExpr),
    rename_in_infos(OInfo,Renamings,NInfo),
    rename_bt2(Old,Renamings,New).
rename_bt2(identifier(Old),Renamings,identifier(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(lazy_lookup_expr(Old),Renamings,lazy_lookup_expr(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(lazy_lookup_pred(Old),Renamings,lazy_lookup_pred(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(OExpr,Renamings,NExpr) :-
    syntaxtransformation(OExpr,Subs,TNames,NSubs,NExpr),
    get_texpr_exprs(TNames,Names),
    remove_renamings(Names,Renamings,NRenamings),
    rename_bt_l(Subs,NRenamings,NSubs).
rename_bt_l([],_,[]).
rename_bt_l([Old|ORest],Renamings,[New|NRest]) :-
    rename_bt(Old,Renamings,New),
    rename_bt_l(ORest,Renamings,NRest).

remove_renamings([],Renamings,Renamings).
remove_renamings([identifier(Name)|Rest],Old,New) :-
    ( select(rename(Name,_),Old,Inter1) -> true  % Name no longer visible to renaming
          ; Old = Inter1),
    (member(rename(_OldName,Name),Inter1) ->
        gensym('__FRESH__',FRESHID),
        %print(variable_capture_in_rename(Name,from(OldName),FRESHID)),nl,
        Inter2 = [rename(Name,FRESHID)|Inter1]
      ; Inter2 = Inter1),
    remove_renamings(Rest,Inter2,New).

rename_in_infos(Old,Renamings,New) :-
    ( has_info_to_rename(Old) ->
        maplist(rename_in_infos2(Renamings),Old,New)
    ; otherwise ->
        Old = New).
rename_in_infos2(Renamings,OInfo,NInfo) :-
    ( infos_to_rename(OInfo,OIds,SortedNIds,NInfo) ->
        rename_ids(OIds,Renamings,NIds),
        sort(NIds,SortedNIds)
    ; otherwise ->
        OInfo = NInfo).

rename_ids([],_,[]).
rename_ids([OId|Orest],Renamings,[NId|Nrest]) :-
    rename_id(OId,Renamings,NId),
    rename_ids(Orest,Renamings,Nrest).
rename_id(Old,Renamings,New) :-
    ( memberchk(rename(Old,New),Renamings) -> true  % we could use ord_member if we sort !
    ; otherwise -> Old=New).

has_info_to_rename(Infos) :-
    member(I,Infos),infos_to_rename(I,_,_,_),!.

infos_to_rename(modifies(O),O,N,modifies(N)).
infos_to_rename(reads(O),O,N,reads(N)).
infos_to_rename(non_det_modifies(O),O,N,non_det_modifies(N)).
infos_to_rename(modifies_locals(O),O,N,modifies_locals(N)).
infos_to_rename(reads_locals(O),O,N,reads_locals(N)).
infos_to_rename(used_ids(O),O,N,used_ids(N)).
%infos_to_rename(lambda_result(O),[O],[N],lambda_result(N)). % whould we no longer assume that we have lambda result, as predicate has possibly changed!?


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove type information for transformations

remove_bt(b(Expr,Type,Infos),Expr,NExpr,b(NExpr,Type,Infos)).

remove_bt_and_used_ids(b(OldExpr,T,Infos),OldExpr,NewExpr,b(NewExpr,T,NewInfos)) :-
    delete(Infos,used_ids(_),NewInfos). % invalidate used_ids info

%remove_bt_l([],[],[],[]).
%remove_bt_l([OT|OTRest],[O|ORest],[N|NRest],[NT|NTRest]) :-
%   remove_bt(OT,O,N,NT),
%    remove_bt_l(OTRest,ORest,NRest,NTRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% traversations
% Takes a predicate or expression or substitution and extracts:
%  the expression itself Expr, its type Type, the syntaxnode information Infos
%  + the subexpressions as a list Subs and the local identifiers declared
syntaxtraversion(b(Expr,Type,Infos),Expr,Type,Infos,Subs,Names) :- !,
    safe_syntaxelement(Expr,Subs,Names,_,_).
syntaxtraversion(IExpr,Expr,Type,Infos,Subs,Names) :-
    add_internal_error('Not properly wrapped', syntaxtraversion(IExpr,Expr,Type,Infos,Subs,Names)),
    fail.


map_over_full_bexpr_no_fail(P,BExpr) :-
   syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
   call(P,Expr), % the predicate should not fail
   (Subs=[] -> true ; maplist(map_over_full_bexpr_no_fail(P),Subs)).

map_over_bexpr(P,BExpr) :-
   syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
   (call(P,Expr) % should probably fail so that by backtrack we recurse
    ;
    member(Sub,Subs), map_over_bexpr(P,Sub)
   ).
% same as above but gets typed expressions:
map_over_typed_bexpr(P,BExpr) :-
   syntaxtraversion(BExpr,_Expr,_,_,Subs,_TNames),
   (call(P,BExpr)
    ;
    member(Sub,Subs), map_over_typed_bexpr(P,Sub)
   ).
% same as above but returns value:
map_over_typed_bexpr(P,BExpr,Result) :-
   syntaxtraversion(BExpr,_Expr,_,_,Subs,_TNames),
   (call(P,BExpr,Result)
    ;
    member(Sub,Subs), map_over_typed_bexpr(P,Sub,Result)
   ).
% this one gets TNames (locally introduced variables as parameter)
map_over_typed_bexpr_with_names(P,BExpr) :-
   syntaxtraversion(BExpr,_Expr,_,_,Subs,TNames),
   (call(P,BExpr,TNames)
    ;
    member(Sub,Subs), map_over_typed_bexpr_with_names(P,Sub)
   ).

% same as map_over_expr but provides an accumulator passed top-down to children; needs to be used by failure driven loop

map_over_bexpr_top_down_acc(P,BExpr,TDAcc) :-
   syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
   (call(P,Expr,TDAcc,NewAcc)
     -> member(Sub,Subs), map_over_bexpr_top_down_acc(P,Sub,NewAcc)
     ;  member(Sub,Subs), map_over_bexpr_top_down_acc(P,Sub,TDAcc)
   ).
% now a version which gets the typed predicate as argument

map_over_type_bexpr_top_down_acc(P,BExpr,TDAcc) :-
   syntaxtraversion(BExpr,_Expr,_,_,Subs,_TNames),
   (call(P,BExpr,TDAcc,NewAcc)
     -> member(Sub,Subs), map_over_type_bexpr_top_down_acc(P,Sub,NewAcc)
     ;  member(Sub,Subs), map_over_type_bexpr_top_down_acc(P,Sub,TDAcc)
   ).


reduce_over_bexpr(P,BExpr,InitialValue,ResultValue) :-
   syntaxtraversion(BExpr,Expr,_,_,Subs,_TNames),
   call(P,Expr,InitialValue,I1), % print(reduce(P,Expr,InitialValue,I1)),nl,
   scanlist(reduce_over_bexpr(P),Subs,I1,ResultValue).

% apply a predicate over a syntax tree (bottom-up)

transform_bexpr(Pred,b(Expr,Type,Info),NewBExpr) :-
   syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
   l_transform_bexpr(Subs,Pred,NSubs),
   (call(Pred,b(NewExpr1,Type,Info),NewBExpr) -> true ; NewBExpr = b(NewExpr1,Type,Info)).

l_transform_bexpr([],_,[]).
l_transform_bexpr([SubH|T],Pred,[TSubH|TT]) :-
   transform_bexpr(Pred,SubH,TSubH),
   l_transform_bexpr(T,Pred,TT).

% apply a predicate over a syntax tree (bottom-up), and provide scoping info about local ids

transform_bexpr_with_scoping(Pred,b(Expr,Type,Info),NewBExpr) :-
  transform_bexpr_with_scoping2(Pred,b(Expr,Type,Info),NewBExpr,[]).
transform_bexpr_with_scoping2(Pred,b(Expr,Type,Info),NewBExpr,LocalIds) :-
   syntaxtransformation(Expr,Subs,Names,NSubs,NewExpr1),
   get_texpr_ids(Names,QuantifiedNewIds), list_to_ord_set(QuantifiedNewIds,SQuantifiedNewIds),
   ord_union(LocalIds,SQuantifiedNewIds,NewLocalIds),
   l_transform_bexpr_with_scoping(Subs,Pred,NSubs,NewLocalIds),
   (call(Pred,b(NewExpr1,Type,Info),NewBExpr,LocalIds) -> true ; NewBExpr = b(NewExpr1,Type,Info)).

l_transform_bexpr_with_scoping([],_,[],_).
l_transform_bexpr_with_scoping([SubH|T],Pred,[TSubH|TT],LocalIds) :-
   transform_bexpr_with_scoping2(Pred,SubH,TSubH,LocalIds),
   l_transform_bexpr_with_scoping(T,Pred,TT,LocalIds).


% apply a predicate over a syntax tree (bottom-up) with Accumulator result
% Accumulator is constructed bottom up; Pred receives *all* accumulators of sub expressions

transform_bexpr_with_bup_accs(Pred,b(Expr,Type,Info),NewBExpr,EmptyAcc,Acc) :-
   syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
   l_transform_bexpr_with_bup_accs(Subs,Pred,NSubs,EmptyAcc,SubAccs),
   (call(Pred,b(NewExpr1,Type,Info),NewBExpr,SubAccs,Acc) -> true
    ; NewBExpr = b(NewExpr1,Type,Info), Acc = EmptyAcc).

l_transform_bexpr_with_bup_accs([],_,[],_,[]).
l_transform_bexpr_with_bup_accs([SubH|T],Pred,[TSubH|TT],EmptyAcc,[Acc1|RestAcc]) :-
   transform_bexpr_with_bup_accs(Pred,SubH,TSubH,EmptyAcc,Acc1),
   l_transform_bexpr_with_bup_accs(T,Pred,TT,EmptyAcc,RestAcc).

% apply a predicate over a syntax tree (bottom-up) with Accumulator result
% a single Accumulator is passed along

transform_bexpr_with_acc(_Pred,E,NewBExpr,InAcc,Acc) :- var(E),!,
   NewBExpr=E, Acc=InAcc.
transform_bexpr_with_acc(Pred,b(Expr,Type,Info),NewBExpr,InAcc,Acc) :-
   syntaxtransformation(Expr,Subs,_Names,NSubs,NewExpr1),
   l_transform_bexpr_with_acc(Subs,Pred,NSubs,InAcc,SubAcc),
   (call(Pred,b(NewExpr1,Type,Info),NewBExpr,SubAcc,Acc) -> true
    ; NewBExpr = b(NewExpr1,Type,Info), Acc = SubAcc).

l_transform_bexpr_with_acc([],_,[],Acc,Acc).
l_transform_bexpr_with_acc([SubH|T],Pred,[TSubH|TT],InAcc,ResAcc) :-
   transform_bexpr_with_acc(Pred,SubH,TSubH,InAcc,Acc1),
   l_transform_bexpr_with_acc(T,Pred,TT,Acc1,ResAcc).

min_max_integer_value_used(BExpr,Min,Max) :-
   min_max_integer_value_used(BExpr,none,none,Min,Max).
min_max_integer_value_used(BExpr,IMin,IMax,Min,Max) :-
   reduce_over_bexpr(min_max_aux,BExpr,minmax(IMin,IMax),minmax(Min,Max)).

min_max_aux(sequence_extension(L),minmax(Min,Max),minmax(NMin,NMax)) :- !,
  length(L,Len), % we use implicitly numbers from 1..Len
  ((number(Min),1>Min) -> NMin=Min ; NMin=1),
  ((number(Max),Len<Max) -> NMax=Max ; NMax=Len).
min_max_aux(integer(N),minmax(Min,Max),minmax(NMin,NMax)) :- !,
  ((number(Min),N>Min) -> NMin=Min ; NMin=N),
  ((number(Max),N<Max) -> NMax=Max ; NMax=N).
min_max_aux(_,V,V).

% check if a B expression uses something like NAT,NAT1,INT, MAXINT or MININT.
uses_implementable_integers(BExpr) :-
   map_over_bexpr(uses_implementable_integers_aux,BExpr).

uses_implementable_integers_aux(maxint).
uses_implementable_integers_aux(minint).
uses_implementable_integers_aux(integer_set(X)) :-
   (X='NAT1' ; X='NAT' ; X='INT').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some checks
check_if_typed_predicate(b(Pred,X,_)) :- ground(X), X=pred,  % at runtime there can be value(X) with variables inside !
    syntaxelement(Pred,_,_,_,_,pred).
check_if_typed_expression(b(Expr,Type,_)) :-
    syntaxelement(Expr,_,_,_,_,expr),
    Type \== pred, Type \== subst, ground(Type).
check_if_typed_substitution(b(Subst,X,_)) :- ground(X), X=subst,
    syntaxelement(Subst,_,_,_,_,subst).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% transformations

syntaxtransformation(Expr,Subs,Names,NSubs,NExpr) :-
    functor(Expr,F,Arity),
    functor(NExpr,F,Arity),
    safe_syntaxelement(Expr,Subs,Names,Lists,Constant),
    all_same_length(Lists,NLists),
    syntaxelement(NExpr,NSubs,_,NLists,Constant,_).

safe_syntaxelement(Expr,Subs,Names,Lists,Constant) :-
    ( syntaxelement(Expr,SubsX,NamesX,Lists,ConstantX,_) ->
        Subs=SubsX, Names=NamesX, Constant=ConstantX
    ; otherwise ->
        functor(Expr,F,Arity),
        add_error_fail(bsyntaxtree,'Uncovered syntax element: ', F/Arity)
        ).

% syntaxelement(Expr,SubExprs,Identifiers,Lists,Constant,Type):
% Expr: the expression itself
% SubExprs: a list of sub-expressions
% Identifiers: a list of identifiers that are newly introduced (e.g. by a quantifier)
% Lists: A list of lists in the expression, to prevent infinite loops when having variable parts
% Constant: A part of the expression that is not a sub-expression (e.g. the number in integer(...))
% Type: Fundamental type of the element (predicate, expression, etc)

% predicates
syntaxelement(truth,                 [],   [], [], [], pred).
syntaxelement(falsity,               [],   [], [], [], pred).
syntaxelement(unknown_truth_value(Msg),[],   [], [], [Msg], pred). % artificial, e.g., created by well_def_analyser
syntaxelement(conjunct(A,B),         [A,B],[], [], [], pred).
syntaxelement(negation(A),           [A],  [], [], [], pred).
syntaxelement(disjunct(A,B),         [A,B],[], [], [], pred).
syntaxelement(implication(A,B),      [A,B],[], [], [], pred).
syntaxelement(equivalence(A,B),      [A,B],[], [], [], pred).
syntaxelement(equal(A,B),            [A,B],[], [], [], pred).
syntaxelement(not_equal(A,B),        [A,B],[], [], [], pred).
syntaxelement(member(A,B),           [A,B],[], [], [], pred).
syntaxelement(not_member(A,B),       [A,B],[], [], [], pred).
syntaxelement(subset(A,B),           [A,B],[], [], [], pred).
syntaxelement(subset_strict(A,B),    [A,B],[], [], [], pred).
syntaxelement(not_subset(A,B),       [A,B],[], [], [], pred).
syntaxelement(not_subset_strict(A,B),[A,B],[], [], [], pred).
syntaxelement(less_equal(A,B),       [A,B],[], [], [], pred).
syntaxelement(less(A,B),             [A,B],[], [], [], pred).
syntaxelement(less_equal_real(A,B),  [A,B],[], [], [], pred).
syntaxelement(less_real(A,B),        [A,B],[], [], [], pred).
syntaxelement(greater_equal(A,B),    [A,B],[], [], [], pred).
syntaxelement(greater(A,B),          [A,B],[], [], [], pred).
syntaxelement(forall(Ids,D,P),       [D,P|Ids],Ids,[Ids], [], pred).
syntaxelement(exists(Ids,P),         [P|Ids],  Ids,[Ids], [], pred).
syntaxelement(finite(A),             [A],  [], [], [], pred/only_typecheck).
syntaxelement(partition(S,Es),       [S|Es],[],[Es],[],pred).
syntaxelement(kodkod(PId,Ids),       Ids,[],[Ids],PId, pred).
syntaxelement(external_pred_call(F,Args),Args,[],[Args],F,pred).

% expressions
syntaxelement(value(V),                [],   [], [], V,  expr).
syntaxelement(operation_call_in_expr(Id,As), As, [], [As], [Id], expr). % Do not treat Id as a sub-expression for find_identifier_uses, ...
%syntaxelement(operation_call_in_expr(Id,As), [Id|As], [], [As], [], expr). % was like this, but changed to avoid op(.) ids in find_identifier_uses
syntaxelement(boolean_true,            [],   [], [], [], expr).
syntaxelement(boolean_false,           [],   [], [], [], expr).
syntaxelement(max_int,                 [],   [], [], [], expr).
syntaxelement(min_int,                 [],   [], [], [], expr).
syntaxelement(empty_set,               [],   [], [], [], expr).
syntaxelement(bool_set,                [],   [], [], [], expr).
syntaxelement(float_set,                [],   [], [], [], expr).
syntaxelement(real(I),                 [],   [], [], I,  expr).
syntaxelement(real_set,                [],   [], [], [], expr).
syntaxelement(string_set,              [],   [], [], [], expr).
syntaxelement(convert_bool(A),         [A],  [], [], [], expr).
syntaxelement(convert_real(A),         [A],  [], [], [], expr).
syntaxelement(convert_int_floor(A),    [A],  [], [], [], expr).
syntaxelement(convert_int_ceiling(A),  [A],  [], [], [], expr).
syntaxelement(add(A,B),                [A,B],[], [], [], expr).
syntaxelement(add_real(A,B),           [A,B],[], [], [], expr).
syntaxelement(minus(A,B),              [A,B],[], [], [], expr).
syntaxelement(minus_real(A,B),         [A,B],[], [], [], expr).
syntaxelement(minus_or_set_subtract(A,B),[A,B],[], [], [], expr/only_typecheck).
syntaxelement(unary_minus(A),          [A],  [], [], [], expr).
syntaxelement(unary_minus_real(A),     [A],  [], [], [], expr).
syntaxelement(multiplication(A,B),     [A,B],[], [], [], expr).
syntaxelement(multiplication_real(A,B),[A,B],[], [], [], expr).
syntaxelement(mult_or_cart(A,B),       [A,B],[], [], [], expr/only_typecheck).
syntaxelement(cartesian_product(A,B),  [A,B],[], [], [], expr).
syntaxelement(div(A,B),                [A,B],[], [], [], expr).
syntaxelement(div_real(A,B),           [A,B],[], [], [], expr).
syntaxelement(floored_div(A,B),        [A,B],[], [], [], expr).
syntaxelement(modulo(A,B),             [A,B],[], [], [], expr).
syntaxelement(power_of(A,B),           [A,B],[], [], [], expr).
syntaxelement(power_of_real(A,B),      [A,B],[], [], [], expr).
syntaxelement(successor,               [],   [], [], [], expr).
syntaxelement(predecessor,             [],   [], [], [], expr).
syntaxelement(max(A),                  [A],  [], [], [], expr).
syntaxelement(min(A),                  [A],  [], [], [], expr).
syntaxelement(card(A),                 [A],  [], [], [], expr).
syntaxelement(couple(A,B),             [A,B],[], [], [], expr).
syntaxelement(pow_subset(A),           [A],  [], [], [], expr).
syntaxelement(pow1_subset(A),          [A],  [], [], [], expr).
syntaxelement(fin_subset(A),           [A],  [], [], [], expr).
syntaxelement(fin1_subset(A),          [A],  [], [], [], expr).
syntaxelement(interval(A,B),           [A,B],[], [], [], expr).
syntaxelement(union(A,B),              [A,B],[], [], [], expr).
syntaxelement(intersection(A,B),       [A,B],[], [], [], expr).
syntaxelement(set_subtraction(A,B),    [A,B],[], [], [], expr).
syntaxelement(general_union(A),        [A],  [], [], [], expr).
syntaxelement(general_intersection(A), [A] , [], [], [], expr).
syntaxelement(relations(A,B),          [A,B],[], [], [], expr).
syntaxelement(identity(A),             [A],  [], [], [], expr).
syntaxelement(event_b_identity,        [],  [], [], [], expr). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(reverse(A),              [A],  [], [], [], expr).
syntaxelement(first_projection(A,B),   [A,B],[], [], [], expr/only_typecheck).
syntaxelement(first_of_pair(A),        [A],  [], [], [], expr).
syntaxelement(event_b_first_projection(A),[A],  [], [], [], expr/only_typecheck).
syntaxelement(event_b_first_projection_v2,[],  [], [], [], expr/only_typecheck). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(second_projection(A,B),  [A,B],[], [], [], expr/only_typecheck).
syntaxelement(event_b_second_projection_v2,[],  [], [], [], expr/only_typecheck). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(second_of_pair(A),       [A],  [], [], [], expr).
syntaxelement(event_b_second_projection(A),[A],  [], [], [], expr/only_typecheck).
syntaxelement(composition(A,B),        [A,B],[], [], [], expr).
syntaxelement(ring(A,B),               [A,B],[], [], [], expr/only_typecheck).
syntaxelement(direct_product(A,B),     [A,B],[], [], [], expr).
syntaxelement(parallel_product(A,B),   [A,B],[], [], [], expr).
syntaxelement(trans_function(A),       [A],  [], [], [], expr).
syntaxelement(trans_relation(A),       [A],  [], [], [], expr).
syntaxelement(iteration(A,B),          [A,B],[], [], [], expr).
syntaxelement(reflexive_closure(A),    [A],  [], [], [], expr).
syntaxelement(closure(A),              [A],  [], [], [], expr).
syntaxelement(domain(A),               [A],  [], [], [], expr).
syntaxelement(range(A),                [A],  [], [], [], expr).
syntaxelement(image(A,B),              [A,B],[], [], [], expr).
syntaxelement(domain_restriction(A,B), [A,B],[], [], [], expr).
syntaxelement(domain_subtraction(A,B), [A,B],[], [], [], expr).
syntaxelement(range_restriction(A,B),  [A,B],[], [], [], expr).
syntaxelement(range_subtraction(A,B), [A,B],[], [], [], expr).
syntaxelement(overwrite(A,B),          [A,B],[], [], [], expr).
syntaxelement(partial_function(A,B),   [A,B],[], [], [], expr).
syntaxelement(total_function(A,B),     [A,B],[], [], [], expr).
syntaxelement(partial_injection(A,B),  [A,B],[], [], [], expr).
syntaxelement(total_injection(A,B),    [A,B],[], [], [], expr).
syntaxelement(partial_surjection(A,B), [A,B],[], [], [], expr).
syntaxelement(total_surjection(A,B),   [A,B],[], [], [], expr).
syntaxelement(total_bijection(A,B),    [A,B],[], [], [], expr).
syntaxelement(partial_bijection(A,B),  [A,B],[], [], [], expr).
syntaxelement(total_relation(A,B),     [A,B],[], [], [], expr).
syntaxelement(surjection_relation(A,B),[A,B],[], [], [], expr).
syntaxelement(total_surjection_relation(A,B),[A,B],[], [], [], expr).
syntaxelement(seq(A),                  [A],  [], [], [], expr).
syntaxelement(seq1(A),                 [A],  [], [], [], expr).
syntaxelement(iseq(A),                 [A],  [], [], [], expr).
syntaxelement(iseq1(A),                [A],  [], [], [], expr).
syntaxelement(perm(A),                 [A],  [], [], [], expr).
syntaxelement(empty_sequence,          [],   [], [], [], expr).
syntaxelement(size(A),                 [A],  [], [], [], expr).
syntaxelement(first(A),                [A],  [], [], [], expr).
syntaxelement(last(A),                 [A],  [], [], [], expr).
syntaxelement(front(A),                [A],  [], [], [], expr).
syntaxelement(tail(A),                 [A],  [], [], [], expr).
syntaxelement(rev(A),                  [A],  [], [], [], expr).
syntaxelement(concat(A,B),             [A,B],[], [], [], expr).
syntaxelement(insert_front(A,B),       [A,B],[], [], [], expr).
syntaxelement(insert_tail(A,B),        [A,B],[], [], [], expr).
syntaxelement(restrict_front(A,B),     [A,B],[], [], [], expr).
syntaxelement(restrict_tail(A,B),      [A,B],[], [], [], expr).
syntaxelement(general_concat(A),       [A],  [], [], [], expr).
syntaxelement(function(A,B),           [A,B],[], [], [], expr).
syntaxelement(external_function_call(F,Args),Args,[],[Args],F,expr).
syntaxelement(identifier(I),           [],   [], [], I,  expr).
syntaxelement(lazy_lookup_expr(I),           [],   [], [], I,  expr).
syntaxelement(lazy_lookup_pred(I),           [],   [], [], I,  pred).
syntaxelement(integer(I),              [],   [], [], I,  expr).
syntaxelement(integer_set(T),          [],   [], [], T,  expr).
syntaxelement(string(S),               [],   [], [], S,  expr).
syntaxelement(set_extension(L),        L,    [], [L], [], expr).
syntaxelement(sequence_extension(L),   L,    [], [L], [], expr).
syntaxelement(comprehension_set(Ids,P),[P|Ids],  Ids,[Ids], [], expr).
syntaxelement(event_b_comprehension_set(Ids,E,P),[E,P|Ids],  Ids,[Ids], [], expr/only_typecheck).
syntaxelement(lambda(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(general_sum(Ids,P,E),             [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(general_product(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(quantified_union(Ids,P,E),        [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(quantified_intersection(Ids,P,E), [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(struct(Rec), [Rec], [], [], [], expr).
syntaxelement(rec(Fields), FContent, [], [FContent], FNames, expr) :- syntaxfields(Fields,FContent, FNames).
syntaxelement(record_field(R,I), [R], [], [], I, expr).
syntaxelement(assertion_expression(Cond,ErrMsg,Expr), [Cond,Expr], [], [], ErrMsg, expr).
syntaxelement(typeset, [], [], [], [], expr/only_typecheck).

syntaxelement(tree(A),                 [A],  [], [], [], expr).
syntaxelement(btree(A),                [A],  [], [], [], expr).
syntaxelement(const(A,B),              [A,B],[], [], [], expr).
syntaxelement(top(A),                  [A],  [], [], [], expr).
syntaxelement(sons(A),                 [A],  [], [], [], expr).
syntaxelement(prefix(A),               [A],  [], [], [], expr).
syntaxelement(postfix(A),              [A],  [], [], [], expr).
syntaxelement(sizet(A),                [A],  [], [], [], expr).
syntaxelement(mirror(A),               [A],  [], [], [], expr).
syntaxelement(rank(A,B),               [A,B],[], [], [], expr).
syntaxelement(father(A,B),             [A,B],[], [], [], expr).
syntaxelement(son(A,B,C),              [A,B,C],[], [], [], expr).
syntaxelement(subtree(A,B),            [A,B],[], [], [], expr).
syntaxelement(arity(A,B),              [A,B],[], [], [], expr).
syntaxelement(bin(A),                  [A],[], [], [], expr).
syntaxelement(bin(A,B,C),              [A,B,C],[], [], [], expr).
syntaxelement(infix(A),                [A],  [], [], [], expr).
syntaxelement(left(A),                 [A],  [], [], [], expr).
syntaxelement(right(A),                [A],  [], [], [], expr).

% substitutions
syntaxelement(skip,                    [],   [], [], [], subst).
syntaxelement(precondition(A,B),       [A,B],[], [], [], subst).
syntaxelement(assertion(A,B),          [A,B],[], [], [], subst).
syntaxelement(if_elsif(A,B),           [A,B],[], [], [], subst/elsif).
syntaxelement(while(A,B,C,D),          [A,B,C,D],[], [], [], subst).
% used only internally in the interpreter, contains last value of variant:
syntaxelement(while1(A,B,C,D,E),          [A,B,C,D],[], [], E, subst).
syntaxelement(select_when(A,B),        [A,B],[], [], [], subst/when).
syntaxelement(block(S),[S],[],[],[], subst/only_typecheck).
syntaxelement(assign(Lhs,Rhs),Exprs,[],[Lhs,Rhs], [], subst) :- append(Lhs,Rhs,Exprs).
syntaxelement(assign_single_id(Id,Rhs),[Id,Rhs],[],[], [], subst).
syntaxelement(any(Ids,P,S),[P,S|Ids],Ids,[Ids], [], subst).
syntaxelement(var(Ids,S),  [S|Ids],  Ids,[Ids], [], subst).
syntaxelement(if(Ifs),   Ifs,  [], [Ifs], [], subst).
syntaxelement(parallel(Ss), Ss, [], [Ss], [], subst).
syntaxelement(sequence(Ss), Ss, [], [Ss], [], subst).
syntaxelement(becomes_element_of(Ids,E), [E|Ids], [], [Ids], [], subst).
syntaxelement(becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst).  % Ids are new value, Ids$0 is old value
syntaxelement(evb2_becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst/only_typecheck).
syntaxelement(let(Ids,P,S), [P,S|Ids], Ids, [Ids], [], subst).
syntaxelement(operation_call(Id,Rs,As), [Id|Exprs], [], [Rs,As], [], subst) :- append(Rs,As,Exprs).
syntaxelement(case(E,Eithers,Else), [E,Else|Eithers], [], [Eithers], [], subst).
syntaxelement(case_or(Es,S), [S|Es], [], [Es], [], subst/caseor).
syntaxelement(choice(Ss), Ss, [], [Ss], [], subst).
syntaxelement(select(Whens), Whens, [], [Whens], [], subst).
syntaxelement(select(Whens,Else), [Else|Whens], [], [Whens], [], subst).
syntaxelement(operation(I,Rs,As,B), [I,B|Ids], Ids, [Rs,As], [], subst) :- append(Rs,As,Ids).
syntaxelement(external_subst_call(F,Args),Args,[],[Args],F,subst).

% elements of a VALUES clause
syntaxelement(values_entry(I,E),[I,E],[],[],[],values_entry).

% syntax for Event-B events
syntaxelement(rlevent(I,Sec,St,Ps,G,Ts,As,VWs,PWs,Ums,Rs), Subs, [], [Ps,Ts,As,VWs,PWs,Ums,Rs], [I,Sec], subst) :-
    append([[St],Ps,[G],Ts,As,VWs,PWs,Ums,Rs],Subs).
syntaxelement(witness(I,P), [I,P], [], [], [], witness).

% extended syntax for Z
syntaxelement(let_predicate(Ids,As,Pred),  Exprs, Ids, [Ids,As], [], pred) :- append([Ids,As,[Pred]],Exprs).
syntaxelement(let_expression(Ids,As,Expr), Exprs, Ids, [Ids,As], [], expr) :- append([Ids,As,[Expr]],Exprs).
syntaxelement(let_expression_global(Ids,As,Expr), Exprs, Ids, [Ids,As], [], expr) :-  % version used by b_compiler
   append([Ids,As,[Expr]],Exprs).
syntaxelement(lazy_let_expr(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], expr) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(lazy_let_pred(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], pred) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(lazy_let_subst(TID,A,Expr), Exprs, [TID], [[TID],[A]], [], subst) :-
   append([[TID],[A],[Expr]],Exprs).
syntaxelement(if_then_else(If,Then,Else),[If,Then,Else], [], [], [], expr).
syntaxelement(compaction(A), [A],  [], [], [], expr).
syntaxelement(mu(A),         [A],  [], [], [], expr).
syntaxelement(bag_items(A),  [A],  [], [], [], expr).

syntaxelement(freetype_set(Id), [], [], [], Id, expr).
syntaxelement(freetype_case(Type,Case,Expr), [Expr], [], [], [Type,Case], pred).
syntaxelement(freetype_constructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).
syntaxelement(freetype_destructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).

syntaxelement(ordinary,    [], [], [], [], status).
syntaxelement(anticipated(Variant), [Variant], [], [], [], status).
syntaxelement(convergent(Variant),  [Variant], [], [], [], status).

% Just one ID expected
syntaxelement(recursive_let(Id,C),[Id,C],[Id],[],[], expr). % Note: Id is not really introduced !


% fields of records
%syntaxfields(Fields,C,_) :- var(Fields),var(C),var(N),!, add_internal_error('Illegal call: ',syntaxfields(Fields,C,N)),fail.
syntaxfields([],[],[]).
syntaxfields([field(N,C)|Rest],[C|CRest],[N|NRest]) :- syntaxfields(Rest,CRest,NRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% helper for declaration of quantified identifiers

% has_declared_identifier/2 returns a list of identifiers which are declared in
% this AST node. The main difference to the Names variable when doing a
% syntaxtraversion/6 or similiar is that the identifiers are described by
% predicates.
has_declared_identifier(TExpr,Ids) :-
    get_texpr_expr(TExpr,Expr),
    ( default_declaration(Expr,_,Ids,_)
    ; non_default_declaration(Expr,Ids)).

add_declaration_for_identifier(b(Expr,Type,Infos),Decl,b(NExpr,Type,NewInfos)) :-
    %delete(Infos,used_ids(_),NewInfos), % we cannot
    ( default_declaration(Expr,Predicate,Ids,Constant) ->
        same_functor(Expr,NExpr),
        conjunct_predicates([Decl,Predicate],NPredicate),
        default_declaration(NExpr,NPredicate,Ids,Constant)
    ; non_default_declaration(Expr,Ids) ->
        add_non_default_declaration(Expr,Decl,NExpr)
    ),
    add_used_ids(Infos,Ids,Decl,NewInfos).

% add used ids of a predicate within quantification of Ids to current used_ids info; if it is there
add_used_ids(Infos,Ids,Pred,NewInfos) :- update_used_ids(Infos,OldUsed,NewInfos,NewUsed),
   !, % a field needs updating
   find_identifier_uses(Pred,[],NewIds), get_texpr_ids(Ids,UnsortedIds),sort(UnsortedIds,SIds),
   ord_subtract(NewIds,SIds,NewIds2),
   ord_union(NewIds2,OldUsed,NewUsed).
add_used_ids(I,_,_,I).

update_used_ids(Infos,OldUsed,NewInfos,NewUsed) :- select(OldInfo,Infos,I1),
    used_ids_like_info(OldInfo,F,OldUsed),!,
    used_ids_like_info(NewInfo,F,NewUsed),NewInfos = [NewInfo|I1].

% info fields which contain used_ids information
used_ids_like_info(used_ids(UsedIds),used_ids,UsedIds).
used_ids_like_info(reads(UsedIds),reads,UsedIds).

:- use_module(probsrc(btypechecker), [prime_identifiers/2]).

default_declaration(forall(Ids,D,P),D,Ids,P).
default_declaration(exists(Ids,P),P,Ids,[]).
default_declaration(comprehension_set(Ids,P),P,Ids,[]).
default_declaration(lambda(Ids,P,E),P,Ids,E).
default_declaration(general_sum(Ids,P,E),P,Ids,E).
default_declaration(general_product(Ids,P,E),P,Ids,E).
default_declaration(quantified_union(Ids,P,E),P,Ids,E).
default_declaration(quantified_intersection(Ids,P,E),P,Ids,E).
default_declaration(any(Ids,P,S),P,Ids,S).
default_declaration(becomes_such(Ids,P),P,Ids,[]).
default_declaration(evb2_becomes_such(Ids,P),P,Primed,[]) :- nl,print(evb2(Ids,Primed)),nl,nl, % no longer used, as it is translated to becomes_such
   prime_identifiers(Ids,Primed).
default_declaration(rlevent(I,Sec,St,Ps,G,Ts,As,VWs,PWs,Ums,Rs),G,Ps,
                        [I,Sec,St,Ps,Ts,As,VWs,PWs,Ums,Rs]).

non_default_declaration(operation(_I,Rs,As,_TBody),Ids) :-
    append(Rs,As,Ids).

add_non_default_declaration(operation(I,Rs,As,TBody),Decl,operation(I,Rs,As,NTBody)) :-
    ( remove_bt(TBody,precondition(P,S),precondition(NP,S),NTBody) ->
        conjunct_predicates([Decl,P],NP)
    ; otherwise ->
        create_texpr(precondition(Decl,TBody),subst,[],NTBody)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pattern definitions (not yet finished)

bsyntax_pattern(Expr,TExpr) :-
    var(Expr),!,Expr=TExpr.
bsyntax_pattern(-Expr,TExpr) :-
    !,remove_all_infos(Expr,TExpr).
bsyntax_pattern(Expr,TExpr) :-
    functor(Expr,b,3),!,Expr=TExpr.
bsyntax_pattern(Expr:Type/Info,TExpr) :-
    !,bsyntax_pattern2(Expr,Type,Info,TExpr).
bsyntax_pattern(Expr:Type,TExpr) :-
    !,bsyntax_pattern2(Expr,Type,_Info,TExpr).
bsyntax_pattern(Expr/Info,TExpr) :-
    !,bsyntax_pattern2(Expr,_Type,Info,TExpr).
bsyntax_pattern(Expr,TExpr) :-
    !,bsyntax_pattern2(Expr,_Type,_Info,TExpr).

bsyntax_pattern2(Pattern,Type,Info,TExpr) :-
    functor(Pattern,Functor,Arity),
    functor(R,Functor,Arity),
    create_texpr(R,Type,Info,TExpr),
    syntaxelement(Pattern,PSubs,_,PLists,Const,EType),
    syntaxelement(R,RSubs,_,RLists,Const,EType),
    ( EType==pred -> Type=pred
    ; EType==subst -> Type=subst
    ; otherwise -> true),
    all_same_length(PLists,RLists),
    maplist(bsyntax_pattern,PSubs,RSubs).

all_same_length([],[]).
all_same_length([A|Arest],[B|Brest]) :-
    ( var(A),var(B) ->
        add_error_fail(bsyntaxtree,'At least one list should contain nonvar elements for all_same_length',[A,B])
    ; otherwise ->
        same_length(A,B)),
    all_same_length(Arest,Brest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% strip an AST into a more compact form (without b/3 terms)

strip_and_norm_ast(TExpr,SNExpr) :-
    get_texpr_expr(TExpr,Expr),
    strip_and_norm_ast_aux(Expr,SNExpr).

:- use_module(tools,[safe_univ_no_cutoff/2]).
strip_and_norm_ast_aux(Expr,Res) :-
   comm_assoc_subs(Expr,Op,Subs,[]), !, % associative & commutative operator detected
   maplist(strip_and_norm_ast_aux,Subs,NSubs), % Subs are already unwrapped
   sort(NSubs,Sorted), % in case there are associative operators that are not commutative: insert is_commutative check
   safe_univ_no_cutoff(Res,[Op|Sorted]).
strip_and_norm_ast_aux(Expr,Res) :-
   assoc_subs(Expr,Op,Subs,[]), !, % associative operator detected
   maplist(strip_and_norm_ast_aux,Subs,NSubs), % Subs are already unwrapped
   safe_univ_no_cutoff(Res,[Op|NSubs]).
strip_and_norm_ast_aux(Expr,SNExpr) :-
    syntaxtransformation(Expr,Subs,_,NSubs,SExpr),
    strip_and_norm_ast_l(Subs,NSubs),
    norm_strip(SExpr,SNExpr).

strip_and_norm_ast_l([],[]).
strip_and_norm_ast_l([TExpr|Trest],[NExpr|Nrest]) :-
    strip_and_norm_ast(TExpr,NExpr),
    strip_and_norm_ast_l(Trest,Nrest).

norm_strip(greater_equal(A,B),less_equal(B,A)) :- !.
norm_strip(greater(A,B),less(B,A)) :- !.
norm_strip(set_extension(NL),set_extension(SNL)) :- !,
    sort(NL,SNL).
norm_strip(Old,New) :-
    functor(Old,Functor,2),
    is_commutative(Functor),!,
    arg(1,Old,OA),
    arg(2,Old,OB),
    ( OA @> OB -> New =.. [Functor,OB,OA]
    ; New=Old).
norm_strip(Old,Old).
% TO DO: flatten associative operators into lists !

comm_assoc_subs(conjunct(TA,TB),conjunct) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   comm_assoc_subs(A,conjunct), comm_assoc_subs(B,conjunct).
comm_assoc_subs(disjunct(TA,TB),disjunct) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   comm_assoc_subs(A,disjunct), comm_assoc_subs(B,disjunct).
comm_assoc_subs(add(TA,TB),add) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   comm_assoc_subs(A,add), comm_assoc_subs(B,add).
comm_assoc_subs(multiplication(TA,TB),multiplication) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   comm_assoc_subs(A,multiplication), comm_assoc_subs(B,multiplication).
comm_assoc_subs(union(TA,TB),union) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   comm_assoc_subs(A,union), comm_assoc_subs(B,union).
comm_assoc_subs(intersection(TA,TB),intersection) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   comm_assoc_subs(A,intersection), comm_assoc_subs(B,intersection).
comm_assoc_subs(Expr,Op) --> {nonvar(Op)},[Expr]. % base case for other operators

% detect just associative operators
assoc_subs(concat(TA,TB),concat) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   assoc_subs(A,concat), assoc_subs(B,concat).
assoc_subs(composition(TA,TB),composition) --> !,
   {get_texpr_expr(TA,A), get_texpr_expr(TB,B)},
   assoc_subs(A,composition), assoc_subs(B,composition).
assoc_subs(Expr,Op) --> {nonvar(Op)},[Expr]. % base case for other operators

is_commutative(conjunct).
is_commutative(disjunct).
is_commutative(equivalence).
is_commutative(equal).
is_commutative(not_equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).

% check if two type expressions are same modulo info fields and reordering of commutative operators
% same_texpr does not reorder wrt commutativity.
same_norm_texpr(TExpr1,TExpr2) :-
   strip_and_norm_ast(TExpr1,N1),
   strip_and_norm_ast(TExpr2,N1).


% small utility to get functor of texpr:
get_texpr_functor(b(E,_,_),F,N) :- !, functor(E,F,N).
get_texpr_functor(E,_,+) :- add_error_fail(get_texpr_functor,'Not a typed expression: ',E).

% -------------------------

% check if a ProB type is a set type:
is_set_type(set(Type),Type).
is_set_type(seq(Type),couple(integer,Type)).
% should we have a rule for any ?? : in all cases using is_set_type any seems not possible; better that we generate error message in get_set_type or get_texpr_set_type
% is_set_type(any,any).

get_set_type(TypeX,Res) :-
   (is_set_type(TypeX,SetType) -> Res=SetType ; add_error_fail(get_set_type,'Not a set type: ',TypeX)).

get_texpr_set_type(X,Res) :- get_texpr_type(X,TypeX),
   get_set_type(TypeX,Res).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if a set contains all elements of a type, i.e., maximal type
is_just_type(Expr) :- is_just_type(Expr,[]).
% is_just_type(+Expr,+RTT):
%  like is_just_type/1 but RTT is a list of identifiers (without type information)
%  of variables or constants that are known to be types, too.
%  E.g. is_just_type(a ** INTEGER) would fail but is_just_type(a ** INTEGER,[a]) would
%  succeed.
is_just_type(b(E,T,I),RTT) :- is_just_type3(E,T,I,RTT).


is_just_type3(identifier(G),set(Type),Infos,RefsToTypes) :-
    ( Type = global(G), memberchk(given_set,Infos)
    ; memberchk(G,RefsToTypes)),!.
is_just_type3(freetype_set(_),_,_,_).
is_just_type3(pow_subset(E),_,_,RTT) :- is_just_type(E,RTT).
is_just_type3(fin_subset(E),T,_,RTT) :- \+ is_infinite_type(T),is_just_type(E,RTT).
is_just_type3(integer_set('INTEGER'),set(integer),_,_).
is_just_type3(bool_set,set(boolean),_,_).
is_just_type3(real_set,set(real),_,_).
is_just_type3(string_set,set(string),_,_).
is_just_type3(cartesian_product(A,B),_,_,RTT) :-
    is_just_type(A,RTT),is_just_type(B,RTT).
is_just_type3(mult_or_cart(A,B),_,_,RTT) :-    % is_just_type/1 could be called before the cleanup
    is_just_type(A,RTT),is_just_type(B,RTT).   % of an expression has finished
is_just_type3(relations(A,B),_,_,RTT) :-
    is_just_type(A,RTT),is_just_type(B,RTT).
is_just_type3(struct(b(rec(Fields),_,_)),_,_,RTT) :-
    maplist(field_is_just_type(RTT),Fields).
is_just_type3(typeset,_,_,_).
is_just_type3(comprehension_set(_,b(truth,_,_)),_,_,_).
is_just_type3(value(Val),ST,_,_) :-  is_set_type(ST,Type), is_maximal_value(Val,Type). % also see: quick_is_definitely_maximal_set(Val).

field_is_just_type(RTT,field(_,Set)) :- is_just_type(Set,RTT).

is_maximal_value(V,_) :- var(V),!,fail.
is_maximal_value(global_set(GS),Type) :-  (GS=='INTEGER' -> true ; Type==global(GS)). % we could call is_maximal_global_set(GS) or b_global_set (but requires compiled)

% Generate a custom matching / visitor predicate
% bsyntaxtree:gen_visitor(bexpr_variables_aux,bexpr_variables)
/*
gen_visitor(Pred, BodyPred) :- syntaxelement(Expr,SubsX,_NamesX,_Lists,_ConstantX,_),
    print_clause(Pred,BodyPred,Expr,SubsX),
    fail.
gen_visitor(Pred, BodyPred) :- nl.

print_clause(Pred,BodyPred,Expr,SubsX) :- E=..[Op|SubsX], format('~w(~w) :- ',[Pred,Expr]).
*/

/*
 * create_recursive_compset(+Ids,+Cond,+Type,+Infos,-RecId,-TCompSet)
 *  creates a recursive comprehension set:
 *   Ids      is the list of the introduced typed identifiers
 *   Cond     is the condition (a typed predicate)
 *   Infos    are additional informations for the comprehension set syntax element
 *   RecId    is the identifier (untyped, an atom) that can be used in Cond to refer to the
 *            comprehension set recursively. Usually RecId is used in Cond as a variable
 *   TCompSet is the generated recursive comprehension set, the recursion parameter is introduced
 *            by a recusive(Id,CompSet) syntax element
 */
create_recursive_compset(Ids,Cond,Type,Infos,RecId,TCompSet) :-
    unique_id("recursive.",RecId),
    add_symbolic_annotation(Infos,RInfos),
    create_texpr(comprehension_set(Ids,Cond),Type,RInfos,TCompSet1),
    create_typed_id(RecId,Type,TRecId),
    create_texpr(recursive_let(TRecId,TCompSet1),Type,[],TCompSet).

unique_typed_id(Prefix,Type,TId) :-
    unique_id(Prefix,Id),
    create_typed_id(Id,Type,TId).


mark_bexpr_as_symbolic(b(E,T,I),b(E2,T,RI)) :-
     add_symbolic_annotation(I,RI),
     mark_aux(E,E2).
mark_aux(union(A,B),R) :- !, R=union(SA,SB),
   mark_bexpr_as_symbolic(A,SA), mark_bexpr_as_symbolic(B,SB).
% union is currently the only operator that is kept symbolically
mark_aux(A,A).

add_symbolic_annotation([H|I],R) :- H == prob_annotation('SYMBOLIC'),!, R=[H|I].
add_symbolic_annotation(I,R) :- !, R=[prob_annotation('SYMBOLIC')|I].

% identifier_sub_ast(+TExpr,+Identifier,-SubPosition):
% Find occurences of an identifier "Identifier" in an expression "TExpr"
% Returns a list of positions (beginning with 0) that describes the position of a sub-expression
% that contains all occurrences of Identifier in TExpr.
% E.g. [1,0,1] means "second sub-expression of TExpr, then the first sub-expression of that,
% then the second sub-expression of that".
% When an identifier has multiple occurences, the position of the sub-expression is returned
% where all occurences of it are located.
% If the identifier is not found, the call fails.
% Sub-expressions are meant like the ones syntaxtraversion or syntaxtransformation return.
identifier_sub_ast(TExpr,Identifier,SubPosition) :-
    syntaxtraversion(TExpr,Expr,_Type,_Infos,AllSubs,Names),
    identifier_sub_ast_aux(Expr,AllSubs,Names,Identifier,SubPosition).
identifier_sub_ast_aux(identifier(Identifier),[],[],Identifier,[]) :- !.
identifier_sub_ast_aux(_Expr,AllSubs,Names,Identifier,SubPosition) :-
    get_texpr_id(TId,Identifier),nonmember(TId,Names),
    % For each sub-expression E in the list AllSubs create a term Pos-E
    % where Pos is E's position (starting with 0) in AllSubs
    foldl(annotate_pos,AllSubs,AllAnnotatedSubs,0,_),
    % Introduced identifiers occur alsa as sub-expressions, remove those
    foldl(select_sub,Names,AllAnnotatedSubs,RealAnnSubs),
    convlist(identifier_sub_ast_aux2(Identifier),RealAnnSubs,SubPositions),
    ( SubPositions = [SubPosition] -> true
    ; SubPositions = [_,_|_] -> % More then one occurrences - make this the found node
        SubPosition = []).
identifier_sub_ast_aux2(Identifier,Pos-TExpr,[Pos|SubPosition]) :-
    identifier_sub_ast(TExpr,Identifier,SubPosition).
annotate_pos(TExpr,I-TExpr,I,I2) :- I2 is I+1.
select_sub(Name,AnnotatedSubs,Result) :-
    selectchk(_Pos-Name,AnnotatedSubs,Result).

% exchange_ast_position(+SubPosition,+OldTExpr,-OldInner,+NewInner,-NewTExpr):
% exchanges a sub-expression in the expression "OldTExpr".
% SubPosition is a list of positions like identifier_sub_ast/3 returns it.
% OldInner is the sub-expression found in OldTExpr.
% NewInner is the new sub-expression.
% NewTExpr is the new expression that originates from replacing OldInner by NewInner.
exchange_ast_position([],Old,Old,New,New).
exchange_ast_position([Pos|RestPos],OldTExpr,OldInner,NewInner,NewTExpr) :-
    remove_bt_and_used_ids(OldTExpr,OldExpr,NewExpr,NewTExpr), % also invalidates used_ids info
    syntaxtransformation(OldExpr,Subs,_Names,NSubs,NewExpr),
    nth0(Pos,Subs, OldSelected,Rest),
    nth0(Pos,NSubs,NewSelected,Rest),
    exchange_ast_position(RestPos,OldSelected,OldInner,NewInner,NewSelected).


% -----------------------

:- public check_used_ids/2.
% a simple checker, only checks used_ids info fields for entire typed expression:
check_used_ids(TExpr,PP) :- %format('CHECK AST for ~w~n',[PP]),
    map_over_typed_bexpr(check_used_ids_texpr_fail(PP),TExpr).
check_used_ids(_,_).
check_used_ids_texpr_fail(PP,E) :- check_used_ids_texpr(PP,E),!,fail.
check_used_ids_texpr(PP,b(P,T,I)) :-  member(used_ids(UIds1),I),!,
    (find_identifier_uses(b(P,T,I),[],UIds2) -> true
      ; add_internal_error('find_identifier_uses failed',PP),fail),
    (UIds1==UIds2 -> true
     ; format('*** Wrong used_ids Info (~w)!!~n  Used_ids: ~w~n  Comp_ids: ~w~n',[PP,UIds1,UIds2]),
       translate:print_bexpr(b(P,T,I)),nl,
       add_error(check_used_ids,'Wrong used_ids: ',PP,I)
    ).
check_used_ids_texpr(_,_).

% a simple checker to see if an AST is well-formed:
% can be tested e.g. as follows: b_get_invariant_from_machine(I), check_ast(I).
% called in prob_safe_mode by clean_up_section

check_ast(TE) :- check_ast(false,TE).
check_ast(AllowVars,TExpr) :- %nl,print('CHECK AST'),nl,
    map_over_typed_bexpr(check_ast_texpr(AllowVars),TExpr).
check_ast(_,_).

:- use_module(typing_tools,[valid_ground_type/1]).
%check_ast_texpr(X) :- print(check(X)),nl,fail.
check_ast_texpr(AllowVars,b(E,Type,Infos)) :-
   (check_expr(E,Type,Infos) -> true ; add_error(check_ast_texpr,'Invalid Expr: ', b(E,Type,Infos)), trace, check_expr(E,Type,Infos)),
   (check_type(Type,AllowVars) -> true
    ; add_error(check_ast_texpr,'Invalid Type for: ',b(E,Type,Infos),Infos)),
   safe_functor(E,F),
   check_ast_typing(E,Type),
   (check_infos(Infos,F) -> true ; add_error(check_ast_texpr,'Invalid Infos: ',b(E,Type,Infos))),
   fail. % to force backtracking in map_over_typed_bexpr

safe_functor(V,R) :- var(V),!,R='$VAR'.
safe_functor(E,F/N) :- functor(E,F,N).

% check whether the type term is ok
check_type(X,AllowVars) :- var(X),!,
  (AllowVars==true -> true ; add_error(check_type,'Variable type: ',X),fail).
check_type(pred,_) :- !.
check_type(subst,_) :- !.
check_type(op(Paras,Returns),AllowVars) :- !,
    maplist(check_normal_type(AllowVars),Paras),
    maplist(check_normal_type(AllowVars),Returns).
check_type(T,AllowVars) :- check_normal_type(AllowVars,T).

check_normal_type(AllowVars,T) :-
 (AllowVars \== true -> valid_ground_type(T)
   ; ground(T) -> valid_ground_type(T)
   ; true).  % TO DO: call valid_ground_type but pass AllowVars

% check whether type is compatible with operators
check_ast_typing(member(_,B),Type) :- !, get_texpr_type(B,T),check_set_type(T,member), check_type(Type,pred,member).
check_ast_typing(greater_equal(A,B),Type) :- !,
     get_texpr_type(A,TA),check_type(TA,integer,greater_equal),
     get_texpr_type(B,TB),check_type(TB,integer,greater_equal), check_type(Type,pred,member).
% TO DO: use rules from typechecker
check_ast_typing(_,_).

check_set_type(set(_),_) :- !.
check_set_type(seq(_),_) :- !.
check_set_type(Type,Func) :- add_error(check_ast_typing,'Invalid type for operator: ',Func:Type).

check_type(Type,Type,_) :- !.
check_type(Type,_,Func) :- add_error(check_ast_typing,'Unexpected type for operator: ',Func:Type).

check_infos(X,F) :- var(X),!, add_error(check_infos,'Info field list not terminated: ',F:X),fail.
check_infos([],_).
check_infos([H|_],F) :- \+ ground(H),!, add_error(check_infos,'Info field not ground: ',F:H),fail.
check_infos([[H|T]|_],F) :- !, add_error(check_infos,'Info field contains nested list: ',F:[H|T]),fail.
check_infos([cse_used_ids(H)|T],F) :- member(cse_used_ids(H2),T),!,
   add_error(check_infos,'Multiple cse_used_ids entry: ',F:[H,H2]),fail.
check_infos([used_ids(H)|T],F) :- member(used_ids(H2),T),!,
   add_error(check_infos,'Multiple used_ids entry: ',F:[H,H2]),fail.
check_infos([_|T],F) :- check_infos(T,F).

check_expr(Expr,Type,Infos) :- nonmember(contains_wd_condition,Infos),
   sub_expression_contains_wd_condition(Expr,Sub),
   TE = b(Expr,Type,Infos),
   (Type = subst -> fail % AST cleanup is not called for substitutions; WD-info not available for substitutions at the moment
                    %translate:translate_substitution(TE,PS)
    ; translate_bexpression(TE,PS)),
   functor(Expr,Functor,_),
   functor(Sub,SubFunctor,_),
   tools:ajoin(['Node for AST node ',Functor,' does not contain WD info from Subexpression ',SubFunctor,' :'],Msg),
   add_warning(check_expr,Msg,PS,TE).
check_expr(member(LHS,RHS),Type,Infos) :- is_just_type(RHS),
   get_preference(optimize_ast,true),
   !,
   TE = b(member(LHS,RHS),Type,Infos),
   translate:translate_bexpression(TE,PS),
   add_warning(check_expr,'AST contains redundant typing predicate: ',PS,TE).
check_expr(identifier(ID),_,_) :- illegal_id(ID),!,
   add_error(check_infos,'Illegal identifier: ', identifier(ID)).
check_expr(lazy_lookup_expr(ID),_,_) :- illegal_id(ID),!,
   add_error(check_infos,'Illegal identifier: ', lazy_lookup_expr(ID)).
check_expr(lazy_lookup_pred(ID),_,_) :- illegal_id(ID),!,
   add_error(check_infos,'Illegal identifier: ', lazy_lookup_pred(ID)).
check_expr(exists(Parameters,Condition),pred,Infos) :- !,
   b_interpreter:get_exists_used_ids(Parameters,Condition,Infos,_Used). % this will do some checks
check_expr(value(V),Type,_) :- !, check_bvalue(V,Type).
check_expr(_,_,_).

illegal_id(ID) :- var(ID),!.
illegal_id(op(ID)) :- !, \+ atom(ID).
illegal_id(ID) :- \+ atom(ID).

:- use_module(btypechecker, [unify_types_strict/2, couplise_list/2]).
:- use_module(avl_tools,[check_is_non_empty_avl/1]).
% TO DO: we could check type more
check_bvalue(V,_) :- var(V),!.
check_bvalue(avl_set(A),Type) :- !, unify_types_strict(Type,set(_)),
    check_is_non_empty_avl(A).
check_bvalue(closure(_,T,B),Type) :- !,
    couplise_list(T,CT), unify_types_strict(set(CT),Type),
    check_ast(B).
check_bvalue(_,_).

% -----------------------


indent_ws(X) :- X<1,!.
indent_ws(X) :- print('  '), X1 is X-1, indent_ws(X1).

print_ast_td(b(E,T,I),Level,L1) :-
   indent_ws(Level),
   (E=identifier(_)
     -> format('~w (~w) -> ~w~n',[E,T,I])
     ;  functor(E,F,N),
        format('~w/~w (~w) -> ~w~n',[F,N,T,I])
    ),
   L1 is Level+1.
print_ast(TExpr) :-
   (map_over_type_bexpr_top_down_acc(print_ast_td,TExpr,0),fail ; true).

% ---------------------

:- dynamic count_id_usage/2.
single_usage_id_count(Expr) :- uses_an_identifier(Expr,Id),
   retract(count_id_usage(Id,Count)),
   !,
   Count=0,
   C1 is Count+1,
   assert(count_id_usage(Id,C1)).
single_usage_id_count(_).

% check if an identifier is used at most once
single_usage_identifier(ID,ExprOrPredicates,Count) :-
   retractall(count_id_usage(_,_)),
   assert(count_id_usage(ID,0)),
   % TO DO: take care of naming; do not count occurences when we enter scope defining ID
   maplist(single_usage_cnt,ExprOrPredicates),
   retract(count_id_usage(ID,Count)).

single_usage_cnt(ExprOrPredicate) :- map_over_full_bexpr_no_fail(single_usage_id_count,ExprOrPredicate).


