% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(chr_set_membership,[chr_subset/2, chr_not_subset/2,
                              chr_subset_strict/2, chr_not_subset_strict/2,
                              chr_member/2, chr_not_member/2]).

:- use_module(probsrc(module_information),[module_info/2]).
:- module_info(group,experimental).
:- module_info(description,'Provide interface to CHR set reasoning.').

:- use_module(library(chr)).
:- chr_option(check_guard_bindings,off).
:- chr_option(optimize,full).

% restore operators overwritten by chr library
% otherwise, unary operators might not be parseable
:- op(200, fy, [+,-]).
:- op(500, yfx, \).

% assumes we use CLP(FD) to complement it

chr_subset(S1,S2) :- subset(S1,S2).
chr_not_subset(S1,S2) :- not_subset(S1,S2).

chr_subset_strict(S1,S2) :- subset_strict(S1,S2) .
chr_not_subset_strict(S1,S2) :- not_subset_strict(S1,S2).

chr_member(E,S) :- member_of(E,S).
chr_not_member(E,S) :- not_member_of(E,S).
% TODO: reification-trigger: when can we infer membership or non-membership

:- chr_constraint subset/2, not_subset/2.
:- chr_constraint subset_strict/2, not_subset_strict/2.
:- chr_constraint member_of/2, not_member_of/2.

:- public '@'/2.
contradiction @ subset(X,Y), not_subset(X,Y) <=> fail.
reflexivity   @ subset(X,X) <=> true.
antisymmetry  @ subset(X,Y), subset(Y,X) <=> X = Y.
idempotence   @ subset(X,Y) \ subset(X,Y) <=> true.
transitivity  @ subset(X,Y), subset(Y,Z) ==> subset(X,Z).
%checking      @ subset(X,Y) <=> nonvar(X),nonvar(Y) | X =< Y.

contradiction   @ subset_strict(X,Y), not_subset_strict(X,Y) <=> fail.
contradiction   @ subset_strict(X,Y), not_subset(X,Y) <=> fail.
antireflexivity @ subset_strict(X,X) <=> fail.
idempotence     @ subset_strict(X,Y) \ subset_strict(X,Y) <=> true.
transitivity    @ subset_strict(X,Y), subset(Y,Z) ==> subset_strict(X,Z).
transitivity    @ subset(X,Y), subset_strict(Y,Z) ==> subset_strict(X,Z).
transitivity    @ subset_strict(X,Y), subset_strict(Y,Z) ==> subset_strict(X,Z).
%checking        @ lt(X,Y) <=> nonvar(X),nonvar(Y) | X < Y.

idempotence     @ member_of(X,Y) \ member_of(X,Y) <=> true.
idempotence     @ not_member_of(X,Y) \ not_member_of(X,Y) <=> true.
contradiction @ member_of(X,Y), not_member_of(X,Y) <=> fail.
contradiction @ member_of(X,Y), not_member_of(X,Z), subset(Y,Z) <=> fail.
contradiction @ member_of(X,Y), not_member_of(X,Z), subset_strict(Y,Z) <=> fail.

% member_of(X,Y) => not_empty(Y)
% not_empty(X), empty(X) <=> fail.
% subset(X,Y), not_empty(X) => not_empty(Y).
% subset_strict(X,Y) => not_empty(Y).
