/* -*- Mode:Prolog; coding:ISO-8859-1; -*- */
:- module(csp_basic,[safe_less_than/2, safe_less_than_equal/2,
                     safe_less_than/3, safe_less_than_equal/3,
                     enumerate_csp_int/3]).

:- use_module(probsrc(module_information)).
:- module_info(group,csp).
:- module_info(description,'Basic operations on CSP values.').

:- use_module(probsrc(self_check)).
:- assert_must_succeed( (csp_basic:safe_less_than(X,Y), X=1,Y=2)).
:- assert_must_fail( (csp_basic:safe_less_than(X,Y), X=1,Y=1)).
:- assert_must_fail( (csp_basic:safe_less_than(X,Y), X=2,Y=1)).

:- block safe_less_than(-,?),safe_less_than(?,-).
safe_less_than(X,Y) :- X<Y.

:- assert_must_succeed( (csp_basic:safe_less_than(X,Y,R), X=1,Y=2,R==true)).
:- assert_must_succeed( (csp_basic:safe_less_than(X,Y,R), X=2,Y=2,R==false)).
:- block safe_less_than(-,?,?),safe_less_than(?,-,?).
safe_less_than(X,Y,R) :- (X<Y -> R=true ; R=false).

:- assert_must_succeed( (csp_basic:safe_less_than_equal(X,Y), X=1,Y=2)).
:- assert_must_succeed( (csp_basic:safe_less_than_equal(X,Y), X=1,Y=1)).
:- assert_must_fail( (csp_basic:safe_less_than_equal(X,Y), X=2,Y=1)).
:- block safe_less_than_equal(-,?),safe_less_than_equal(?,-).
safe_less_than_equal(X,Y) :- X=<Y.

:- assert_must_succeed( (csp_basic:safe_less_than_equal(X,Y,R), X=1,Y=2,R==true)).
:- assert_must_succeed( (csp_basic:safe_less_than_equal(X,Y,R), X=2,Y=2,R==true)).
:- assert_must_succeed( (csp_basic:safe_less_than_equal(X,Y,R), X=2,Y=1,R==false)).
:- block safe_less_than_equal(-,?,?),safe_less_than_equal(?,-,?).
safe_less_than_equal(X,Y,R) :- (X=<Y -> R=true ; R=false).

:- assert_must_succeed( (csp_basic:enumerate_csp_int(X,1,10), X==5)).
enumerate_csp_int(N,X,Y) :- X=<Y,
  (N=X ; X1 is X+1, enumerate_csp_int(N,X1,Y)).
