% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(fd_utils_clpfd,[eq_fd/2, neq_fd/3,
   in_fd/3, in_fd_wf/4,
   /* neq_fdpair/4, not_in_fd/3, not used anymore */
   enum_fd/3, enum_fd_linear/3]).

 :- use_module(module_information).
:- module_info(group,kernel).
:- module_info(description,'Basic checks for SET elements (equality, ...); maps to CLP(FD).').

/* File: fd_utils_clpfd.pl */
/* Created: 21/3/06 by Michael Leuschel */


:- use_module(self_check).

:- use_module(library(clpfd)).

:- assert_must_succeed(( fd_utils_clpfd:eq_fd(2,2) )).
:- assert_must_fail(( fd_utils_clpfd:eq_fd(2,3) )).

eq_fd(X,Y) :- X=Y.

%eq_fd(X,Y) :- %print_message(informational,eq_fd(X,Y)),
% %((var(X),frozen(X,G)) -> print_message(informational,frozen(X,G)) ; true),
% X = Y,
% print_message(informational,done_eq(X,Y)).


:- assert_must_succeed(( fd_utils_clpfd:in_fd(X,1,3),
                         fd_utils_clpfd:neq_fd(X,1,'Name'), fd_utils_clpfd:neq_fd(X,3,'Name'), X==2 )).

:- assert_must_succeed(( fd_utils_clpfd:neq_fd(1,2,'Name') )).
:- assert_must_fail(( fd_utils_clpfd:neq_fd(2,2,'Name') )).

neq_fd(X,Y,_Type) :-  %print(neq_fd(X,Y,_Type)),nl,
   % sometimes no CLP(FD) constraints set up for X or Y; can mean that we only detect inconsistencies much later
   %(var(X) -> b_global_sets:global_type(fd(X,_Type),_Type) ; true), % commenting it in causes test 496 to fail;
   %(var(Y) -> b_global_sets:global_type(fd(Y,_Type),_Type) ; true), % TO DO: investigate
   X #\= Y,
   ((var(X),var(Y)) -> dif(X,Y) ; true). % even for finite domains  X in 1..10, Y in 1..10, X#\=Y, X=Y. does not fail;



/*
:- assert_must_succeed(( fd_utils_clpfd:neq_fdpair(2,3,2,4) )).
:- assert_must_succeed(( fd_utils_clpfd:neq_fdpair(2,3,3,3) )).
:- assert_must_fail(( fd_utils_clpfd:neq_fdpair(2,3,2,3) )).

neq_fdpair(X1,Y1,X2,Y2) :-
  when((?=(X1,X2);?=(Y1,Y2)),
        (?=(X1,X2) -> (X1 \= X2 -> true ; dif(Y1,Y2))
                   ;  (Y1 \= Y2 -> true ; dif(X1,X2))
        )).
*/

:- use_module(kernel_waitflags,[add_fd_variable_for_labeling/2]).

in_fd(X,Low,Up) :- number(Low), number(Up), !, X in Low..Up.
in_fd(X,Low,Up) :- X #>=Low, X#=<Up. %when((ground(Low),ground(Up)), X in Low..Up).

in_fd_wf(X,Low,Up,WF) :-  number(Low), number(Up), !, X in Low..Up,
    in_fd_wf2(X,Low,Up,WF).
in_fd_wf(X,Low,Up,WF) :- X #>=Low, X#=<Up, in_fd_wf2(X,Low,Up,WF).

:- block in_fd_wf2(?,-,?,?), in_fd_wf2(?,?,-,?).
in_fd_wf2(X,_Low,_Up,WF) :-
  (ground(X) -> true
    ; %Size is Up+1-Low,
      %fd_size(X,Size),
      add_fd_variable_for_labeling(X,WF) % will automatically do enumeration
  ).

/* TO DO: check if we have to add FD variables to kernel_waitflags store using label_clpfd_vars
 in other circumstances;
   Provide support for x: {aa,bb,cc} ... constraints
*/




enum_fd(X,Low,Up) :- % we could use clpfd labeling
    preferences:get_preference(randomise_enumeration_order,true)
    -> %print(random_enum(X,Low,Up)),nl,
       enum_fd_random(X,Low,Up)
    ;  enum_fd_linear(X,Low,Up).
enum_fd_linear(X,Low,Up) :-
    Low =< Up
    -> ( %print(enum_fd(X,Low,Up)),translate:print_clpfd_variable(X),nl,
        X=Low %, tools:print_bt_message(done_enum_fd(X,Low,Up))
           ;
          %print(continue(X,Low,Up)),nl,
          % TO DO:  We could assert that X>Low and/or try to trigger all equalities for X=y with y<=Low
          L1 is Low+1, enum_fd_linear(X,L1,Up)
        )
     ;  fail.


:- use_module(library(random)).
:- use_module(extension('random_permutations/random_permutations')).
enum_fd_random(Var,Low,Up) :-
    init_random_permutations,
    IntervalLength is Up - Low + 1,
    get_num_bits(IntervalLength,MaxIdx,NumBits),
    get_masks(NumBits,LeftMask,RightMask),
    % the seed relies on the random predicate, not on now/1, thus prob can be made deterministic by setting a central random seed
    random(TempSeed),
    Seed is floor(TempSeed * 10000),
    enum_fd_random_aux(Var,0,MaxIdx,Low,Up,Seed,NumBits,LeftMask,RightMask).

enum_fd_random_aux(Var,CurIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask) :-
    random_permutation_element(CurIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask,Drawn,NextIdx),
    (Var=Drawn ; enum_fd_random_aux(Var,NextIdx,MaxIdx,From,To,Seed,NumBits,LeftMask,RightMask)).


% not really required, would be a type error
%:- block not_in_fd(-,?,?), not_in_fd(?,-,?), not_in_fd(?,?,-).
%not_in_fd(X,Low,Up) :-
%   (X < Low ; X > Up),!.
