% (c) 2012-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(interval_calc,[interval_addition/3
                        ,interval_subtraction/3
                        ,interval_multiplication/3
                        ,interval_negation/2
                        ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,analysis).
:- module_info(description,'This module provides methods to calculate with interval expressions.').

:- use_module(inf_arith).

interval_addition(irange(Am,AM2),irange(Bm,BM2),irange(Cm,CM2)) :-
    Cm infis Am+Bm, CM2 infis AM2+BM2.
interval_subtraction(irange(Am,AM2),irange(Bm,BM2),irange(Cm,CM2)) :-
    Cm infis Am-BM2, CM2 infis AM2-Bm.
interval_multiplication(irange(Am,AM2),irange(Bm,BM2),irange(Cm,CM2)) :-
    MultExprs = [Am*Bm,Am*BM2,AM2*Bm,AM2*BM2],
    get_min(MultExprs,Cm),
    get_max(MultExprs,CM2).
interval_negation(irange(Am,AM2),irange(Cm,CM2)) :-
    CM2 infis -Am, Cm infis -AM2.

get_min([A],Min) :- !, Min infis A.
get_min([A|Rest],Min) :-
    get_min(Rest,Rmin), Min infis min(A,Rmin).
get_max([A],Max) :- !,Max infis A.
get_max([A|Rest],Max) :-
    get_max(Rest,Rmax), Max infis max(A,Rmax).
