% (c) 2016 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_frozen_info,[kfi_domain/2, kfi_range/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module detects information about variables by using frozen.').

% light-weight CHR-style propagations
% ideally we should rewrite this to use CHR itself or attributes
% indeed, the code below is very limited: will find only one domain/range co-routine,
%  is quite brittle wrt the particular pending co-routine, may be expensive if lot of co-routines attached,...

:- use_module(debug).

kfi_domain(V,Domain) :- var(V), frozen(V,Frozen),
   attached_domain_info(Frozen,V,Domain), %print(attached(V,Domain)),nl,nl,
   debug_println(5,found_kfi_attached_domain(V,Domain)).

attached_domain_info((A,B),Var,Dom) :-
   (attached_domain_info(A,Var,R) -> Dom=R ; attached_domain_info(B,Var,Dom)).
attached_domain_info(kernel_mappings:binary_in_kernel_call2(_Module,Pred,X,Arg1,Arg2,_WF,_WF0),Var,Dom) :-
   binary_in_domain(Pred,X,Arg1,Arg2,Var,Dom).


% to do: add other infos/co-routines
binary_in_domain(total_function_wf,X,TFDom,_TFRange,Var, Dom) :- Var==X, Dom=TFDom.
binary_in_domain(total_injection_wf,X,TFDom,_TFRange,Var, Dom) :- Var==X, Dom=TFDom.
binary_in_domain(total_surjection_wf,X,TFDom,_TFRange,Var, Dom) :- Var==X, Dom=TFDom.
binary_in_domain(total_bijection_wf,X,TFDom,_TFRange,Var, Dom) :- Var==X, Dom=TFDom.
binary_in_domain(total_relation_wf,X,TFDom,_TFRange,Var, Dom) :- Var==X, Dom=TFDom.
binary_in_domain(total_surjection_relation_wf,X,TFDom,_TFRange,Var, Dom) :- Var==X, Dom=TFDom.

% ------------

% TO DO: for surjections we can also do kfi_range(V,Range)

kfi_range(V,Ran) :- var(V), frozen(V,Frozen),
   attached_range_info(Frozen,V,Ran), print(attached_range(V,Ran)),nl,nl,debug_println(5,found_range(V,Ran)).

attached_range_info((A,B),Var,Ran) :-
   (attached_range_info(A,Var,R) -> Ran=R ; attached_range_info(B,Var,Ran)).
attached_range_info(kernel_mappings:binary_in_kernel_call2(_Module,Pred,X,Arg1,Arg2,_WF,_WF0),Var,Ran) :-
   binary_in_range(Pred,X,Arg1,Arg2,Var,Ran).


binary_in_range(total_surjection_wf,X,_TFDom,TFRange,Var, Ran) :- Var==X, Ran=TFRange.
binary_in_range(total_bijection_wf,X,_TFDom,TFRange,Var, Ran) :- Var==X, Ran=TFRange.
binary_in_range(total_surjection_relation_wf,X,_TFDom,TFRange,Var, Ran) :- Var==X, Ran=TFRange.
binary_in_range(partial_surjection_wf,X,_TFDom,TFRange,Var, Ran) :- Var==X, Ran=TFRange.
binary_in_range(surjection_relation_wf,X,_TFDom,TFRange,Var, Ran) :- Var==X, Ran=TFRange.
