% Heinrich Heine Universitaet Duesseldorf
% (c) 2009-2020 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(prob2_interface,
   [initialise_specification/0,
    get_version/7,
    get_prob_total_number_of_errors/1,

    load_classical_b_from_list_of_facts/2,
    update_preferences_from_spec/0, update_preferences_from_spec/1,

    load_event_b_project/4,

    load_cspm_spec_from_cspm_file/1,
    load_z_spec_from_fuzz_file/1,
    load_z_spec_from_tex_file/1,
    load_xtl_spec_from_prolog_file/1,
    load_alloy_spec_from_term/2,

    reset_animator/0, clear_animator/0, start_animation/0,

    serialize/2,
    deserialize/2,

    get_state/2,
    get_b_state/2,

    compute_operations_for_state/2,
    prob2_execute_custom_operations/6,

    get_op_from_id/4,

    is_initialised_state/1,
    is_initialised_b_state/1,
    state_property/3,
    op_timeout_occurred/2,

    get_state_errors/2,

    evaluate_formula/3, evaluate_formulas/3,
    evaluate_registered_formulas/3,
    register_prob2_formula/2, register_prob2_formulas/2,
    deactivate_prob2_formula/1, deactivate_prob2_formulas/1,
    activate_prob2_formula/1, activate_prob2_formulas/1,
    unregister_prob2_formula/1, unregister_prob2_formulas/1,
    get_animation_image_list/1, get_animation_image_matrix_for_state/6,
    get_react_to_item_right_click_options_for_state/4,
    react_to_item_right_click_option_for_state/6,

    evaluate_formula_typecheck/5,

    get_states_for_predicate/3,

    filter_states_for_predicate/3,

    get_top_level_formulas/1,
    insert_formula_for_expansion/2,
    prob2_evaluate_bvisual2_formulas/3,
    prob2_expand_bvisual2_formula/3,

    do_modelchecking/5,
    set_goal_for_model_checking/1,
    compute_efficient_statespace_stats/3,
    compute_coverage/5,
    get_modelchecking_coverage/5,
    get_statistics/2,
    prob2_deadlock_freedom_check/2,
    prob2_invariant_check/2,
    prob2_redundant_invariants/2,



    get_enable_matrix/2,

    prob2_do_ltl_modelcheck/4,

    find_trace_to_node/2,
    find_trace_from_node_to_node/3,
    find_state_for_predicate/3,

    cbc_disprove/5, cbc_disprove/6,
    cbc_solve/4, cbc_solve/5, % deprecated
    cbc_solve_with_opts/5,
    pretty_print_predicate/3,
    cbc_generate_test_cases/3,
    prob2_find_test_path/4,



    check_csp_assertions/3,

    list_eclipse_preferences/1,
    list_all_eclipse_preferences/1, % also includes advanced eclipse preferences
    list_current_eclipse_preferences/1,
    get_eclipse_preference/2,
    set_eclipse_preference/2,

    get_signature_merge_state_space/2,
    get_transition_diagram/2,

    write_dotty_transition_diagram/2,
    write_dotty_signature_merge/2,
    write_dot_for_state_viz/2,
    write_dotty_state_space/1,
    is_dotty_command/1, write_dotty/2,
    is_dotty_command_for_expr/1, write_dotty_for_expr/3,
    get_dot_commands_in_state/2, call_dot_command_in_state/4,
    get_table_commands_in_state/2, call_table_command_in_state/4,

    get_error_messages/2,
    get_error_messages_with_span_info/1,
    get_error_messages_with_span_info/2,

    es_copy_from_statespace/2,
    es_eval/4,
    evaluate_expression_prolog/3,

    generate_trace_until_condition_fulfilled/4,
    execute_model/5, execute_model/6,
    get_unsat_core_with_fixed_conjuncts/3,
    get_minimum_unsat_core_with_fixed_conjuncts/3,

    prob2_construct_trace/6,

    symbolic_model_check/2,

    % synthesis
    start_synthesis_from_ui_/13,
    start_synthesis_single_operation_from_ui_/11,
    get_valid_and_invalid_equality_predicates_for_operation_/6,
    get_valid_and_invalid_equality_predicates_for_invariants_/4,
    get_invariant_violating_vars_from_examples_/3,
    adapt_machine_code_for_operations_/2,
    reset_synthesis_context_/0,
    generate_data_from_machine_operation_/6,
    generate_synthesis_data_from_predicate_untyped_/5,
    generate_operation_data_from_machine_path_/4,

    evalb_evaluate_formula/2,
    evalb_evaluate_tautology/2,

    get_pretty_print/1, get_pretty_print_unicode/1,

    get_primed_predicate/2,
    get_weakest_precondition/3,
    before_after_predicate/2,

    get_machine_operation_names/1, get_machine_operation_infos/1,
    get_machine_identifiers/2,
    get_machine_files/1,

    get_possible_completions/3,
    get_possible_fuzzy_matches/2,

    prob2_extended_static_check/1,
    prob2_check_well_definedness/2
   ]).


:- use_module(module_information).
:- module_info(group,cli).
:- module_info(description,'This module provides the new ProB2 Prolog interface to Java and other languages (usually called via socket server).').

:- use_module(state_space,              [time_out_for_node/1, time_out_for_node/3,
                                         time_out_for_invariant/1, max_reached_for_node/1,
                                         max_reached_or_timeout_for_node/1,
                                         visited_expression/2, transition/4, visited_expression_id/1,
                                         set_context_state/1, clear_context_state/0,
                                         state_error/3, current_state_id/1,
                                         get_state_space_stats/3 %, state_space_initialise/0
                                         ]).
% new profiler
%:- use_module('../extensions/profiler/profiler.pl').


:- use_module(symmetry_marker,          [precompile_marker_typing_info/0]).
:- use_module(value_persistance,        [initialise_operation_caching/0]).
:- use_module(eventhandling,            [announce_event/1]).
%:- use_module(self_check,               [turn_off_run_time_type_checks/0]).
:- use_module(version,                  [version/4, revision/1, lastchangeddate/1]).
:- use_module(pref_definitions,         [b_get_preferences_from_machine/0, b_get_preferences_from_machine/1]).
:- use_module(bsyntaxtree,              [conjunct_predicates/2, get_texpr_ids/2,
                                         get_texpr_expr/2,
                                         find_identifier_uses/3]).
:- use_module(evalstores,               [evalstore_bindings/2, evalstore_create_store_by_state/2, evalstore_eval_formula/4]).
%:- use_module(model_checker,            [is_deadlocked/1]).
:- use_module(state_space_exploration_modes,[set_depth_breadth_first_mode/1]).
:- use_module(bmachine,                 [b_type_expression/5, b_get_machine_operation/4, b_get_machine_operation/6,
                                         b_type_open_predicate/5, b_top_level_operation/1,
                                         get_machine_operation_additional_identifiers/2,
                                         b_load_machine_from_list_of_facts/2,
                                         b_set_eventb_project_flat/3, load_additional_information/1,
                                         b_machine_precompile/0, other_spec_precompile/0,
                                         b_set_parsed_typed_machine_goal/1]).
:- use_module(probltlsrc(ltl),          [ltl_model_check_with_ce/4,preprocess_formula/2]).
:- use_module(probltlsrc(ltl_tools),    [typecheck_temporal_formula/3]).
:- use_module(probltlsrc(ltl_verification), [evaluate_ltl_formula/6]).
%:- use_module(b_interpreter,            [b_compute_expression_nowf/4]).
:- use_module(b_global_sets,            [add_prob_deferred_set_elements_to_store/3]).
:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2,
                                                     ignore_user_interrupt_det/1]).
:- use_module(library(lists),           [maplist/3, maplist/4]).
:- use_module(library(codesio),         [read_from_codes/2, write_term_to_codes/4]).
:- use_module(specfile,                 [expand_const_and_vars_to_full_store/2, animation_mode/1,
                                         csp_mode/0,
                                         b_or_z_mode/0, set_currently_opened_file/1,
                                         set_currently_opening_file/1, set_failed_to_open_file/1,
                                         set_currently_opened_package/1,
                                         set_animation_mode/1, set_animation_minor_mode/1]).
:- use_module(translate,                [translate_bvalue/2,
                                         suppress_rodin_positions/1, reset_suppress_rodin_positions/1,
                                         translate_event/2, translate_bstate/2, translate_state_error/2,
                                         translate_bexpression/2, translate_bvalue_with_limit/3,
                                         translate_event_error/2, explain_event_trace/2, explain_state_error/3]).
:- use_module(error_manager,            [add_error_and_fail/3, get_all_errors_and_reset/1,
                                         get_all_errors_with_span_info_and_reset/1,
                                         add_error/3, add_error/2, add_error_and_fail/2,
                                         add_message/2, %reset_errors/0,
                                         add_warning/3,
                                         add_all_perrors/1,
                                         get_error/2,
                                         catch_enumeration_warning_exceptions/2,
                                         real_error_occurred/0,
                                         warning_occurred/0]).
:- use_module(tools,                    [string_concatenate/3, split_atom/3,
                                         safe_atom_codes/2,ajoin/2]).
%:- use_module(kernel_tools,             [map_over_bvalue/3]). % not used ??
:- use_module(bvisual2,                 [bv_include_variables/0, bv_get_top_level/1,
                                         bv_insert_formula/3, bv_expand_formula/3,
                                         bv_formula_description/2, bv_get_values/3,
                                         bv_get_formula_functor_symbol/2, bv_formula_labels/2,
                                         bv_is_typed_predicate/1, bv_is_typed_formula/1]).
:- bv_include_variables.
:- use_module(preferences,              [eclipse_preference/2, get_preference/2,
                                         preference_description/2, preference_val_type/2,
                                         preference_default_value/2, preference_category/2,
                                         set_preference/2, advanced_eclipse_preference/2,
                                         obsolete_eclipse_preference/1]).
:- use_module(disprover,                [disprove/5, disprove_with_opts/6]).
:- use_module(solver_interface,         [solve_predicate/5,
                                         call_with_smt_mode_enabled/1]).
:- use_module(probcspsrc(haskell_csp),  [parse_and_load_cspm_file/1]).
:- use_module(state_space_reduction,    [reset_ignored_events/0, set_ignored_events/1,
                                         compute_signature_merge/0, compute_transition_diagram/1,
                                         get_reduced_node/4, reduced_trans/5,
                                         generate_node_color/5, generate_node_labels/3,
                                         generate_transition_label/3, generate_transition_color_and_style/6,
                                         write_signature_merge_to_dotfile/2]).
:- use_module(graph_canon,              [print_cstate_graph/2]).
:- use_module(debug, [debug_println/2]).
:- use_module(unsat_cores, [unsat_chr_core_with_fixed_conjuncts_auto_time_limit/4,
                            minimum_unsat_core_with_fixed_conjuncts/3]).
:- use_module(smt_solvers_interface(smt_solvers_interface), [smt_solve_predicate_in_state/5, smt_solve_predicate/4]).

:- use_module(typechecker).
:- use_module(self_check).

:- use_module(probsrc('synthesis/deep_learning/predicate_data_generator'), [generate_synthesis_data_from_predicate_untyped/5]).
:- use_module(probsrc('synthesis/deep_learning/operation_data_generator'), [generate_operation_data_from_machine_path/4,
                                                                            generate_data_from_machine_operation/6]).

:- use_module(synthesis(b_synthesis),[start_synthesis_from_ui/13,
                                      start_synthesis_single_operation_from_ui/11,
                                      reset_synthesis_context/0]).
:- use_module(synthesis(synthesis_util),[get_valid_and_invalid_equality_predicates_for_operation/6,
                                         get_valid_and_invalid_equality_predicates_for_invariants/4,
                                         get_invariant_violating_vars_from_examples/3,
                                         adapt_machine_code_for_operations/2]).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/**
Access information about the current version of the ProB core.

#### called by:
* ProB 2.0: GetVersionCommand
*/
get_version(Major,Minor,Service,Qualifier,GitRevision,LastChangedDate,PrologInfo) :-
   version(Major,Minor,Service,Qualifier),
   revision(GitRevision),  lastchangeddate(LastChangedDate), prolog_flag(version,PrologInfo).


% get the total number of errors of ProB's Prolog code; cannot be accidentally reset
% useful for validation reports,...
:- use_module(error_manager,[get_total_number_of_errors/1]).
get_prob_total_number_of_errors(X) :- get_total_number_of_errors(X).

/* ------------------------- */
/* B Loading Interface       */
/* ------------------------- */

/**
Loads a classical b model.

#### called by:
* ProB 2.0: LoadBProjectCommand
*/
:- mode load_classical_b_from_list_of_facts(+MainFilename,+ListOfFacts).
load_classical_b_from_list_of_facts(MainFilename,ListOfFacts) :-
   clear_animator,
   set_animation_mode(b),
   set_currently_opening_file(MainFilename),
   (b_load_machine_from_list_of_facts(MainFilename,ListOfFacts)
    -> set_currently_opened_file(MainFilename), % or should we use set_currently_opened_package
       update_preferences_from_spec
    ;  set_failed_to_open_file(MainFilename), fail).

/**
update preferences from SET_PREF Definitions in B machines:
should be called after loading a model and before start_animation
*/
update_preferences_from_spec :- (b_or_z_mode -> b_get_preferences_from_machine ; true).
update_preferences_from_spec(List) :- (b_or_z_mode -> b_get_preferences_from_machine(List) ; List=[]).

/* ------------------------- */
/* Event-B Loading Interface */
/* ------------------------- */

/**
Loads an Event-B model.

#### called by:
* ProB Plugin: de.prob.eventb.translator.internal.EventBTranslator, DisproverLoadCommand
* ProB 2.0: de.prob.model.eventb.translate.EventBModelTranslator (LoadEventBProjectCommand)
*/
load_event_b_project(Machines,Contexts,Proofs,Errors) :-
    clear_animator,
    set_animation_mode(b), set_animation_minor_mode(eventb),
    b_set_eventb_project_flat(Machines,Contexts,Proofs),
    load_additional_information(Proofs),
    Errors = [],
    set_currently_opened_package(event_b_project).

/* ------------------------- */
/* CSP Loading Interface     */
/* ------------------------- */

/**
Takes a path to a CSPM specification and loads the file using the CSPM parser.

#### called by:
* ProB 2.0: LoadCSPCommand
*/
:- mode load_cspm_spec_from_cspm_file(+CSPMFile).
load_cspm_spec_from_cspm_file(CSPMFile) :-
   clear_animator,
   set_animation_mode(cspm),
   set_currently_opening_file(CSPMFile),
   parse_and_load_cspm_file(CSPMFile),
   !,
   set_currently_opened_file(CSPMFile).
load_cspm_spec_from_cspm_file(CSPMFile) :- set_failed_to_open_file(CSPMFile),fail.


/* ------------------------- */
/*   Z Loading Interface     */
/* ------------------------- */

:- use_module(bmachine,[b_set_typed_machine/2]).
:- use_module(parsercall,[call_fuzz_parser/2]).
:- use_module(prozsrc(proz),[open_proz_file/2]).
load_fuzz_file_internal(FuzzFile) :-
  open_proz_file(FuzzFile,BMachine),
  b_set_typed_machine(BMachine,FuzzFile).

load_z_spec_from_fuzz_file(FuzzFile) :-
  clear_animator,
  set_animation_mode(b), set_animation_minor_mode(z),
  set_currently_opening_file(FuzzFile),
  (load_fuzz_file_internal(FuzzFile)
    -> set_currently_opened_file(FuzzFile)
    ; set_failed_to_open_file(FuzzFile),fail
  ).

load_z_spec_from_tex_file(TexFile) :-
  clear_animator,
  call_fuzz_parser(TexFile,FuzzFile),
  load_z_spec_from_fuzz_file(FuzzFile).

/* ------------------------- */
/*   XTL Loading Interface     */
/* ------------------------- */

:- use_module(xtl_interface,[open_xtl_file/1]).
load_xtl_spec_from_prolog_file(PrologFile) :-
  clear_animator,
  set_animation_mode(xtl),
  set_currently_opening_file(PrologFile),
  (open_xtl_file(PrologFile)
    -> set_currently_opened_file(PrologFile)
    ; set_failed_to_open_file(PrologFile),fail
  ).

:- use_module(probsrc('alloy2b/alloy2b'),[load_alloy_model/2]).
load_alloy_spec_from_term(PrologTerm,AlloyFile) :-
   clear_animator,
   load_alloy_model(PrologTerm,AlloyFile).
/* ------------------------- */
/* Animation                 */
/* ------------------------- */

reset_animator :- announce_event(reset_specification). % just reset; keeping same spec
clear_animator :- reset_animator,announce_event(clear_specification).

start_animation :- %reset_errors, % we no longer reset_errors here !
                   initialise_specification,!.
start_animation :- add_error(start_animation,'Start Animation Failed').
   %NOW done by events: reset_flow,    state_space_initialise, reduce_graph_reset, reset_runtime_profiler,
   %  reset_refinement_checker reset_model_checker,
   %  reset_dynamics, % new profiler

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_prob2_interface,
                           'Reset prob2_interface caches.').
reset_prob2_interface :- retractall(cached_registered_formula_result(_,_,_)).


initialise_specification :-  /* call once before starting animation or model checking */
   announce_event(start_initialising_specification),
   ( b_or_z_mode -> b_machine_precompile
   ; otherwise ->
       preferences:get_preference(symmetry_mode,X),
       ( X=off -> true
       ; otherwise ->
           add_message(initialise_specification,'Symmetry can only be used for B & Z specifications'),
           preferences:set_preference(symmetry_mode,off) /* Turn symmetry off */
       ),
       %tools:print_bt_message(other_spec_precompile),
       other_spec_precompile
   ),
      %TODO(DP,11.8.2008)
   % now done in state_space: user:initialise_operation_not_yet_covered,
   % now done via event_handline: reset_external_functions,
   preferences:get_preference(symmetry_mode,SymMode),
   precompile_marker_typing_info, % from symmetry_marker.pl
   ( SymMode=nauty ->
       (graph_canon:initialise_nauty -> true ; preferences:set_preference(symmetry_mode,off))
   ; otherwise -> graph_canon:clear_nauty),
   announce_event(specification_initialised),
   ( initialise_operation_caching -> true
   ; otherwise -> add_error_and_fail(b_load_machine,'Initialising operation caching failed')).

/* ------------------------- */
/* Handling States           */
/* ------------------------- */

/**
Produces a prolog serialization of a given state id

#### called by:
* ProB Plugin: SerializeStateCommand
* ProB 2.0: SerializeStateCommand
*/
:- mode serialize(+Id,-SerializedState).
serialize(Id,SerializedState) :-
    visited_expression(Id,State),
    write_term_to_codes(State,Codes,".",[quoted(true)]),
    atom_codes(SerializedState,Codes).

/**
Deserialize a state and add it if necessary

#### called by:
* ProB Plugin: DeserializeStateCommand
* ProB 2.0: DeserializeStateCommand
*/
:- mode deserialize(-NewId,+SerializedState).
:- use_module(state_space_exploration_modes,[get_id_of_node_and_add_if_required/4]).
deserialize(NewId,SerializedState) :-
    atom_codes(SerializedState,Codes),
    read_from_codes(Codes,State),
    get_id_of_node_and_add_if_required(State,NewId,_Res,root).

get_state(ID,BState) :- get_state(ID,BState,_,_).
get_state(es(ID),Store,full_initialised_state,-1) :- % use a store from the evalstore module instead of
    !,evalstore_bindings(ID,Store).  % a state from the state space
get_state(ID,BState,Kind,ConstID) :-
    if(visited_expression(ID,State),true,(ID=root,State=root)),
    get_state2(State,BState,Kind),
    get_constants_id(State,ID,ConstID).
get_state2(csp_and_b(CSPState,BState),[bind('CSP_CONTROLLER',CSPState)|ExpandedBState],Kind) :- !,
    get_state2(BState,ExpandedBState,Kind).
get_state2(concrete_constants(BState),BState,constants_only_state) :- !.
get_state2(root,BState,empty_state) :- !, BState=[].
get_state2(State,Res,full_initialised_state) :-
    expand_const_and_vars_to_full_store(State,BState),
    (BState=[_|_] -> Res=BState ; Res=[] /* we have [] or a CSP process or similar */).

get_constants_id(const_and_vars(ConstID,_),_,Res) :- !,Res=ConstID.
get_constants_id(concrete_constants(_),ConstID,Res) :- !,Res=ConstID.
get_constants_id(_,_,-1).

/**
Get the predicate representation of a BState

#### called by:
* ProB 2.0: GetBStateCommand
*/
get_b_state(StateId, BState) :-
    get_state(StateId, S), translate_bstate(S, BState).

/**
Compute the enabled operations (without the backtrack options) for a given state id.
Transitions is a list of operation tuples with the form op(TransitionId,SrcId,DestId)

#### called by:
* ProB 2.0: GetEnabledOperationsCommand
*/
:- mode compute_operations_for_state(+StateID,-Transitions).
compute_operations_for_state(StateID, Transitions) :-
   visited_expression_id(StateID),
   tcltk_interface:tcltk_compute_options(StateID,OpsSTAndIDs),
   (create_simple_op_terms(OpsSTAndIDs,StateID,Transitions) -> true
     ; add_error_and_fail(prob2_interface, 'Creating op terms failed', OpsSTAndIDs)).

create_simple_op_terms([],_,[]).
create_simple_op_terms([(Id,Op,Dst)|T],StateID,[op(Id,Name,StateID,Dst)|FT]) :-
    extract_op_name(Op,Name),
    create_simple_op_terms(T,StateID,FT).

set_current_state(ID) :- current_state_id(ID),!. % no need to jump
set_current_state(ID) :- /* jumps to the given node; can be backtracked */
   visited_expression_id(ID),
   tcltk_interface:tcltk_goto_state(jump,ID).

/**
Calculates an operation given a predicate from the user.

#### called by:
* ProB 2.0: GetOperationByPredicateCommand
*/
:- use_module(bmachine,[b_is_operation_name/1]).
:- mode prob2_execute_custom_operations(+CurID, +OpName, +ParsedPredicate, +MaxNrOfSolutions, -TOperations, -Errors).
prob2_execute_custom_operations(_CurID, _OpName, _Pred, MaxNrOfSolutions, TOps, Errs) :-
    MaxNrOfSolutions < 1,
    !,
    TOps = [],Errs = ['max nr of solutions too small'].
prob2_execute_custom_operations(_CurID, OpName, _Pred, _Max, TOps, Errs) :-
    \+ valid_op_name(OpName),
    !,
    TOps = [],
    (b_is_operation_name(OpName) -> ajoin(['Not a Top-Level Operation ', OpName], Msg)
       ; ajoin(['Unknown Operation ', OpName], Msg)),
    Errs = [Msg].
prob2_execute_custom_operations(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, ErrorsOut) :-
    set_current_state(CurID),
    prob2_execute_custom_operations_aux(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, ErrorsOut).

% special case for TRUE=TRUE predicate to avoid parser overhead
prob2_execute_custom_operations_aux(CurID, OpName, Pred, MaxNrOfSolutions, TOperations, ErrorsOut) :-
    raw_truth(Pred),!,
    findall(TO, prob2_execute_custom_operation(CurID,OpName,TO,MaxNrOfSolutions), TOperations),
    clear_context_state,
    (TOperations = []
     -> gen_exec_error_message(CurID,OpName,truth,Msg),
        ErrorsOut = [Msg]
     ;  ErrorsOut = []).
prob2_execute_custom_operations_aux(CurID, OpName, ParsedPredicate, MaxNrOfSolutions, TOperations, ErrorsOut) :-
    ( is_special_op(OpName,_)
      ->  Scope = [prob_ids(visible),variables,external_library(all_available_libraries)]
      ;   %b_top_level_operation(OpName),
          get_machine_operation_additional_identifiers(OpName,AdditionalIds),
          % adds parameters, results and x$0 refers to variable value of variable x before operation
          Scope = [identifier(AdditionalIds),prob_ids(visible),variables,external_library(all_available_libraries)]
          % prob scope allows one to use identifiers for deferred set elements
    ),
    b_type_expression(ParsedPredicate, Scope, _, TypedPred, Errors),
    (Errors = [] ->
        findall(TO, prob2_execute_custom_operation_with_predicate(CurID,OpName,TypedPred,TO,MaxNrOfSolutions),
                TOperations),
        retractall(bmachine:b_machine_temp_predicate(_)),
        clear_context_state,
        (TOperations = []
         -> gen_exec_error_message(CurID,OpName,TypedPred,Msg),
            ErrorsOut = [Msg]
         ;  ErrorsOut = [])
    ;   TOperations = [],
        ErrorsOut = [typechecker_errors(Errors)]
    ).

:- use_module(specfile,[get_operation_name/2]).
gen_exec_error_message(root,OpName,_,Msg) :- \+ is_special_op(OpName,_),!,
   ajoin(['Machine is not initialised, could not execute operation ', OpName], Msg).
gen_exec_error_message(CurID,OpName,_,Msg) :- is_concrete_constants_state_id(CurID),
   \+ is_special_op(OpName,_),!,
   ajoin(['Machine is not initialised, could not execute operation ', OpName], Msg).
gen_exec_error_message(CurID,'$setup_constants',_,Msg) :- is_concrete_constants_state_id(CurID),
    !,
    Msg = 'Constants are already set up, could not execute SETUP_CONSTANTS '.
gen_exec_error_message(CurID,OpName,_,Msg) :-
    CurID \= root,
    \+ is_concrete_constants_state_id(CurID),
    is_special_op(OpName,OpS), !,
    ajoin(['Machine is already initialised, cannot execute ', OpS], Msg).
gen_exec_error_message(ID,OpName,truth,Msg) :- !,
    (is_special_op(OpName,OpS) -> true ; OpS=OpName),
    ajoin(['Could not execute operation ', OpS, ' in state ', ID], Msg).
gen_exec_error_message(ID,OpName,_,Msg) :- (is_special_op(OpName,OpS) -> true ; OpS=OpName),
    (transition(ID,Op,_,_), get_operation_name(Op,OpName)
      -> M2 = ' (but a transition for operation exists)' ; M2 = ''),
    ajoin(['Could not execute operation ', OpS, ' in state ', ID, ' with additional predicate',M2], Msg).
is_special_op('$setup_constants','SETUP_CONSTANTS').
is_special_op('$initialise_machine','INITIALISATION').

% true if a raw expression definitely represents truth
raw_truth(equal(_,E1,E2)) :- raw_equals(E1,E2).
raw_equals(boolean_true(_),boolean_true(_)).
raw_equals(boolean_false(_),boolean_false(_)).
raw_equals(integer(_,I),integer(_,I)).

valid_op_name('$initialise_machine').
valid_op_name('$setup_constants').
valid_op_name(OpName) :- b_top_level_operation(OpName).
%valid_op_name(OpName) :- b_get_promoted_machine_operations(OpName).

/**

Calculates a trace given a list of operation names and a list of guards.
In case of errors, a partial trace is generated that jumps over the errornous operation / event.
In addition, a list of integers should be given as argument that make possible to execute some
operations in the oprations list multiple times or as long as they are disabled (in this case one should give -1).
E.g. by calling prob2_construct_trace(0,[e1,e2,e3],[TRUE,TRUE,TRUE],[2,1,-1],OpsOut,ErrOut) the predicate calculates
a trace starting at the state with the ID 0 where e1 is executed sequentially 2 times, after that e2 once,
and finally e3 until it becomes disabled.

#### called by:
* ProB 2.0: ConstructTraceCommand
*/

prob2_construct_trace(CurId,Names,Preds,Nrs,OpsOut,ErrOut) :-
    construct_trace(CurId,Names,Preds,Nrs,[],OpsOut,[],ErrOut).

construct_trace(_CurId,[],[],[],OpsOut,OpsOut,ErrOut,ErrOut).
construct_trace(CurId,[Name|Names],[Pred|Preds],[NrName|NrNames],CurOpsOut,OpsOut,CurErrOut,ErrOut) :-
    execute_nr_of_custom_operations(CurId,Name,Pred,NrName,NewId,ListOfOps,ListOfErrs),
    append(CurOpsOut,ListOfOps,NewOpsOut),
    append(CurErrOut,ListOfErrs,NewErrOut),
    construct_trace(NewId,Names,Preds,NrNames,NewOpsOut,OpsOut,NewErrOut,ErrOut).

execute_nr_of_custom_operations(CurId,_Name,_Pred,0,CurId,[],[]).
execute_nr_of_custom_operations(CurId,Name,Pred,Nr,NewId,ListOfOps,ListOfErrs) :-
    ( Nr = -1 ->
        execute_until_disabled(CurId,Name,Pred,NewId,ListOfOps,ListOfErrs)
    ; otherwise -> % Nr is greater than 0
        prob2_execute_custom_operations(CurId,Name,Pred,1,ListContainingSingleOp,SingleErr),
        ( ListContainingSingleOp = [] ->
            NextId=CurId
        ;   ListContainingSingleOp = [op(_Trans,_Name,_From,To)],
            NextId = To
        ),
        append(ListContainingSingleOp,Ops,ListOfOps),
        append(SingleErr,Errs,ListOfErrs),
        NewNr is Nr - 1,
        execute_nr_of_custom_operations(NextId,Name,Pred,NewNr,NewId,Ops,Errs)
    ).


execute_until_disabled(CurId,OpName,Pred,NewId,ListOfOps,ListOfErrs) :-
    execute_until_disabled(CurId,OpName,Pred,NewId,[],ListOfOps,ListOfErrs).

execute_until_disabled(CurId,OpName,Pred,NewId,CurListOfOps,ListOfOps,ListOfErrs) :-
    prob2_execute_custom_operations(CurId,OpName,Pred,1,ListContainingSingleOp,SingleErr),
    ( ListContainingSingleOp = [] -> % reached a state where OpName is not executed
        NewId=CurId,
        ListOfOps=CurListOfOps,
        ajoin(['Could not execute Operation ', OpName, ' with additional predicate'], Msg),
        (member(Msg,SingleErr) -> % Error message generated because OpName is disabled at some state
            ListOfErrs = []
        ; otherwise -> % unexpected error occurred, need to be reported
            ListOfErrs = SingleErr
        )
    ;   ListContainingSingleOp = [op(Trans,Name,From,To)],
        ( member(op(_T,_N,To,_To),CurListOfOps) -> % in case of a loop we terminate, otherwise we will go forever
            ListOfOps = CurListOfOps,
            NewId=From
        ; otherwise ->
            NextId = To,
            append(CurListOfOps,[op(Trans,Name,From,To)],NewCurListOfOps),
            execute_until_disabled(NextId,OpName,Pred,NewId,NewCurListOfOps,ListOfOps,ListOfErrs)
        )
    ).

% TO DO: check if we need to recompute the operation effect: if the ParsedPredicate is TRUE
%  (TRUE = equal(none,integer(none,1),integer(none,1)) ) & MaxNrOfSolutions <=
%        what has already been used previously, we can simply reuse transition from the state space

prob2_execute_custom_operation(CurID,OpName,Transition,Max) :-
    statistics(walltime, [CurTime1,_]),
    set_context_state(CurID),
    visited_expression(CurID,InState),
    %add_prob_deferred_set_elements_to_store(InState1, InState, visible), % does not work for root, const_and_vars
    specfile:compute_operation_effect_max(InState,OpName,Operation,NewState,_TransPathInfo,Max),
    % logger:writeln_log(sol(OpName,NewState)), %%
    tcltk_interface:add_trans_id(CurID,Operation,NewState,[],NewID,[],TransId),
    % logger:writeln_log(ids(OpName,Operation,NewID,TransId)), %%
    statistics(walltime, [CurTime2,_]),
    log_executed_operation(CurID,Operation,NewID,CurTime1,CurTime2), % TODO? : log errors as comments
    Transition = op(TransId,OpName,CurID,NewID).

:- use_module(succeed_max,[succeed_max_call_id/3]).
:- use_module(b_state_model_check,[execute_operation_by_predicate_in_state/5]).
:- use_module(b_global_sets,[inline_prob_deferred_set_elements_into_bexpr/2]).
:- use_module(bmachine,[assert_temp_typed_predicate/1,reset_temp_predicate/0]).
prob2_execute_custom_operation_with_predicate(CurID,OpName,Predicate,Transition,Max) :-
    on_exception(enumeration_warning(_,_,_,_,_),
      prob2_execute_custom_operation_with_predicate_aux(CurID,OpName,Predicate,Transition,Max),
      (add_warning(prob2_execute_custom_operation_with_predicate,'Enumeration warning when executing operation with predicate: ',OpName),
       fail)).
prob2_execute_custom_operation_with_predicate_aux(CurID,OpName,Predicate,Transition,Max) :-
    b_top_level_operation(OpName),
    !,
    statistics(walltime, [CurTime1,_]),
    set_context_state(CurID),
    visited_expression(CurID,InState),
    succeed_max_call_id(prob2_execute_custom_operation,
                 execute_operation_by_predicate_in_state(InState,OpName,Predicate,Operation,NewState),Max),
    tcltk_interface:add_trans_id(CurID,Operation,NewState,[],NewID,[],TransId),
    statistics(walltime, [CurTime2,_]),
    log_executed_operation(CurID,Operation,NewID,CurTime1,CurTime2),
    Transition = op(TransId,OpName,CurID,NewID).
prob2_execute_custom_operation_with_predicate_aux(CurID,OpName,Predicate,Transition,MaxNrOfSolutions) :-
        % temporary fix: old style execution for INIT, setup_constants; INIT does not work yet for Event-B !
        inline_prob_deferred_set_elements_into_bexpr(Predicate,CP),
        %translate:print_bexpr(CP),nl,
        assert_temp_typed_predicate(CP),
        call_cleanup(prob2_execute_custom_operation(CurID,OpName,Transition,MaxNrOfSolutions),
                     reset_temp_predicate).

:- dynamic last_trace_id/2. % if this fact is set, we assume we will log to a trace file

:- use_module(preferences,[get_prob_application_type/1]).
:- use_module(bmachine,[b_absolute_file_name_relative_to_main_machine/2]).
:- register_event_listener(start_initialising_specification,reset_last_trace_id,
                           'Reset for logging.').
reset_last_trace_id :- retractall(last_trace_id(_,_)),
   (get_preference(prob2_trace_file,File),File \= '',
    get_prob_application_type(Type), Type \= tcltk, % logging not working, would overwrite Trace file
    b_absolute_file_name_relative_to_main_machine(File,AFile)
     -> reset_trace_file(AFile,WritableFile),
        assert(last_trace_id(root,WritableFile)) ; true).

% log for later replay
log_executed_operation(CurID,Operation,NewID,CurTime1,CurTime2) :-
  retract(last_trace_id(LastID,File)), !,
   (LastID=CurID -> true ; add_jump(LastID,CurID,File)),
   Delta is CurTime2 - CurTime1,
   detailed_print_to_trace_file(File,Operation,CurID,NewID,Delta),
   % TO DO: print some timing info,... in comments
   assert(last_trace_id(NewID,File)).
log_executed_operation(_,_,_,_,_).

:- use_module(state_space,[is_concrete_constants_state_id/1]).
add_jump(X,X,_) :- !.
add_jump(root,X,File) :- is_concrete_constants_state_id(X), % in case we missed a setup_constants
   !, print_to_trace_file(File,'$setup_constants').
add_jump(X,Y,File) :- nl, nl, print(add_jump(X,Y,File)),nl,
   (X=root ; is_concrete_constants_state_id(X)), % in case we missed an initialisation
   !,
   print_to_trace_file(File,'$initialise_machine').
add_jump(LastID,CurID,File) :- print_to_trace_file(File,'$jump'(LastID,CurID)).

:- use_module(tools_printing,[print_time_stamp/1]).
reset_trace_file(File,WritableFile) :-
   (get_preference(prob2_trace_file_gen_unique_name,true)
     -> open(File,write,Stream,[if_exists(generate_unique_name)]),
        stream_property(Stream,file_name(WritableFile))
     ;  open(File,write,Stream), WritableFile=File
    ),
   write(Stream,'% Created: '), print_time_stamp(Stream), nl(Stream),
   revision(GitRevision),
   format(Stream,'% ProB GIT revision: ~w~n',[GitRevision]),
   close(Stream).
print_to_trace_file(File,Term) :-
   open(File,append,Stream),
   portray_clause(Stream,Term),
   close(Stream).
detailed_print_to_trace_file(File,Term,CurID,NewID,Delta) :-
   open(File,append,Stream),
   format(Stream,'~n% Executed ~w -> ~w in ~w ms  at  ',[CurID,NewID,Delta]), print_time_stamp(Stream), nl(Stream),
   get_state_space_stats(NrNodes, NrTrans, NrProcessed),
   statistics(memory_used,V), VMb is V / 1000000, % used instead of deprecated 1048576
   format(Stream,'% State space: nodes = ~w, transitions = ~w, processed = ~w, MB used: ~3f~n',[NrNodes,NrTrans,NrProcessed,VMb]),
   portray_clause(Stream,Term), % TO DO: also export in JSON format; drawback is we need to add closing brackets at the end, whereas in Prolog format we always have a valid trace file
   close(Stream).


/**
Extracts information about the parameters and return values for the
specified transition.

#### called by:
* ProB 2.0: GetOpFromId
*/
:- mode get_op_from_id(+Id,+Truncate,-Params,-RetVals).
get_op_from_id(Id,Truncate,Params,RetVals) :-
    transition(_Src,Op,Id,_Dest),
    create_op_tuple(Op,Truncate,Params,RetVals).

/**
Creates an operation tuple from transition id, source id, an op term, and a destination id.
If creation is unsuccessful, an error is added and the predicate fails.
See comment for extract_op_tuple for a description of OpTuple
*/
:- mode create_op_tuple(+OpTerm, +Truncate, -Params, -RetVals).
create_op_tuple(OpTerm, Truncate, Params, RetVals) :-
    (extract_op_tuple(OpTerm, Truncate, Params, RetVals) -> true;
        add_error_and_fail(prob2_interface, 'Could not create OpTuple ', extract_op_tuple(OpTerm, Params, RetVals))).

/**
Extracts the parameters and return values of the operations
*/
:- mode extract_op_tuple(+OpTerm, +Truncate, -Params, -RetVals).
extract_op_tuple(OpTerm, Truncate, Params, RetVals) :-
    (animation_mode(cspm) ->
        extract_csp_op(OpTerm, Params, RetVals)
    ;   extract_b_op(OpTerm, Truncate, Params, RetVals)
    ).

/**
Extracts information for a CSP operation.
*/
:- mode extract_csp_op(+OpTerm, -Params, -RetVals).
extract_csp_op(OpTerm, Params, []) :-
    %OpTuple = op(TransId, Src, Dest, Name, params(Source, Pretty), return([],[]), TargetState), /* TODO: CSP operations do not have return types, is this correct??? */
    %extract_csp_name_and_args(OpTerm, _Name, _Source),
    translate_event(OpTerm, PPEvent),
    split_atom(PPEvent,['.','!'],[_Op|Params]).

:- mode extract_csp_name_and_args(+OpTerm, -Name, -Args).
extract_csp_name_and_args(io(Args,ChName,_SPAN),ChName,Args).
extract_csp_name_and_args(start_cspm(Name), Name, []).
extract_csp_name_and_args(start_cspm_MAIN,'start_cspm_MAIN',[]).
extract_csp_name_and_args(tick(_),tick,[]).
extract_csp_name_and_args(tau(_),tau,[]).
extract_csp_name_and_args(_OP,'?',[]).

/**
Extracts information for a B operation.
*/
:- mode extract_b_op(+OpTerm, +Truncate, -Params, -RetVals).
extract_b_op(OpTerm, Truncate, Params, RetVals) :-
    extract_b_op_infos(OpTerm, _Name, PSource, RSource),
    maplist(pretty_print_bvalue(Truncate), PSource, Params),
    maplist(pretty_print_bvalue(Truncate), RSource, RetVals).

:- mode extract_b_op_infos(+Term, -Name, -Arguments, -RetVals).
extract_b_op_infos(Term, Name, Arguments, RetVals) :-
    (Term = '-->'(OpTerm,RetVals) -> true;
        OpTerm = Term, RetVals = []),
    OpTerm =.. [Name|Arguments].

extract_op_name(OpTerm,Name) :-
    (animation_mode(cspm) ->
        extract_csp_name_and_args(OpTerm, Name, _Args)
    ;   extract_b_op_infos(OpTerm, Name, _Params, _RetVals)
    ).

pretty_print_bvalue_unicode(Truncate,Formula,Result) :-
    set_unicode_mode,
    call_cleanup(pretty_print_bvalue(Truncate,Formula,Result),unset_unicode_mode).
pretty_print_bvalue(truncate,Formula,Result) :-
    translate_bvalue_with_limit(Formula,600,Result).
    %translate_bvalue(Formula, Result).
pretty_print_bvalue(expand,Formula,Result) :-
    translate_bvalue_with_limit(Formula,-1,Result).

/* ------------------------- */
/* Boolean Properties        */
/* ------------------------- */

:- use_module(state_space,[visited_state_corresponds_to_setup_constants_b_machine/1]).
is_constants_set_up_state(ID) :-
  b_or_z_mode
    -> visited_state_corresponds_to_setup_constants_b_machine(ID)
    ; true.

is_initialised_state(ID) :- b_or_z_mode,!,
   is_initialised_b_state(ID).
is_initialised_state(ID) :-
    visited_expression_id(ID), ID \= root.

:- use_module(state_space,[visited_state_corresponds_to_initialised_b_machine/1]).
is_initialised_b_state(ID) :-
    visited_state_corresponds_to_initialised_b_machine(ID).


invariantKO(StateID) :-  is_initialised_state(StateID),
    get_state(StateID,State),
    tcltk_interface:check_invariantKO(StateID,State),
    state_space:invariant_violated(StateID).

/**
Takes a state id and finds all of the operations for which a timeout occurred

#### called by:
* ProB Plugin: GetTimeoutedOperationsCommand
* ProB 2.0: GetOperationsWithTimeout
*/
op_timeout_occurred(StateID,OpNameList) :-
    findall(OpName,time_out_for_node(StateID,OpName,_Type),OpNameList).

inv_timeout_occurred(StateID) :- time_out_for_invariant(StateID).
timeout_occurred(StateID) :-  time_out_for_node(StateID).
max_operations_reached(StateID) :- max_reached_for_node(StateID).

/**
Finds the status for a given property
Properties can be: invariantKO, timeout_occurred, max_operations_reached, initialised
Statuses are expected to be boolean values: either true or false

#### called by:
* ProB Plugin: CheckBooleanPropertyCommand
* ProB 2.0: CheckBooleanPropertyCommand
*/
:- mode state_property(+Property,+StateId,-Status).
state_property(Property,StateId,Status) :-
    ( state_property2(Property,StateId) -> Status = true
    ; otherwise -> Status = false).
state_property2(invariantKO,StateId) :- invariantKO(StateId).
state_property2(timeout_occurred,StateId) :- inv_timeout_occurred(StateId) ; timeout_occurred(StateId).
state_property2(max_operations_reached,StateId) :- max_operations_reached(StateId).
state_property2(valid_state,StateId) :- visited_expression_id(StateId).
state_property2(constants_set_up,StateId) :- is_constants_set_up_state(StateId).
state_property2(initialised,StateId) :-
    (specfile:csp_mode ->
        true % in CSP the content of the state is irrelevant for the semantic of the state space
      ; is_initialised_state(StateId)
    ).


/**
Takes a id for a given state in the state space and produces a list of all
state based errors.

#### called by:
* ProB Plugin: GetStateBasedErrorsCommand
* ProB 2.0:    GetStateBasedErrorsCommand
*/
:- mode get_state_errors(+StateId,-Errors).
get_state_errors(StateId,Errors) :-
  findall(E, (state_error(StateId,_,E),E \== invariant_violated), Errs),
  convert_errors(Errs,Errors).
convert_errors([],[]).
convert_errors([Error|InRest],[error(Event,Short,Long)|OutRest]) :-
  ( Error = eventerror(Event,EError,Trace) ->
      translate_event_error(EError,Short),
      explain_event_trace(Trace,LongStr),
      safe_atom_codes(Long,LongStr)
  ; otherwise ->
      Event = '*unknown*',
      translate_state_error(Error,Short),
      explain_state_error(Error,_Span,LongStr),
      safe_atom_codes(Long,LongStr)),
  convert_errors(InRest,OutRest).

/* ------------------------- */
/* Evaluate Formulas         */
/* ------------------------- */

% term in string (see String argument) should be terminated by full-stop!!!
evaluate_csp_expression_string(PlClause,R) :-
    read_from_codes(PlClause,PlTerm),
    evaluate_csp_expression_aux(PlTerm,R).

evaluate_csp_expression_aux(PlTerm,R) :-
    haskell_csp:evaluate_expression(PlTerm,Res),
    translate:translate_cspm_expression(Res, R).
evaluate_csp_expression_aux(Other,_R) :-
    add_error_and_fail(prob2_interface, 'Unexpected CSP Expression: ', Other).

/**
Evaluate a formula given a state id and the raw prolog representation of the formula

#### Params:
* StateId - state for which the formula should be evaluated
* Element - Triple in the form eval(Formula,Type,Code)
    * Formula - Prolog representation of non-typechecked formula
    * Type - csp,'#PREDICATE',or '#EXPRESSION'
    * Code - String representation of Formula
    * Truncate - if the result should be truncated or expanded
* Result - Triple result(Value,Solutions,Code)
    * Value - String rep of the value calculated by ProB or tuple: v(SRes,Term)
        * SRes - String rep of the value calculated by ProB
        * Term - Prolog representation of the value
    * Solutions - List of solutions as a triple: bind(Name,Solution,PPSol)
        * Name - String free variable from formula
        * Solution - Prolog representation of the solution bound to the name
        * PPSol - String representation of the solution bound to the name
    * Code - String representation of evaluated formula
           - errors(Res,ListOfErrors)
#### called by:
ProB 2.0: EvaluateFormulaCommand
*/

:- mode evaluate_formula(+StateId,+Element,-R).
evaluate_formula(StateId,Element,R) :-
    setCurrentState(StateId),
    get_bstate_with_deferred_sets(StateId,State,Kind,_),
    evaluate_formula_in_state(Kind,State,StateId,Element,R).


:- assert_must_succeed((X=identifier(unknown,x),
      evaluate_formulas(root,
      [eval(comprehension_set(unknown,[X],greater(unknown,mult_or_cart(unknown,X,X),integer(unknown,2))),
      '#EXPRESSION','{x|x * x > 2}',truncate)],Res),
      Res = [result(Str,[])],
      (Str == '/*@symbolic*/ {x|x * x > 2}' ; Str == '{x|x * x > 2}'))).

% a list version of the above: advantage: state expanded only once
:- mode evaluate_formulas(+StateId,+Elements,-Results).
evaluate_formulas(StateId,Elements,Rs) :-
    setCurrentState(StateId),
    get_bstate_with_deferred_sets(StateId,State,Kind,_ConstID),
    maplist(inline_machine(_Machine),Elements,Elements2),
    maplist(evaluate_formula_in_state(Kind,State,StateId),Elements2,Rs).

% useful for certain external functions in LibraryMeta.def, e.g., ENABLED
setCurrentState(ID) :- current_state_id(ID),!. % no need to jump
setCurrentState(ID) :- /* jumps to the given node; can be backtracked */
   visited_expression_id(ID),!,
   tcltk_interface:set_id_as_current_state(ID). % what should we do with history, trace, ...
setCurrentState(ID) :- add_error(prob2_interface,'Unknown state id:',ID).

% detect all expressions and insert machine term variable to avoid generating term multiple times
    % TO DO: maybe also inline Typechecker's type environment
inline_machine(Machine,eval(Raw,Type,Code,Truncate),eval(Raw,for_machine(Machine,Type),Code,Truncate)) :-
    Type \= csp, !.
inline_machine(_,E,E).

%:- use_module(runtime_profiler,[profile_single_call/3]).
:- meta_predicate myprofile(-,-,0).
myprofile(_,_,Call) :- call(Call).
%myprofile(Source,StateId,Call) :- profile_single_call(Source,StateId,Call).
evaluate_formula_in_state(_Kind, _State,_StateId, eval(Raw,csp,_Code,_Truncate), R) :- !,
    ( evaluate_csp_expression_string(Raw,Res) ->
        R = result(Res,[]);
        R = errors('UNKNOWN',['Unexpected CSP Expression'])).
evaluate_formula_in_state(StateKind, State,StateId, eval(Raw,PredOrExpr,_Code,Truncate), R) :- !,
   %logger:writeln_log_time(typechecking_raw(PredOrExpr)),
    myprofile(Raw,StateId,
         prob2_interface:formula_typecheck2_for_eval(PredOrExpr,Raw,Typed,Type,Errors)),
   %logger:writeln_log_time(evaluate_typechecked_b_formula_in_state(StateKind)),
    ( Errors=[] ->
        %print('EVAL:'),nl,translate:print_bexpr(Typed),nl,
        myprofile(prob2_evaluate_formula,StateId,
         prob2_interface:evaluate_typechecked_b_formula_in_state(StateKind,formula(Typed),State,Typed,Type,Truncate,R))
    ; R = errors('ERROR',ErrMsgs), % prob2_kernel does not process error terms here, it wants atoms
      maplist(get_err_msg,Errors,ErrMsgs)).
evaluate_formula_in_state(StateKind, State,StateId, E, R) :-
   add_internal_error('Illegal evaluation term:',evaluate_formula_in_state(StateKind, State,StateId, E, R)),
   R = errors('UNKNOWN',['Illegal evaluation term']).


get_err_msg(error(ErrMsg,_ErrType,_ErrLocations),ErrMsg) :- !.
get_err_msg(X,X).

get_bstate_with_deferred_sets(StateId,State,Kind,ConstID) :-
    get_state(StateId, State1,Kind,ConstID), % TO DO: we could try and get only those identifiers that are really used
    add_prob_deferred_set_elements_to_store(State1, State, visible).

/**
Type checks a given formula.

#### Params:
* PredOrExpr - '#PREDICATE' or '#EXPRESSION'
* RawFormula - Prolog representation of non-typechecked formula
* TypedFormula - Prolog representation of typechecked formula
* Type - type of formula
* Errors - any errors that have occured during typechecking

#### called by:
ProB 2.0: EvaluateFormulaCommand
*/
% can be expensive for big machines, as typechecking calls full_b_machine (20-30 ms for ZC scheduler)

:- mode evaluate_formula_typecheck(+PredOrExpr,+RawFormula,-TypedFormula,-Type,-Errors).
evaluate_formula_typecheck(PredOrExpr,RawFormula,TypedFormula,PPType,Errors) :-
   formula_typecheck2_for_eval(PredOrExpr,RawFormula,TypedFormula,Type,Errors),
   translate:pretty_type(Type,PPType).
evaluate_formula_typecheck(PredOrExpr,RawFormula,TypedFormula,Errors) :-
   formula_typecheck2_for_eval(PredOrExpr,RawFormula,TypedFormula,_Type,Errors).

:- use_module(bmachine,[b_is_variable/2, b_is_constant/2, b_type_expression_for_full_b_machine/6]).
% this predicate returns the internal Prolog type:
:- mode formula_typecheck2_for_eval(+PredOrExpr,+RawFormula,-TypedFormula,-Type,-Errors).
formula_typecheck2_for_eval('#PREDICATE',Raw,Typed,Type,Errors) :-
    !, get_eval_scope(Scope),
    b_type_open_predicate(open(exists),Raw,Scope,Typed,PErrors),
    convert_perrors(PErrors, Errors),
    Type=pred.
formula_typecheck2_for_eval(for_machine(M,'#PREDICATE'),Raw,Typed,Type,Errors) :-
    !,get_eval_scope(Scope),
    bmachine:b_type_open_predicate_for_full_b_machine(M,open(exists),Raw,Scope,Typed,PErrors),
    convert_perrors(PErrors, Errors),
    Type=pred.
formula_typecheck2_for_eval('#EXPRESSION',identifier(Pos,ID),Typed,Type,Errors) :-
    (b_is_variable(ID,Type) ; b_is_constant(ID,Type)),
    !,  % hack to avoid type-checking simple identifiers which just need to be looked up in state
    Errors=[], Typed = b(identifier(ID),Type,[nodeid(Pos)]).
formula_typecheck2_for_eval('#EXPRESSION',Raw,Typed,Type,Errors) :-
    !,get_eval_scope(Scope),
    b_type_expression(Raw,Scope,Type,Typed,PErrors),
    convert_perrors(PErrors, Errors).
formula_typecheck2_for_eval(for_machine(_M,'#EXPRESSION'),identifier(Pos,ID),Typed,Type,Errors) :-
    (b_is_variable(ID,Type) ; b_is_constant(ID,Type)),
    !,  % hack to avoid type-checking simple identifiers which just need to be looked up in state
    Errors=[], Typed = b(identifier(ID),Type,[nodeid(Pos)]).
formula_typecheck2_for_eval(for_machine(M,'#EXPRESSION'),Raw,Typed,Type,Errors) :-
    !,get_eval_scope(Scope),
    b_type_expression_for_full_b_machine(M,Raw,Scope,Type,Typed,PErrors),
    convert_perrors(PErrors, Errors).
formula_typecheck2_for_eval(Else,_Raw,_Typed,_Type,_Errors) :-
    add_error_and_fail(prob2_interface,'expected #PREDICATE or #EXPRESSION instead of:', Else),fail.

% get scope for typechecker:
get_eval_scope([prob_ids(visible),Scope,
                external_library(all_available_libraries)]) :-
     % external_library for VisB and Debugging; TODO: should we only make it available in some contexts
     % Note: the external_library entries must come after Scope
     get_main_eval_scope(Scope).
get_main_eval_scope(assertions_scope) :- get_preference(allow_operation_calls_in_expr,true),!.
get_main_eval_scope(variables).

% FIXME This is a bit hacky, it would be better to implement this properly in error_manager.
convert_perrors(PErrors, Errors) :-
    add_all_perrors(PErrors),
    get_all_errors_with_span_info_and_reset(Errors).

:- use_module(tools_meta,[safe_time_out/3]).
evaluate_typechecked_b_formula_in_state(StateKind,Requirements,State,Typed,Type,Truncate,Result) :-
      (requirements_met(Requirements,StateKind) ->
           evaluate_typechecked_b_formula_in_state(State,Typed,Type,Truncate,Result)
       ;  Result = errors('NOT-INITIALISED',[])
     ).
evaluate_typechecked_b_formula_in_state(State,Typed,Type,Truncate,Result) :-
    % should we use get_computed_preference(debug_time_out,DTO) instead of 1000 ?
    % there seem to be two uses in ProB2: for state view, where we want a short time out, and LISB or maybe REPL where we want a longer time out or no time out
    (Truncate=truncate -> TO = 1000, FORCE=limit(10000) ; TO = 30000, FORCE=force),
    safe_time_out(evaluate_formula_eval(Type,State,Typed,FORCE,Res,Solution,EnumWarning),TO,TimeOutRes),
    (TimeOutRes=time_out -> Res=time_out, EnumWarning=false ; true),
    get_error_messages(_WarningsOnly,Errors), % TO DO: treat warnings_only !
    extract_result(EnumWarning,Errors,Res,Solution,Truncate,Result).

extract_result(true,_,_,_,_,R) :- !, R=enum_warning.
extract_result(_EnumW,[],v(_SRes,Term),[],Truncate,Result) :- !,  % v/2 for value
    pretty_print_bvalue_unicode(Truncate,Term,R),
    Result = result(R,[]).
extract_result(_EnumW,[],SRes,Solutions,Truncate,Result) :- !,
    prettyprint_solutions(Truncate,Solutions,PPSol),
    Result = result(SRes,PPSol).
extract_result(_EnumW,Errors,v(SRes,_),_Sol,_Truncate,Result) :- !,
    Result = errors(SRes, Errors).
extract_result(_EnumW,Errors,SRes,_Sol,_Truncate,Result) :-
    Result = errors(SRes, Errors).

%:- use_module(eval_strings,[eval_predicate/5]).  % TODO: export and refactor
% TO DO: move this into another more general module:
evaluate_formula_eval(pred,State,Typed,_,Res,LocalState,EnumWarning) :-
    !,catch_enumeration_warning_exceptions(
         (eval_predicate(State,Typed,Res,EnumWarning,LocalState)->true),
         (Res = time_out,EnumWarning = true)).
evaluate_formula_eval(_Type,State,Typed,FORCE,Res,[],EnumWarning) :-
    catch_enumeration_warning_exceptions(
        (evaluate_expression(State,Typed,FORCE,SRes,Term)->EnumWarning=false),
        (SRes = time_out, EnumWarning = true)),
    Res = v(SRes,Term).

:- use_module(clpfd_interface,[catch_clpfd_overflow_call2/2]).
:- use_module(error_manager,[enter_new_error_scope/2, exit_error_scope/3,clear_all_errors_in_error_scope/1,
   event_occurred_in_error_scope/1, abort_error_occured_in_error_scope/0]).
:- use_module(store,[normalise_value_for_var/4]).
% a simplified version of eval_strings:eval_expression
% evaluate an expression Typed in an expanded state EState giving string result Result and Prolog value NValue
% FORCE = force or e.g. limit(10000)
evaluate_expression(EState,Typed,FORCE,Result,NValue) :-
        enter_new_error_scope(ScopeID,evaluate_expression),
        clear_all_errors_in_error_scope(ScopeID),
        %replace_expression_by_kodkod_if_enabled(Typed,Typed2),
        catch_clpfd_overflow_call2(b_interpreter:b_compute_expression_nowf(Typed,[],EState,Value),fail), % We could return CLPFD overflow error result
         !,
   %logger:writeln_log_time(normalise_value_for_var(evaluate_expression)),
        normalise_value_for_var(evaluate_expression,FORCE,Value,NValue),
        get_unknown_error_result(ErrRes),
        exit_error_scope(ScopeID,ErrOcc,evaluate_expression),
        (ErrOcc=true
            -> Result = ErrRes
            ;  translate:translate_bvalue_for_expression(NValue,Typed,Result)
        ).
evaluate_expression(_,_,_,ErrRes,error) :-
        get_unknown_error_result(ErrRes),
        exit_error_scope(_ScopeID,_,evaluate_expression).

% which result should be displayed in case of failure/errors:
get_unknown_error_result(ErrRes) :-
        (abort_error_occured_in_error_scope -> ErrRes = 'NOT-WELL-DEFINED' ; ErrRes = 'UNKNOWN').

% a simplified version of eval_strings:eval_predicate_aux
eval_predicate(State, ExTyped,Result,EnumWarning,LocalState) :-
        enter_new_error_scope(ScopeID,eval_predicate), clear_all_errors_in_error_scope(ScopeID),
        (catch_clpfd_overflow_call2(prob2_interface:eval_predicate2(State, ExTyped,LocalState,Res),
                                    fail)
          -> EnumWarning=false
          ;  event_occurred_in_error_scope(enumeration_warning(_,_,_,_,_Critical))
              -> Res= 'UNKNOWN', EnumWarning=true
              ;  eval_predicate2_fail_result(ExTyped,Res),  % either TRUE or FALSE
                 EnumWarning=false
        ),
        get_unknown_error_result(ErrRes),
        exit_error_scope(ScopeID,ErrOcc,eval_predicate),
        (ErrOcc=true
            -> Result = ErrRes
            ;  Result = Res
        ).

:- use_module(probsrc(bsyntaxtree), [safe_create_texpr/4]).
eval_predicate2(State, ExTyped,LocalState,'TRUE') :-
        eval_strings:is_existential_quantifier(ExTyped,Parameters,Typed), % TO DO: move to module
        !,
        test_bool_exists(State, Parameters,Typed,LocalState).
eval_predicate2(State, ExTyped,LocalState,'FALSE') :-
        is_universal_quantifier(ExTyped,Parameters,TypedLHS,TypedRHS),
        !,
        safe_create_texpr(negation(TypedRHS),pred,[try_smt],NegRHS),
        conjunct_predicates([TypedLHS,NegRHS],Conjunction), % replace Kodkod?
        test_bool_exists(State, Parameters,Conjunction,LocalState).
eval_predicate2(State, Typed,LocalState,'TRUE') :- LocalState=[],
        b_interpreter:b_test_boolean_expression_wf(Typed,LocalState,State).

% what should the result be if the above eval_predicate2 fails (without enumeration warning):
eval_predicate2_fail_result(ExTyped,Res) :-
        is_universal_quantifier(ExTyped,_,_,_),!, Res='TRUE'.
eval_predicate2_fail_result(_,'FALSE').


test_bool_exists(EState, Parameters,Typed,LocalState) :-
        kernel_waitflags:init_wait_flags(WF,[test_bool_exists]),
        b_interpreter:set_up_typed_localstate(Parameters,_FreshOutputVars,TypedVals,[],LocalState,positive),
        b_enumerate:b_tighter_enumerate_values(TypedVals,WF),
        b_interpreter:b_test_boolean_expression(Typed,LocalState,EState,WF),
        kernel_waitflags:ground_wait_flags(WF).

is_universal_quantifier(b(forall(Parameters,LHS,RHS),pred,_),Parameters,LHS,RHS).

:- dynamic prob2_formula/4.
:- dynamic inactive_prob2_formula/1.

/**
Type checks a specified formula and asserts so that it can be found later
and evaluated later.

#### called by:
* ProB 2.0: RegisterFormulaCommand
*/

:- mode register_prob2_formula(+FormulaUUID, +Formula).
register_prob2_formula(FormulaUUID, _) :- prob2_formula(FormulaUUID,_,_,_),!.
register_prob2_formula(FormulaUUID, eval(Raw,PredOrExpr,Truncate)) :-
    evaluate_formula_typecheck(PredOrExpr,Raw,Typed,Errors),
    ( Errors=[] ->
        register_prob2_formula2(FormulaUUID, Typed, Truncate)
    ; otherwise ->
        add_error_and_fail(register_prob2_formula, 'Type-Errors: ', Errors)
    ).

:- use_module(bmachine, [determine_type_of_formula/2]).
register_prob2_formula2(FUUID, Typed, Truncate) :-
    determine_type_of_formula(Typed,Requirements),
    assert(prob2_formula(FUUID, Typed, Truncate,Requirements)).

:- mode register_prob2_formulas(+FormulaUUIDs, +Formulas).
register_prob2_formulas(FormulaUUIDs,Formulas) :-
    maplist(inline_machine(_Machine),FormulaUUIDs,FormulaUUIDs2),
    maplist(register_prob2_formula,FormulaUUIDs2,Formulas).


% TO DO: determine upon registering whether a formula reads nothing, just constants, or also variables

% unregister a single or a list of formula ids:
unregister_prob2_formula(FormulaUUID) :- retractall(prob2_formula(FormulaUUID,_,_,_)).
unregister_prob2_formulas(Fs) :- maplist(unregister_prob2_formula,Fs).

% mark certain formulas as inactive; will not be computed by evaluate_registered_formula
% useful when certain information currently not needed (e.g., in state view the formula is not visible/shown)
deactivate_prob2_formulas(Fs) :- maplist(deactivate_prob2_formula,Fs).
deactivate_prob2_formula(FormulaUUID) :- prob2_formula(FormulaUUID,_,_,_),!,
   (inactive_prob2_formula(FormulaUUID) -> print(formula_already_inactive(FormulaUUID)),nl
    ; assert(inactive_prob2_formula(FormulaUUID))).
deactivate_prob2_formula(FormulaUUID) :-
    add_internal_error('Cannot deactivate unknown formula: ',deactivate_prob2_formula(FormulaUUID)).

% mark certain formulas as active again; will not be computed by evaluate_registered_formula
activate_prob2_formulas(Fs) :- maplist(activate_prob2_formula,Fs).
activate_prob2_formula(FormulaUUID) :- prob2_formula(FormulaUUID,_,_,_),!,
   (retract(inactive_prob2_formula(FormulaUUID)) -> true ; print(formula_already_active(FormulaUUID)),nl).
activate_prob2_formula(FormulaUUID) :- add_internal_error('Cannot activate unknown formula: ',activate_prob2_formula(FormulaUUID)).

/**
Takes a state id and a list of formula ids (that were registered already)
via register_prob2_formula predicate. Then it evaluates all of the formulas
for the given state id.

#### called by:
* ProB 2.0: EvaluateRegisteredFormulasCommand
*/
:- mode evaluate_registered_formulas(+StateId, +FormIds, -Results).
evaluate_registered_formulas(StateId,FormIds,Results) :-
    get_state_for_eval(StateId,State,Kind,ConstID),
    maplist(evaluate_registered_formula(State,Kind,ConstID), FormIds, Results).

% some special treatment for XTL, CPS mode to obtain raw Prolog State
get_state_for_eval(StateId,PrologState,full_initialised_state,empty_state) :- animation_mode(M), (M=xtl ; M=cspm),!,
   visited_expression(StateId,PrologState).
get_state_for_eval(StateId,State,Kind,ConstID) :- get_bstate_with_deferred_sets(StateId,State,Kind,ConstID).

:- use_module(probsrc(bsyntaxtree),[get_texpr_type/2]).
evaluate_registered_formula(_,_,_,FormulaUUID,Result) :- inactive_prob2_formula(FormulaUUID),!,
    Result = errors('NOT-INITIALISED',['Inactive Formula']).
%evaluate_registered_formula(State,StateKind,FormulaUUID,Result) :-
%    prob2_formula(FormulaUUID,Typed,Truncate,Requirements),!,
%    tools:ajoin([Requirements,':',StateKind],R), Result = result(R,[]).
evaluate_registered_formula(State,StateKind,ConstID,FormulaUUID,Result) :-
    prob2_formula(FormulaUUID,Typed,Truncate,Requirements),!,
    get_texpr_type(Typed,Type),
    (cached_registered_formula_result(FormulaUUID,ConstID,R) -> Result = R
      ; requirements_met(Requirements,StateKind) ->
           evaluate_typechecked_b_formula_in_state(State,Typed,Type,Truncate,Result),
           store_result_in_cache(Requirements,ConstID,FormulaUUID,Type,Result)
      ;  StateKind=requires_constants ->
           Result = errors('NOT-INITIALISED',['SETUP CONSTANTS FIRST']) %Result = errors('CONSTANTS NOT INITIALISED',[])
      ;  Result = errors('NOT-INITIALISED',['INITIALISE VARIABLES FIRST'])).
evaluate_registered_formula(_State,_StateKind,_ConstID,FormulaUUID,Result) :-
      add_error(prob2_interface,'Unknown formula id:',FormulaUUID),
      Result = errors('INTERNAL ERROR',['Unknown formula id']).

% check whether we can evaluate the formula in the state:
requirements_met(requires_nothing,_).
requirements_met(requires_constants,constants_only_state).
requirements_met(requires_constants,full_initialised_state).
requirements_met(requires_variables,full_initialised_state).
requirements_met(formula(Typed),State) :-
   (State=full_initialised_state -> true
    ; determine_type_of_formula(Typed,Requirements), requirements_met(Requirements,State)).

:- dynamic cached_registered_formula_result/3.
store_result_in_cache(requires_nothing,ConstID,FormulaUUID,Type,Result) :- cache_this_type(Type), !,
    assert(cached_registered_formula_result(FormulaUUID,ConstID,Result)).
store_result_in_cache(requires_constants,ConstID,FormulaUUID,Type,Result) :- cache_this_type(Type), ConstID>=0,!,
    assert(cached_registered_formula_result(FormulaUUID,ConstID,Result)).
store_result_in_cache(_,_,_,_,_).

cache_this_type(pred). % currently we only cache predicates, anyway: it makes no sense to cache identifiers (their values are stored in the state anyway)


% API for showing Tk Animation Images in Java FX (or other):
:- use_module(state_viewer_images,[get_animation_images/1,
    get_animation_image_grid/6, get_react_to_item_right_click_options/4, react_to_item_right_click/6]).
get_animation_image_list(ImageList) :- get_animation_images(ImageList).
% Format: [image_file(0,'images/empty_box_white.gif'),...,image_file(6,'images/F.gif')]

get_animation_image_matrix_for_state(ID,Matrix,MinRow,MaxRow,MinCol,MaxCol) :-
   (get_animation_image_grid(ID,M,M1,M2,M3,M4)
    -> Matrix=M, MinRow=M1, MaxRow=M2, MinCol=M3, MaxCol=M4
     ; Matrix=[], MinRow = -1, MaxRow = 0, MinCol = -1, MaxCol = 0).
% Format: [entry(1,1,image(0)),... entry(2,3,text(some_atom)),...],  entry(Row,Col,ImgOrText)

% returns a list of atoms (strings) for the various options that are available in the state ID
% for Row/Col (Y/X)
get_react_to_item_right_click_options_for_state(ID,Row,Col,Options) :-
   get_react_to_item_right_click_options(ID,Col,Row,Options).

% should be called for one option provided by get_react_to_item_right_click_options_for_state
react_to_item_right_click_option_for_state(ID,Row,Col,Option,TransitionID,NewID) :-
   react_to_item_right_click(ID,Col,Row,Option,TransitionID,NewID).

% | ?- prob2_interface:get_react_to_item_right_click_options_for_state(3,1,1,L).
% L = ['Set(1,1,1)','Set(1,1,2)','Set(1,1,3)','Solve'|...] ?
% | ?- prob2_interface:react_to_item_right_click_option_for_state(3,1,1,'Set(1,1,1)',NewID).
% Performed: Set(int(1),int(1),int(1))
% NewID = ..., TransitionID=...

/**
Takes a predicate and finds a list of all state ids for which the
predicate holds. The states that are not intitialized (i.e. root) are
included in the list. The list that is returned is therefore the union
of the uninitialised states and the states for which the predicate holds.

#### called by:
* ProB 2.0: GetStatesFromPredicate
*/
:- mode get_states_for_predicate(+Raw,-States,-Errors).
get_states_for_predicate(Raw,States,Errors) :-
      evaluate_formula_typecheck('#PREDICATE',Raw,Typed,Errors),
      findall(StateId,get_state_for_predicate(StateId,Typed),States).

get_state_for_predicate(StateId,Typed) :-
      state_space:visited_expression(StateId,StatePacked),
      ( is_initialised_state(StateId) ->
          expand_const_and_vars_to_full_store(StatePacked,State),
          eval_predicate(State,Typed,'TRUE',_,_)
          ;
          true
      ).

/**
This cycles through all of the solutions and extracts the string
representation of the solutions.
*/
prettyprint_solutions(_Truncate,[],[]).
prettyprint_solutions(Truncate,[bind(Name,Res)|T],[PP|R]) :-
    pretty_print_bvalue(Truncate,Res,PPRes), % pretty_print_bvalue_unicode ?
    PP = solution(Name,PPRes),
    prettyprint_solutions(Truncate,T,R).

/**
Allows Strings to be parsed and evaluated. Used for EvalB

#### called by:
* ProB 2.0: RemoteEvaluateCommand
*/
:- use_module(eval_strings).
evalb_evaluate_formula(Codes, result(R,Solutions,Quantor,Vars,Warn)):-
    eval_strings:eval_codes(Codes,exists,R,Warn,LocalState,TypeInfo),
    extract_typeinfo(TypeInfo,Quantor,Vars),
    prettyprint_solutions(truncate,LocalState,Solutions).

evalb_evaluate_tautology(Codes, result(R,Solutions,Quantor,Vars,Warn)):-
    eval_strings:eval_codes(Codes,forall,R,Warn,LocalState,TypeInfo),
    extract_typeinfo(TypeInfo,Quantor,Vars),
    prettyprint_solutions(truncate,LocalState,Solutions).

extract_typeinfo(expression(_),expression,[]).
extract_typeinfo(predicate(forall(BL)),forall,L) :- translate_list(BL,L).
extract_typeinfo(predicate(exists(BL)),exists,L) :- translate_list(BL,L).
extract_typeinfo(predicate(no_outer_quantifier),predicate,[]).

translate_list([],[]).
translate_list([H|T],[HT|TT]) :- translate_bexpression(H,HT), translate_list(T,TT).

/**
Takes a list of state ids and a predicate and finds all of the states
for which the predicate is true

#### called by:
* ProB 2.0: FilterStatesForPredicateCommand
*/
:- mode filter_states_for_predicate(+Raw,+States,-Filtered).
filter_states_for_predicate(Raw,States,Filtered) :-
      evaluate_formula_typecheck('#PREDICATE',Raw,Typed,Errors),
      (Errors=[] ->
        filter_states_for_typed_predicate(Typed,States,Filtered)
        ;
        Filtered = errors(Errors)
      ).

filter_states_for_typed_predicate(_,[],[]).
filter_states_for_typed_predicate(Typed,[Id|States],Filtered) :-
      get_bstate_with_deferred_sets(Id,State,Kind,_),
      ( Kind = full_initialised_state ->   %is_initialised_state(Id) ->
          (eval_predicate(State,Typed,'TRUE',_,_) ->
              Filtered = [Id|Rest] ;
              Filtered = Rest
          )
          ; Filtered = [Id|Rest]
      ),
      filter_states_for_typed_predicate(Typed,States,Rest).

/* ------------------------- */
/* Formula Expansion         */
/* ------------------------- */

/**
Gets all top-level formula IDs from bvisual2.

#### called by:
* ProB 2.0: GetTopLevelFormulasCommand
*/
:- mode get_top_level_formulas(-TopIds).
get_top_level_formulas(TopIds) :-
    suppress_rodin_positions(C),
    bv_get_top_level(TopIds),
    reset_suppress_rodin_positions(C).

/**
Inserts a formula into bvisual2 module
The formula is inserted as a child of level "user" in bvisual2

#### called by:
* ProB 2.0: InsertFormulaForVisualizationCommand
*/
:- mode insert_formula_for_expansion(+AST,-Id).
insert_formula_for_expansion(Typed,Id) :- Typed = b(_,_,_),!,
    suppress_rodin_positions(C),
    bv_insert_formula(Typed,user,Id),
    reset_suppress_rodin_positions(C).

insert_formula_for_expansion(AST,Id) :-
    suppress_rodin_positions(C),
    b_type_expression(AST,[variables],_,Typed,Errors),
    ( Errors == [] ->
        bv_insert_formula(Typed,user,Id),
        reset_suppress_rodin_positions(C)
    ; otherwise ->
        add_all_perrors(Errors), reset_suppress_rodin_positions(C),
        fail).

/**
Uses the bvisual2 module to evaluate the given formulas.
Unlike expand_formula_nonrecursive/6, this predicate only evaluates the formulas,
but doesn't return their label, description, or subformulas.

#### called by:
* ProB 2.0: EvaluateBVisual2FormulasCommand
*/
:- mode prob2_evaluate_bvisual2_formulas(+Ids,+StateId,-Values).
prob2_evaluate_bvisual2_formulas(Ids,StateId,Values) :-
    bv_get_values(Ids,StateId,Values).

/**
Uses the bvisual2 module to expand a specified formula,
optionally also expanding child formulas recursively and/or evaluating all expanded formulas.
FormulaId must be a bvisual2 formula ID,
as returned for example by get_top_level_formulas/1, insert_formula_for_expansion/2, or prob2_expand_bvisual2_formula/3.

#### called by:
* ProB 2.0: ExpandBVisual2FormulaCommand
*/
:- mode prob2_expand_bvisual2_formula(+FormulaId,+Options,-ExpandedFormula).
prob2_expand_bvisual2_formula(FormulaId,Options,ExpandedFormula) :-
    % Check that no unknown or duplicate options were passed.
    (selectchk(evaluate(_), Options, Options1) -> true ; Options1 = Options),
    (selectchk(recursive, Options1, Options2) -> true ; Options2 = Options1),
    Options2 == [],

    suppress_rodin_positions(C),
    expand_bvisual2_formula_internal(Options,FormulaId,ExpandedFormula),
    reset_suppress_rodin_positions(C).

expand_bvisual2_formula_internal(Options,FormulaId,formula(Entries)) :-
    findall(Attribute, bvisual2_formula_attribute(FormulaId,Options,Attribute), Entries).

bvisual2_formula_attribute(FormulaId,_,id(FormulaId)).
bvisual2_formula_attribute(FormulaId,_,description(Description)) :-
    bv_formula_description(FormulaId,Description).
bvisual2_formula_attribute(FormulaId,_,functor_symbol(FunctorSymbol)) :-
    bv_get_formula_functor_symbol(FormulaId,FunctorSymbol).
bvisual2_formula_attribute(FormulaId,_,rodin_labels(Labels)) :-
    bv_formula_labels(FormulaId,Labels).
bvisual2_formula_attribute(FormulaId,Options,value(EvaluatedValue)) :-
    memberchk(evaluate(StateId), Options),
    % TODO Refactor to combine multiple bv_get_values calls?
    bv_get_values([FormulaId], StateId, [EvaluatedValue]).
bvisual2_formula_attribute(FormulaId,_,type(Type)) :-
    (bv_is_typed_predicate(FormulaId) ->
        Type = predicate;
        (bv_is_typed_formula(FormulaId) -> Type = expression; Type = other)).
bvisual2_formula_attribute(FormulaId,Options,Attribute) :-
    bv_expand_formula(FormulaId,Label,ChildrenIds),
    (Attribute = label(Label) ; bvisual2_formula_children_attribute(ChildrenIds,Options,Attribute)).


bvisual2_formula_children_attribute(ChildrenIds,Options,children(ExpandedChildren)) :-
    memberchk(recursive, Options),
    !,
    maplist(expand_bvisual2_formula_internal(Options), ChildrenIds, ExpandedChildren).
bvisual2_formula_children_attribute(ChildrenIds,_,children_ids(ChildrenIds)).


/* ------------------------- */
/* Model Checking            */
/* ------------------------- */

/**
#### do_modelchecking(+MaxNumberOfStatesToCheck, +Time,+Options,-Result,stats(-NrNodes,-NrTrans,-NrProcessed))
* +MaxNumberOfStatesToCheck : maximum number of states to be checked
* +Time : Timeout specified by the user in ms
* +Options : List of options specified by the user. Used for predicate do_modelchecking(Time,Options,Result)
* -NrNodes : total number of nodes in state space. Calculated with get_state_space_stats
* -NrTrans : total number of nodes in state space. Calculated with get_state_space_stats
* -NrProcessed : total number of calculated nodes in state space. Calculated with get_state_space_stats
 When Time Milliseconds have elapsed the modelchecker should stop after its next step

#### called by:
* ProB Plugin: ModelCheckingCommand
* ProB 2.0: ModelCheckingStepCommand
*/
:- mode do_modelchecking(+MaxNumberOfStatesToCheck,+Time,+Options,-Result,-Stats).
do_modelchecking(MaxNumberOfStatesToCheck, Time, Options, Result, stats(NrNodes,NrTrans,NrProcessed)) :-
    statistics(walltime, [CurTime,_]), /* get current time in ms */
    LimitTime is CurTime+Time,
    option_set(find_deadlocks, Options, Deadlock),
    option_set(find_invariant_violations, Options, Invariant),
    option_set(find_assertion_violations, Options, Assertions),
    (option_set(ignore_state_errors, Options, 1) -> FindStateErrors=0 ; FindStateErrors=1),
    option_set(inspect_existing_nodes, Options, InspectExistingNodes),
    option_set(stop_at_full_coverage, Options, StopAtFullCoverage),
    (option_set(breadth_first_search, Options, 1) -> set_depth_breadth_first_mode(breadth_first) ;
     option_set(depth_first_search, Options, 1) -> set_depth_breadth_first_mode(depth_first) ;
     set_depth_breadth_first_mode(mixed)),
    option_set(find_goal, Options, Goal),
    (option_set(partial_guard_evaluation, Options, 1)
      -> preferences:get_preference(pge,PartialGuardEvaluation)
      ;  PartialGuardEvaluation = off),
    (option_set(partial_order_reduction, Options, 1)
      -> preferences:get_preference(por,WithPOR)
      ;  WithPOR = off),
    % no partial guards evaluation
    tcltk_interface:do_model_check(MaxNumberOfStatesToCheck, _, LimitTime, Res, Deadlock, Invariant, Goal,Assertions,
                        FindStateErrors,StopAtFullCoverage, WithPOR, PartialGuardEvaluation, InspectExistingNodes),
    build_modelcheck_return(Res, Result),
    get_state_space_stats(NrNodes, NrTrans, NrProcessed).


/**
Sets the goal for model checking. The Option search_for_goal needs to be set when starting
model checking in order for this to have an effect on the model checking.

#### called by:
* ProB 2.0: SetBGoalCommand
*/
:- mode set_goal_for_model_checking(+Goal).
set_goal_for_model_checking(Goal) :-
    evaluate_formula_typecheck('#PREDICATE',Goal,TypedGoal,Errors),
    (Errors = [] ->
        b_set_parsed_typed_machine_goal(TypedGoal);
        add_error_and_fail(set_goal_for_model_checking,typeerror,Errors)
      ).

/**
Takes an atom and unifies Result with 1 if the atom is in the List.
Otherwise, Result is unified with 0.
*/
:- mode option_set(+Element, +List, -Result).
option_set(Element, List, Result) :-
    ( member(Element,List)
        -> Result = 1
        ;  Result = 0).

:- mode build_modelcheck_return(+MCRes, -JavaResult).
build_modelcheck_return(MCRes, JavaResult) :-
    build_modelcheck_return2(MCRes, JavaResult) -> true
    ; nl, print('### Unknown Model Check Error Result:'), print(MCRes), nl, /* TODO: Why do we print something here? Can we get rid of this print? (Q? from Joy)*/
        current_state_id(State), JavaResult=general_error(State, MCRes).

build_modelcheck_return2(no, not_yet_finished(100000)).
build_modelcheck_return2([timeout,N], not_yet_finished(N1)) :- N1 is 100000 - N.
build_modelcheck_return2(deadlock, deadlock(State)):- current_state_id(State).
build_modelcheck_return2(invariant_violation, invariant_violation(State)):- current_state_id(State).
build_modelcheck_return2(assertion_violation, assertion_violation(State)):- current_state_id(State).
build_modelcheck_return2(state_error, state_error(State)):- current_state_id(State).
build_modelcheck_return2(state_error(_), state_error(State)):- current_state_id(State).
build_modelcheck_return2(goal_found, goal_found(State)) :- current_state_id(State).
build_modelcheck_return2(well_definedness_error, well_definedness_error(State)) :- current_state_id(State).
build_modelcheck_return2(general_error_occurred, general_error(State)):- current_state_id(State).
build_modelcheck_return2(full_coverage, full_coverage).
build_modelcheck_return2(all, Res) :- max_reached_or_timeout_for_node(_),!, Res=ok_not_all_nodes_considered.
build_modelcheck_return2(all, ok).

:- use_module(symbolic_model_checker(bmc),[bmc_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(kinduction), [kinduction_symbolic_model_check/1, tinduction_symbolic_model_check/1]).
:- use_module(symbolic_model_checker(ic3), [ic3_symbolic_model_check/1]).
symbolic_model_check(bmc,Result) :-
   bmc_symbolic_model_check(Result).
symbolic_model_check(kinduction,Result) :-
   kinduction_symbolic_model_check(Result).
symbolic_model_check(tinduction,Result) :-
   tinduction_symbolic_model_check(Result).
symbolic_model_check(ic3,Result) :-
   ic3_symbolic_model_check(Result).

/**
Computes the coverage statistics of the current state space at any given time.
The information of interest includes the total number of nodes and transitions, as well as
a list of statistics about the nodes and operations and a list of the operations that have been uncovered sofar.

#### called by:
* ProB Plugin: ComputeCoverageCommand
* ProB 2.0: ComputeCoverageCommand
*/
:- use_module(coverage_statistics, [compute_the_coverage/5, operation_hit/2,query_node_hit/2, uncovered_operation/1]).

:- mode compute_efficient_statespace_stats(-NrNodes, -NrTrans, -NrProcessed).
compute_efficient_statespace_stats(NrNodes, NrTrans, NrProcessed) :-
  get_state_space_stats(NrNodes, NrTrans, NrProcessed).

:- mode compute_coverage(-TotalNodeNr,-TotalTransSum,-NodeStat,-OpStat,-Uncovered).
compute_coverage(TotalNodeNr,TotalTransSum,NodeStat,OpStat,Uncovered)  :-
    compute_the_coverage(_,TotalNodeNr,TotalTransSum,false,false),
    findall(S2,(operation_hit(OpS,Nr),string_concatenate(':',Nr,S1),string_concatenate(OpS,S1,S2)),OpStat),
    findall(S2,(query_node_hit(Prop,Nr),string_concatenate(':',Nr,S1),string_concatenate(Prop,S1,S2)),NodeStat),
    findall(OpName, uncovered_operation(OpName),Uncovered).


get_modelchecking_coverage(TotalNodeNr,TotalTransSum,NodeStat,OpStat,Uncovered)  :-
    compute_the_coverage(_,TotalNodeNr,TotalTransSum,false,false),
    findall(entry(OpS,Nr),operation_hit(OpS,Nr),OpStat),
    findall(entry(Prop,Nr),query_node_hit(Prop,Nr),NodeStat),
    findall(OpName, uncovered_operation(OpName),Uncovered).

/**
Takes an option for the statistics and returns the corresponding value.

#### called by:
* ProB 2.0: GetStatisticsCommand
*/

:- mode get_statistics(+Option,-Value).
get_statistics(global_runtime,V) :- statistics(runtime,[V,_]).
get_statistics(global_walltime,V) :- statistics(walltime,[V,_]).
get_statistics(delta_runtime,V) :- statistics(runtime,[_,V]).
get_statistics(delta_walltime,V) :- statistics(walltime,[_,V]).
get_statistics(memory_used,V) :- statistics(memory_used,V).
get_statistics(memory_free,V) :- statistics(memory_free,V).
get_statistics(global_stack_used,V) :- statistics(global_stack_used,V).
get_statistics(local_stack_used,V) :- statistics(local_stack_used,V).
%get_statistics(global_stack_free,V) :- statistics(global_stack_free,V).
%get_statistics(local_stack_free,V) :- statistics(local_stack_free,V).
get_statistics(trail_used,V) :- statistics(trail_used,V).
get_statistics(choice_used,V) :- statistics(choice_used,V).
get_statistics(atoms_used,V) :- statistics(atoms_used,V).
get_statistics(atoms_nbused,V) :- statistics(atoms_nbused,V).
get_statistics(gc_count,V) :- statistics(gc_count,V).
get_statistics(gc_time,V) :- statistics(gc_time,V).

/**
Performs deadlock freedom checking with constraint Predicate and calulates the Result.

#### called by:
* ProB 2.0: ConstraintBasedDeadlockCheckCommand
*/
:- mode prob2_deadlock_freedom_check(+Predicate,-Result).
prob2_deadlock_freedom_check(Predicate,Result) :-
    b_type_expression(Predicate,[variables],pred,TPredicate,Errors),
    ( Errors == [] ->
        prob2_deadlock_freedom_check1(TPredicate,Result)
    ; otherwise ->
        Result = errors(Errors)).
prob2_deadlock_freedom_check1(Predicate,Result) :-
    % always do a deadlock check with SMT mode enabled
    call_with_smt_mode_enabled(prob2_deadlock_freedom_check2(Predicate,Result)).
:- use_module(clpfd_interface,[catch_clpfd_overflow_call1/1]).
:- use_module(b_state_model_check,[cbc_deadlock_freedom_check/3]).
prob2_deadlock_freedom_check2(Predicate,Result) :-
    user_interruptable_call_det(clpfd_interface:catch_clpfd_overflow_call1(
                                 b_state_model_check:cbc_deadlock_freedom_check(State,Predicate,0)),
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = deadlock(Transition,StateId),
        add_artificial_transition(root,deadlock_check,State,StateId,Transition)).
prob2_deadlock_freedom_check2(_Predicate,no_deadlock_found).






/**
Performs invariant checking for either for all operations or a list of operations.

#### called by:
* ProB 2.0: ConstraintBasedInvariantCheckCommand
*/
:- mode prob2_invariant_check(+Ops,-Result).
prob2_invariant_check(all,Result) :-
    findall(OpName,b_is_operation_name(OpName),Ops),
    prob2_invariant_check2(Ops,Result).
prob2_invariant_check(ops(Ops),Result) :-
    prob2_invariant_check2(Ops,Result).
prob2_invariant_check2(Ops,Result) :-
    call_with_smt_mode_enabled(prob2_invariant_check3(Ops,Result,[])).
prob2_invariant_check3([]) --> !.
prob2_invariant_check3([Op|Rest]) -->
    prob2_invariant_check_for_single_op(Op),
    prob2_invariant_check3(Rest).
prob2_invariant_check_for_single_op(OpName,In,Out) :-
    ( clpfd_interface:catch_clpfd_overflow_call1(
             b_state_model_check:state_model_check_invariant(OpName,State1,Operation,State2)) ->
        In = [counterexample(OpName,Trans1,Trans2)|Out],
        atom_concat( invariant_check_ , OpName, RootTrans),
        add_artificial_transition(root,    RootTrans,State1,StateId1,Trans1),
        add_artificial_transition(StateId1,Operation,State2,_StateId2,Trans2)
    ; otherwise ->
        In = Out).


:- use_module(b_state_model_check,[cbc_find_redundant_invariants/2]).
prob2_redundant_invariants(Redundant, Timeout) :-
  cbc_find_redundant_invariants(Redundant, Timeout).

/**
creates a helper transition that is artificially added to the state space (e.g. during deadlock checking)
This transition is added to the state space.
A triple TransitionTuple in the form op(TransId,OpName,SrcId,DstId) for this transition is generated.
*/
:- mode add_artificial_transition(+SrcId,+Operation,+DstState,+DstId,-TransitionTuple).
add_artificial_transition(SrcId,Operation,DstState,DstId,TransitionTuple) :-
    tcltk_interface:tcltk_add_new_transition_transid(SrcId,Operation,DstId,DstState,[],TransId),
    extract_op_name(Operation,OpName),
    TransitionTuple = op(TransId,OpName,SrcId,DstId).

/**
 computes the enabling relation information for the provided operations of interest
 get_enable_matrix(-PairsOfOperations,+EnableResult)
  PairsOfOperations: list of pair(Op1,Op2) of operations pairs for which the enable relation is to be computed
  EnableResult: list of terms enable_edges(Op1,Op2,enable_edges(E,KE,D,KD)) of same length
*/
get_enable_matrix(PairsOfOperations,EnableResult) :-
    maplist(compute_enable_matrix_entry(100),PairsOfOperations,EnableResult).

:- use_module(probsrc(enabling_analysis),[compute_cbc_enable_rel/4]).
compute_enable_matrix_entry(ExtraTimeout,pair(OpName1,OpName2),
                           enable_rel(OpName1,OpName2,
                                      enable_edges(Enable,KeepEnabled,Disable,KeepDisabled))) :-
   compute_cbc_enable_rel(OpName1,OpName2,ExtraTimeout,[Enable,KeepEnabled,Disable,KeepDisabled]).

/**
Performs an LTL model checking step.

#### called by:
* ProB 2.0: LtlCheckingCommand
*/
:- mode prob2_do_ltl_modelcheck(+Formula,+Max,-Result,-Errors).
prob2_do_ltl_modelcheck(Formula,Max,Result,Errors) :-
   typecheck_temporal_formula(Formula,TypeCheckedFormula,Status),
   ( Status=ok ->
       ltl_model_check_with_ce(TypeCheckedFormula,Max,init,Result1),
       prob2_ltl_adapt_operations(Result1,Result),
       get_error_messages(_WarningsOnly,Errors)
   ; otherwise ->
       Result=typeerror,
       get_error_messages(_WarningsOnly,Errors)).

prob2_ltl_adapt_operations(counterexample(CE1,LoopEntry,PathToCE1), Res) :- !,
    Res = counterexample(CE,LoopEntry,PathToCE),
    create_simple_op_terms(PathToCE1,root,PathToCE),
    prob2_ltl_adapt_ce(CE1,CE).
prob2_ltl_adapt_operations(Result,Result).

prob2_ltl_adapt_ce([],[]).
prob2_ltl_adapt_ce([atom(StateId,_,OpTuple)|Irest],[Transition|Orest]) :-
    prob2_ltl_adapt_ce2(OpTuple,StateId,Transition),
    prob2_ltl_adapt_ce(Irest,Orest).
prob2_ltl_adapt_ce2(none,_StateId,none).
prob2_ltl_adapt_ce2((TransId,Action,DestId),StateId,op(TransId,Name,StateId,DestId)) :-
    extract_op_name(Action,Name).

/* ------------------------- */
/* Find Traces               */
/* ------------------------- */

/**
Finds a trace from the root state to the specified state in the current state space.

#### Parameters:
* StateId
* Trace - List of op tuples op(OpId,SrcId,DestId) corresponding to the trace calculated by ProB or atom no_trace_found if the call was unsuccessful

#### called by:
* ProB 2.0: GetShortestTraceCommand
*/
:- mode find_trace_to_node(+StateId,-Trace).
find_trace_to_node(StateId, Trace) :-
    find_trace_from_node_to_node(root, StateId, Trace).

/**
Finds a trace from one state to a goal state in the current state space.

#### Parameters:
* FromId - Id of source node
* ToId - Id of destination node
* Trace - List of op tuples op(OpId,SrcId,DestId) corresponding to the trace calculated by ProB or atom no_trace_found if the call was unsuccessful

#### called by:
* ProB 2.0: GetShortestTraceCommand
*/
:- mode find_trace_from_node_to_node(+FromId,+ToId,-Trace).
find_trace_from_node_to_node(FromId, ToId, Trace) :-
    (tcltk_interface:find_shortest_trace_to_node(FromId, ToId, OpIDs, _TraceIDs) ->
        trace_to_op_triple(OpIDs,FromId,Trace);
        Trace = no_trace_found
    ).

/**
Translates a list of transition ids to a list of op tuples op(TransId,OpName,SrcId,DestId)
*/
:- mode trace_to_op_triple(+ListOpIds,+CurID,-ListTuples).
trace_to_op_triple([],_,[]).
trace_to_op_triple([OpID|T], SrcID, [op(OpID,Name,SrcID,Dest)|OpT]) :-
    transition(SrcID,Action,OpID,Dest),
    extract_op_name(Action,Name),
    trace_to_op_triple(T,Dest,OpT).

/**
Takes a given predicate and finds a state in the state space that satisfies the predicate.
A helper transition is then added to go to the goal state.

#### called by:
* ProB 2.0: FindValidStateCommand
*/
:- use_module(b_state_model_check,[b_set_up_valid_state_with_pred/4]).
find_state_for_predicate(Predicate,UseInvariant,Result) :-
    evaluate_formula_typecheck('#PREDICATE',Predicate,TPredicate,Errors),
    ( Errors == [] ->
        find_state_for_predicate1(TPredicate,UseInvariant,Result)
    ; otherwise ->
        Result = errors(Errors)).
find_state_for_predicate1(Predicate,UseInvariant,Result) :-
    user_interruptable_call_det(clpfd_interface:catch_clpfd_overflow_call1(
                                  b_state_model_check:b_set_up_valid_state_with_pred(State,Predicate,UseInvariant,none)), % TODO: pass UseConstantsFromStateID
                                InterruptResult),!,
    ( InterruptResult = interrupted ->
        Result = interrupted
    ;  State = time_out ->
        Result = interrupted
    ; otherwise ->
        Result = state_found(Transition,StateId),
        add_artificial_transition(root,find_valid_state,State,StateId,Transition)).
find_state_for_predicate1(_Predicate,_,no_valid_state_found).

/**
#### called by:
* ProB Plugin (de.prob.eventb.disprover.core): DisproverCommand
*/
:- mode cbc_disprove(+Goal,+AllHypotheses,+SelectedHypotheses,+TimeoutFactor,-OutResult).
cbc_disprove(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,OutResult) :-
    disprove(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,OutResult),
    % remove the warning regarding double check, in rodin it is shown in the proof tree anyway
    (get_error(warning(disprover_inconsistent_hypotheses),_) ; true).

% the same with explicit Options:
:- mode cbc_disprove(+Goal,+AllHypotheses,+SelectedHypotheses,+TimeoutFactor,-OutResult).
cbc_disprove(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,Options,OutResult) :-
    disprove_with_opts(Goal,AllHypotheses,SelectedHypotheses,TimeoutFactor,Options,OutResult),
    % remove the warning regarding double check, in rodin it is shown in the proof tree anyway
    (get_error(warning(disprover_inconsistent_hypotheses),_) ; true).

/**
Takes a predicates and produces a result from the selected solver
#### called by:
* ProB 2.0: CbcSolveCommand
*/
:- mode cbc_solve(+Solver,+Predicate,-Identifiers,-Result).
cbc_solve(Solver,Predicate,Identifiers,Result) :-
    cbc_solve(Solver,Predicate,_State,Identifiers,Result).
cbc_solve(Solver,Predicate,State,Identifiers,Result) :-
    cbc_solve(Solver,Predicate,State,1,Identifiers,Result).
cbc_solve(Solver,Predicate,State,TimeoutFactor,Identifiers,Result) :-
    cbc_solve_type(Solver,Predicate,TPredicate),
    find_identifier_uses(TPredicate,[],Identifiers),
    (atomic(State) -> Options = [solve_in_visited_state(State)] ; LState=State),
    cbc_solve_typed(Solver,TPredicate,LState,TimeoutFactor,Options,Result).

:- assert_must_succeed((cbc_solve_with_opts('PROB',[truncate(10)],
                        equal(none,identifier(none,x),integer(none,10)),Identifiers,Result),
                       Identifiers == [x],
                        Result == solution([binding(x,int(10),'10')]))).

% This should be the new entry point for CbcSolveCommand, cbc_solve is no longer used
:- mode cbc_solve_with_opts(+Solver,+Options,+Predicate,-Identifiers,-Result).
cbc_solve_with_opts(Solver,Options,Predicate,Identifiers,Result) :-
    maplist(prob2_interface:check_cbc_solve_opts,Options),
    cbc_solve_type(Solver,Predicate,TPredicate),
    find_identifier_uses(TPredicate,[],Identifiers),
    (select(timeout_factor(TimeoutFactor),Options,Opts2) -> true ; TimeoutFactor=1,Opts2=Options),
    cbc_solve_typed(Solver,TPredicate,_,TimeoutFactor,Opts2,Result).

check_cbc_solve_opts(full_machine_state) :- !.
check_cbc_solve_opts(solve_in_visited_state(ID)) :- !,atomic(ID).
check_cbc_solve_opts(timeout_factor(Nr)) :- !,number(Nr).
check_cbc_solve_opts(truncate(Nr)) :- !,number(Nr). % truncate pretty printing
check_cbc_solve_opts(truncate) :- !. % truncate pretty printing
check_cbc_solve_opts(force_evaluation) :- !. % force evaluation of symbolic results
check_cbc_solve_opts(IO) :- add_internal_error('Illegal cbc_solve option:',IO).

cbc_solve_typed('PROB',Predicate,State,TimeoutFactor,Options,Result) :-
    solve_predicate(Predicate,State,TimeoutFactor,['SMT','CLPFD'|Options],Result).
cbc_solve_typed('KODKOD',Predicate,State,TimeoutFactor,Options,Result) :-
    solve_predicate(Predicate,State,TimeoutFactor,['KODKOD','SMT','CLPFD'|Options],Result).
cbc_solve_typed('SMT_SUPPORTED_INTERPRETER',Predicate,State,TimeoutFactor,Options,Result) :-
    solve_predicate(Predicate,State,TimeoutFactor,['SMT_SUPPORTED_INTERPRETER','SMT','CLPFD'|Options],Result).
cbc_solve_typed(SOLVER,Predicate,_StateID,_,Options,Result) :- recognised_smt_solver(SOLVER,InternalName),
    (member(solve_in_visited_state(ID),Options) ->
      smt_solve_predicate_in_state(ID,InternalName,Predicate,_State,Result)
      ; smt_solve_predicate(InternalName,Predicate,_State,Result)).

recognised_smt_solver('Z3',z3).
recognised_smt_solver('CVC4',cvc4).

:- use_module(preferences,[temporary_set_preference/3,reset_temporary_preference/2]).
cbc_solve_type('KODKOD',Pred,TPred) :- !,
    temporary_set_preference(try_kodkod_on_load,true,C),
    call_cleanup(cbc_solve_type2(Pred,TPred),
                 reset_temporary_preference(try_kodkod_on_load,C)).
cbc_solve_type(_,Pred,TPred) :- cbc_solve_type2(Pred,TPred).
cbc_solve_type2(Pred,TPred) :-
    !, get_eval_scope(Scope),
    b_type_open_predicate(no_quantifier,Pred,Scope,TPred,Errors),
    (Errors=[] -> true ; add_error_and_fail(register_prob2_formula, 'Type-Errors: ', Errors)).

/**
Takes a predicates and generates a pretty printed string
*/
:- assert_must_succeed((pretty_print_predicate(equal(none,identifier(none,x),integer(none,1)),[],Result),
                        Result == 'x = 1')).
:- assert_must_succeed((pretty_print_predicate(not_equal(none,identifier(none,x),integer(none,1)),[latex],Result),
                        Result == '\\mathit{x} \\neq  1')).
:- assert_must_succeed((pretty_print_predicate(not_equal(pos(2,-1,1,1,1,7),integer(pos(3,-1,1,1,1,2),1),
                        integer(pos(4,-1,1,6,1,7),2)),[latex,nopt],PPString),
                        PPString == '1 \\neq  2' )).

pretty_print_predicate(Pred,Options,PPString) :-
    select(nopt,Options,Opts2),
    !,
    temporary_set_preference(optimize_ast,false,CHNG1),
    call_cleanup(pretty_print_predicate(Pred,Opts2,PPString),
                 reset_temporary_preference(optimize_ast,CHNG1)).
pretty_print_predicate(Pred,Options,PPString) :-
    get_eval_scope(Scope),
    b_type_open_predicate(no_quantifier,Pred,Scope,TPred,Errors),
    (Errors=[] -> true ; add_error_and_fail(pretty_print_predicate, 'Type-Errors: ', Errors)),
    translate_pred(TPred,Options,PPString).

:- use_module(translate,[set_unicode_mode/0, unset_unicode_mode/0, unicode_mode/0,
           set_latex_mode/0, unset_latex_mode/0, latex_mode/0]).
translate_pred(TPred,Options,PPString) :-
    select(unicode,Options,Opts2), \+ unicode_mode,
    !,
    set_unicode_mode,translate_pred(TPred,Opts2,PPString),unset_unicode_mode.
translate_pred(TPred,Options,PPString) :-
    select(latex,Options,Opts2), \+ latex_mode,
    !,
    set_latex_mode,translate_pred(TPred,Opts2,PPString),unset_latex_mode.
translate_pred(TPred,_,PPString) :-
    translate_bexpression(TPred,PPString).

% Constraint-Based Test-case generation
% TO DO: return values
:- use_module(sap,[cbc_gen_test_cases/5]).
cbc_generate_test_cases(TargetPred,MaxDepth,OutputFile) :-
   % I am not sure whether we should call b_parse_machine_predicate(TargetPred,...) or not
    Events = all,
    cbc_gen_test_cases(Events,TargetPred,MaxDepth,OutputFile,_Uncovered).



:- use_module(sap,[create_testcase_path/5]).
% example call | ?- prob2_interface:prob2_find_test_path([enter1],truth(unknown),200,R).
:- mode prob2_find_test_path(+Events,+EndPredicate,+TimeoutMs,-ResultOpTerms).
% Result is either errors(Errors), timeout, interrupt, infeasible_path, or list of Operation IDs
prob2_find_test_path(Events,EndPredicate,TimeoutMs,ResultOpTerms) :-
    evaluate_formula_typecheck('#PREDICATE',EndPredicate,TPredicate,Errors),
    ( Errors == [] ->
        prob2_find_test_path_aux(Events,TPredicate,TimeoutMs,ResultOpTerms)
     ;  ResultOpTerms = errors(Errors)).

prob2_find_test_path_aux(Events,TPredicate,TimeoutMs,ResOperationIds) :-
    ( create_testcase_path(init,Events,TPredicate,TimeoutMs,Trace)
     -> (is_list(Trace) -> maplist(extract_opid,Trace,ResOperationIds)
         % Trace can be "timeout" or "interrupt", too
           ; ResOperationIds=Trace)
    ; ResOperationIds = infeasible_path).

extract_opid((TransId,OpTerm,StateId,DestId),op(TransId,Name,StateId,DestId)) :- extract_op_name(OpTerm,Name).

/* ------------------------- */
/* Check CSP Assertion       */
/* ------------------------- */

/**
Takes a list of assertions and produce a list of results and result traces.

TODO: We should modify the result traces so that they are useful for ProB 2.0. (or at least document what the result traces mean)

#### called by:
* ProB 2.0: CSPAssertionsCommand
*/
:- mode check_csp_assertions(+Assertions,-Results,-ResultTraces).
check_csp_assertions(Assertions,Results,ResultTraces) :-
    maplist(check_csp_assertion,Assertions,Results,ResultTraces).

check_csp_assertion(AssClause,Res,ResTrace1) :-
    read_from_codes(AssClause,Assertion),
    tcltk_interface:checkAssertion(Assertion,_PP,_Negated,Res,ResTrace),
    (ResTrace = no_counter_example -> ResTrace1 = []; ResTrace1=ResTrace).

/* ------------------------- */
/* Preferences Interface     */
/* ------------------------- */

/**
Returns a list of all the preferences with their current values

#### called by:
* ProB 2.0: GetCurrentPreferencesCommand
*/
:- mode list_current_eclipse_preferences(-L).
list_current_eclipse_preferences(L) :-
    findall(preference(A,B),find_eclipse_preference(A,B),L).

find_eclipse_preference(A,B) :-
   list_all_eclipse_preferences(X),
   member(preference(A,_,_,_,_),X),
   get_eclipse_preference(A,B).

/**
Returns the current value of a specified preference.

#### called by:
* ProB 2.0: GetPreferenceCommand
*/
get_eclipse_preference(PrefS,PrefVal) :-
    if(eclipse_preference(PrefS,PS),
       get_preference(PS,PrefVal),
       add_unknown_preference_error_and_fail(get_eclipse_preference,PrefS)).

/**
Returns a list of all normal eclipse preferences as well as their information
(i.e. type, description, category, and default value)

#### called by:
* ProB Plugin: GetPreferencesCommand
* ProB 2.0: GetDefaultPreferencesCommand
*/
:- mode list_eclipse_preferences(-L).
list_eclipse_preferences(L) :-
    findall(preference(A,B,C,D,E),
       (get_eclipse_preference_infos(A,B,C,D,E),
        \+ advanced_eclipse_preference(A,_)), L).
:- mode list_all_eclipse_preferences(-L). % also includes advanced eclipse preferences
list_all_eclipse_preferences(L) :-
    findall(preference(A,B,C,D,E),get_eclipse_preference_infos(A,B,C,D,E),L).

get_eclipse_preference_infos(PrefString,Type,Description,Category,DefaultValue) :-
    eclipse_preference(PrefString,PS),
    %\+ advanced_eclipse_preference(PrefString,PS), % we now want to show all preferences in ProB 2
    preference_val_type(PS,Type),
    preference_description(PS,Description),
    preference_category(PS,Category),
    preference_default_value(PS,DefaultValue).

/**
Sets a preference

#### called by:
* ProB Plugin: SetPreferenceCommand
* ProB 2.0: SetPreferenceCommand
*/
:- use_module(tools_strings, [convert_cli_arg/2]).
:- mode set_eclipse_preference(+PrefS,+PrefVal).
set_eclipse_preference(PrefS,PrefVal) :-
    convert_cli_arg(PrefVal,Value),
    convert_pref_value(Value,CValue),
    (eclipse_preference(PrefS,P)
     -> (set_preference(P,CValue)
          -> true
           ; add_error(set_eclipse_preference,'Could not set preference: ',Value))
      ; obsolete_eclipse_preference(PrefS)
        -> add_warning(set_eclipse_preference,'Obsolete preference: ',PrefS) % just warn and continue
      ; add_unknown_preference_error_and_fail(set_eclipse_preference,PrefS)
    ).

convert_pref_value('TRUE',V) :- !, V=true.
convert_pref_value('FALSE',V) :- !, V=false.
convert_pref_value(X,X).

:- use_module(tools_matching,[get_possible_preferences_matches_msg/2]).
add_unknown_preference_error_and_fail(Source,PrefS) :-
    obsolete_eclipse_preference(PrefS),!,
    add_error_and_fail(Source,'Obsolete preference: ',PrefS).
add_unknown_preference_error_and_fail(Source,PrefS) :-
    get_possible_preferences_matches_msg(PrefS,FuzzyMsg),!,
    ajoin(['Unknown preference: ',PrefS,'. Did you mean: '],Msg),
    add_error_and_fail(Source,Msg,FuzzyMsg).
add_unknown_preference_error_and_fail(Source,PrefS) :-
    add_error_and_fail(Source,'Unknown preference: ',PrefS).

/* ------------------------- */
/* Apply Graph reduction     */
/* ------------------------- */

/**
Takes a list of ignored events, and applies the signature merge algorithm
from module `state_space_reduction` to the current state space.

#### called by:
* ProB 2.0: ApplySignatureMergeCommand
*/
:- mode get_signature_merge_state_space(+IgnoredEvents,-Space).
get_signature_merge_state_space(IgnoredEvents,Space) :-
    reset_ignored_events,
    set_ignored_events(IgnoredEvents),
    compute_signature_merge,
    findall(node(NodeId,Count,Color,Labels),extract_node_info(NodeId,Count,Color,simple_list,Labels),Nodes),
    findall(trans(TransId,Src,Dest,Label,Style,Color),extract_trans_info(true,TransId,Src,Dest,Label,Style,Color),Trans),
    Space = [Nodes,Trans].

/**
Takes a list of ignored events, and calculates a transition diagram
using module `state_space_reduction` for the current state space.

#### called by:
* ProB 2.0: CalculateTransitionDiagramCommand
*/
:- mode get_transition_diagram(+ParsedExpr,-Space).
get_transition_diagram(ParsedExpr,Space) :-
    evaluate_formula_typecheck('#EXPRESSION',ParsedExpr,TypedExpr,Errors),
    (Errors = [] -> true; add_error_and_fail(get_transition_diagram,typeerror,Errors)),
    compute_transition_diagram(TypedExpr),
    findall(node(NodeId,Count,Color,Labels),extract_node_info(NodeId,Count,Color,gen_label,Labels),Nodes),
    findall(trans(TransId,Src,Dest,Label,Style,Color),extract_trans_info(false,TransId,Src,Dest,Label,Style,Color),Trans),
    Space = [Nodes,Trans].

write_dotty_transition_diagram(Expression,Filename) :-
    write_dotty_for_expr(transition_diagram,Expression,Filename).

write_dotty_signature_merge(IgnoredEvents,Filename) :-
    write_signature_merge_to_dotfile(IgnoredEvents,Filename).

write_dotty_state_space(Filename) :-
    write_dotty(state_space,Filename).

:- use_module(meta_interface,[is_dot_command/1, call_dot_command/3]).
% find out which commands only require a filename:
is_dotty_command(Command) :- is_dot_command(Command).
% call commands which generate a dot file (without requiring further arguments, such as an expression)
write_dotty(Command,Filename) :- OptionalArgs=[],
   call_dot_command(Command,Filename,OptionalArgs).


:- use_module(meta_interface,[is_dot_command_for_expr/1, call_dot_command_for_expr/4]).
% find out which commands only require an expression and a filename:
is_dotty_command_for_expr(Command) :- is_dot_command_for_expr(Command).
% call commands which generate a dot file from an expression:
write_dotty_for_expr(Command,Expr,Filename) :- OptionalArgs=[],
   call_dot_command_for_expr(Command,Expr,Filename,OptionalArgs).


:- use_module(meta_interface,[command_additional_info/2, command_unavailable/2]).

% --------------------------------------------------------
% New preferred API for calling DOT / TABLE commands in ProB2 JFX:
%   get_dot_commands_in_state/2 and call_dot_command_in_state/4
%   get_table_commands_in_state/2 and call_table_command_in_state/4


% example call: prob2_interface:get_dot_commands_in_state(1,List).
get_dot_commands_in_state(StateID,List) :- get_commands_in_state(dot,StateID,List).
get_commands_in_state(Category,StateID,List) :-
   tcltk_interface:tcltk_goto_node_with_id(StateID),
   findall(command(Command,Name,Description,NumberOfFormulaArgs,RelevantEclipsePrefs,AdditionalInfo,AvailMsg),
           (is_a_command(Category,Command,Name,Description,NumberOfFormulaArgs,RelevantPrefs,AdditionalInfo),
            maplist(convert_pref,RelevantPrefs,RelevantEclipsePrefs),
           (command_unavailable(Command,AvailMsg) -> true ; AvailMsg = available)),
           List).
% example call: prob2_interface:get_table_commands_in_state(1,List).
get_table_commands_in_state(StateID,List) :- get_commands_in_state(table,StateID,List).

convert_pref(Pref,Res) :- eclipse_preference(ECLIPSEPREF,Pref),!,Res=ECLIPSEPREF.
convert_pref(P,P) :- print(cannot_convert_pref(P)),nl.

:- use_module(meta_interface,[is_dot_command/6, is_table_command/6]).
is_a_command(dot,Command,Name,Description,NumberOfFormulaArgs,RelevantPreferences,AdditionalInfo) :-
   is_dot_command(Command,Name,Description,NumberOfFormulaArgs,RelevantPreferences,AdditionalInfo).
is_a_command(table,Command,Name,Description,NumberOfFormulaArgs,RelevantPreferences,AdditionalInfo) :-
   is_table_command(Command,Name,Description,NumberOfFormulaArgs,RelevantPreferences,AdditionalInfo).


call_dot_command_in_state(StateID,Command,Formulas,DotFile) :-
   tcltk_interface:tcltk_goto_node_with_id(StateID),
   call_dot_command_for_dotfile(Command,Formulas,DotFile).

% call a dot command for generating a DotFile; used by call_dot_command_in_state
call_dot_command_for_dotfile(Command,[],DotFile) :-
   is_dot_command(Command), !, OptionalArgs=[],
   call_dot_command(Command,DotFile,OptionalArgs).
call_dot_command_for_dotfile(Command,[Expr],DotFile) :-
   is_dot_command_for_expr(Command), !, OptionalArgs=[],
   call_dot_command_for_expr(Command,Expr,DotFile,OptionalArgs).
call_dot_command_for_dotfile(Command,Formulas,DotFile) :-
   add_internal_error('Illegal dot call: ',call_dot_command_for_dotfile(Command,Formulas,DotFile)),
   fail.


% example call: prob2_interface:call_table_command_in_state(1,expr_as_table,[integer(pos,1)],Table).
:- use_module(meta_interface,[call_command/5]).
call_table_command_in_state(StateID,Command,Formulas,TableResult) :-
   (tcltk_interface:tcltk_goto_node_with_id(StateID) -> true
     ; add_error(call_table_command_in_state,'Could not reach state: ',StateID)
    ),
   append(Formulas,[TableResult],ActualArgs),
   OptionalArgs=[],
   debug_println(9,call_command(table,Command,_,ActualArgs,OptionalArgs)),
   call_command(table,Command,_,ActualArgs,OptionalArgs), debug_println(9,result(TableResult)),
   !.
call_table_command_in_state(StateID,Command,Formulas,TableResult) :-
   add_error(call_table_command_in_state,'Table command failed:',call_table_command_in_state(StateID,Command,Formulas,TableResult)),
   TableResult = list([list(['ERROR OCCURED'])]).


/**
Writes a dot representation of the given state to the specified file.

*/
:- mode write_dot_for_state_viz(+StateId, +Filename).
write_dot_for_state_viz(StateId, Filename) :-
    get_state(StateId, State),
    print_cstate_graph(State, Filename).

/**
Generates information about the nodes found during state space reduction

#### Generated Information:
* Count - number of concrete states combined in the abstract state
* Color - the color used to represent this state type in a visualization
* Labels - determine the labels that should appear on a node in a visualization
*/
:- mode extract_node_info(+NodeId,-Count,-Color,+LabelGenerator,-Labels).
extract_node_info(NodeId,Count,Color,LabelGenerator,Labels) :-
    get_reduced_node(AbsState,Count,Witness,NodeId),
    generate_node_color(NodeId,Witness,AbsState,Count,Color),
    generate_node_labels(LabelGenerator,AbsState,Labels).

/**
Generates information about the transitions created during state space reduction

#### Generated Information:
* Src - ID corresponding to the abstract state that is the source of the abstract transition with TransId
* Dest - ID corresponding to the abstract state that is the destination of the abstract transition with TransId
* Label - the label that should appear on a transition in a visualization
* Style - the style that should be applied to a transition of this type in a visualization (e.g. dashed)
* Color - the color used to represent a transition of this type in a visualization
*/
:- mode extract_trans_info(+ShowSelfLoops,+TransId,-Src,-Dest,-Label,-Style,-Color).
extract_trans_info(ShowSelfLoops,TransId,Src,Dest,Label,Style,Color) :-
    reduced_trans(Src,AbsAction,Count,Dest,TransId),
    generate_transition_label(AbsAction,Count,Label),
    generate_transition_color_and_style(ShowSelfLoops,Src,AbsAction,Dest,Color,Style).


/* ------------------ */
/* Get Errors         */
/* ------------------ */

/**
Returns a list of strings containing all error messages (of the error_manager).
Clears all the errors. Returns the empty list if no errors occurred since
the last call to get_error_messages.

#### called by:
* ProB 2.0: GetErrorsCommand
*/
:- mode get_error_messages(-WarningsOnly,-ListOfErrMsgs).
get_error_messages(WarningsOnly,Res) :-
    (real_error_occurred -> WarningsOnly = false, Res = ListOfErrMsgs
     ; warning_occurred -> WarningsOnly = true, Res = ListOfErrMsgs
     ; WarningsOnly = true, Res = [] % ignore messages for now; they appear in Prolog Console; TO DO: support new return type in ProB2
    ),
    ignore_user_interrupt_det((get_all_errors_and_reset(ListOfErrMsgs) ->  true  ; ListOfErrMsgs=[])).

% each error is of the form: error(ErrMsg,ErrType,ErrLocations)
% ErrMsg is an atom (aka string)
% ErrType is warning, internal_error or error
% ErrLocations is a list of terms  error_span(Filename,StartLine,StartCol,EndLine,EndCol)
% Note: Filename is '' when not known
:- mode get_error_messages_with_span_info(-ListOfErrMsgTerms).
get_error_messages_with_span_info(ListOfErrMsgTerms) :-
    ignore_user_interrupt_det(get_all_errors_with_span_info_and_reset(ListOfErrMsgTerms)).

% deprecated in favor of get_error_messages_with_span_info/1
% (WarningsOnly is no longer meaningful now that messages are returned as well)
:- mode get_error_messages_with_span_info(-WarningsOnly,-ListOfErrMsgTerms).
get_error_messages_with_span_info(WarningsOnly,ListOfErrMsgTerms) :-
    (real_error_occurred -> WarningsOnly = false ; WarningsOnly = true),
    get_error_messages_with_span_info(ListOfErrMsgTerms).



/* ------------------------- */
/* Evalstore                 */
/* ------------------------- */

/**
TODO: Please document what this does.

#### called by:
* ProB 2.0: EvalstoreCreateByStateCommand
*/
es_copy_from_statespace(StateId,EvalstoreId) :-
    evalstore_create_store_by_state(StateId,EvalstoreId).

/**
TODO: Please document what this does.

#### called by:
* ProB 2.0: EvalstoreEvalCommand
*/
:- mode es_eval(+EvalstoreId,+UntypedFormula,+Timeout,-Result).
es_eval(EvalstoreId,UntypedFormula,Timeout,Result) :-
    evalstore_eval_formula(EvalstoreId,UntypedFormula,Timeout,Result1), !,
    (Result1 = ok(V,TypedIds,StoreId) ->
        Result = ok(V,PP,Ids,StoreId),
        get_texpr_ids(TypedIds,Ids),
        es_eval_translate(V,PP)
    ; otherwise                       -> Result = Result1).
es_eval_translate(pred(V),V).
es_eval_translate(value(V),PP) :-
    translate_bvalue(V,PP).

% ####################

/**

Animates randomly number of steps until a certain (LTL?) condition is fulfilled.

#### Called by:
* ProB 2.0: ExecuteUntilCommand

#### Arguments
* CurState - the ID of the current state
* Condition - a condition that should be satisfied at some point (this come as parsed term from the ProB2)

#### Generated Information:
* Trace - a list of triples representing a trace in the state space of the model being analysed
* Result - the result found: ltl_found, typeerror, max_nr_of_steps_reached, deadlock
*/
:- mode generate_trace_until_condition_fulfilled(+CurState,+Condition,-Trace,-Result).
generate_trace_until_condition_fulfilled(CurState,Condition,Trace,Result) :-
   typecheck_temporal_formula(Condition,TypeCheckedCondition,Status),
   ( Status=ok -> find_trace1(CurState,TypeCheckedCondition,no_loop,100000,Trace, Result) %FIXME: turn maximal numver of steps into an argument?
   ; otherwise -> get_error_messages(_WOnly,_Errors), % Reset errors. We don't want to throw a ProBError on the Java side because the result is capsuled in Result.
                  Result=typeerror, Trace=[]).

find_trace1(StateId,Ltl,Type,MaxSteps,OpTripleResultTrace, Result) :-
    (MaxSteps =< 0
       -> add_error(find_trace,'Number of maximum animation steps should be a positive integer. The number of steps which was given is ',MaxSteps)
        ; true),
    set_current_state(StateId), !, % can be backtracked
    % negate -> counterexample is the trace we are looking for
    preprocess_formula(Ltl,Ltl2),
    find_trace_aux(StateId,not(Ltl2),Type,0,MaxSteps,ResultTrace,ResultTrace,Names,Result),
    gen_op_triples(ResultTrace,Names,OpTripleResultTrace).
    %(gen_op_triples(ResultTrace,OpTripleResultTrace) -> print(ok(OpTripleResultTrace,result(Result))),nl
    %  ; add_error_and_fail(prob2_interface,'trace not correctly generated',ResultTrace)).

find_trace_aux(CurID,Condition,Type,_N,_Max,StateTransitionHistory,[],[],RESULT) :-
    debug_println(9,checking_ltl_formula(CurID,StateTransitionHistory)),
    evaluate_ltl_formula(Condition,StateTransitionHistory,Type,ltl:check_ap,ltl:callback_tp,EvResult),
    EvResult = false,
    !,
    RESULT=ltl_found.
find_trace_aux(CurID,Condition,Type,N,Max,StateTransitionHistory,STTail,Names,RESULT) :-
    tcltk_interface:tcltk_compute_options(CurID,ActionsAndIDs),
    debug_println(9,opts(ActionsAndIDs)),
    ( N=Max
       -> debug_println(9,'Maximum number of animation steps reached.'),
          Names = [],
          RESULT = maximum_nr_of_steps_reached
    ; pick_action(random,ActionsAndIDs,ActionId,Name,DstID) % pick first one; we could do random
       -> debug_println(9,performing_action(ActionId,from_to(CurID,DstID),opts(ActionsAndIDs))),
          STTail = [strans(CurID,ActionId)|STTail2],
          Names = [Name|Names2],
          N1 is N+1,
          find_trace_aux(DstID,Condition,Type,N1,Max,StateTransitionHistory,STTail2,Names2,RESULT)
    ; otherwise
       -> STTail=[], Names = [], RESULT = deadlock
     ).

gen_op_triples([],[],[]).
gen_op_triples([_X],[_Y],[]).
gen_op_triples([strans(CurID,ActionId),strans(DstID,ActId)|T],[Name|NameT],[op(ActionId,Name,CurID,DstID)|Rest]) :-
   gen_op_triples([strans(DstID,ActId)|T],NameT,Rest).

:- use_module(library(random)).
:- use_module(library(lists)).
pick_action(first,[(ActionId,Term,DstID)|_], ActionId, Name, DstID) :- extract_op_name(Term,Name).
pick_action(random,Options,ActionId, Name, DstID) :-
    length(Options,Len),
    L1 is Len+1,
    random(1,L1,RanChoice),
    debug_println(9,random(RanChoice,Len)),
    nth1(RanChoice,Options,(ActionId,ActionAsTerm,DstID)),
    extract_op_name(ActionAsTerm,Name).


% -------------------------------------

% an execution engine with minimal overhead: states are not stored in visited_expression database, only first enabled operation is taken
:- mode execute_model(+CurStateID,+MaxNrSteps,-TransitionInfo,-ExecutedSteps,-Result).
execute_model(CurStateID,MaxNrSteps,TransitionInfo,ExecutedSteps,Result) :-
  execute_model(CurStateID,MaxNrSteps,[],TransitionInfo,ExecutedSteps,Result).

:- mode execute_model(+CurStateID,+MaxNrSteps,+Options,-TransitionInfo,-ExecutedSteps,-Result).
% Options can contain continue_after_errors, timeout(MS)
% Result is either maximum_nr_of_steps_reached, deadlock, error, internal_error, time_out
execute_model(CurStateID,MaxNrSteps,Options,TransitionInfo,ExecutedSteps,Result) :-
  visited_expression(CurStateID,CurState),
  % we could do something like prepare_state_for_specfile_trans(CurState,CurState2) + ensure that the individual steps do not repack constants
  execute_model_steps(0,CurState,MaxNrSteps,Options,NewState,ExecutedSteps,Result),
  (ExecutedSteps>0
      -> tcltk_interface:tcltk_add_new_transition_transid(CurStateID,'$JUMP'('EXECUTE'(ExecutedSteps)),ToID,NewState,[],TransId),
         debug_println(4,added_transition_for_execute(ExecutedSteps,ToID)),
         tcltk_interface:tcltk_goto_state('$JUMP'('EXECUTE'(ExecutedSteps)),ToID),
         TransitionInfo = op(TransId,'EXECUTE',CurStateID,ToID)
        ; TransitionInfo = none).

execute_model_steps(StepNr,CurState,MaxNrSteps,_Options,NewState,ExecutedSteps,Result) :-
    StepNr >= MaxNrSteps,!,
    NewState=CurState,
    ExecutedSteps=StepNr, Result=maximum_nr_of_steps_reached.
execute_model_steps(StepNr,CurState,MaxNrSteps,Options,NewState,ExecutedSteps,Result) :-
    cli_trans_aux(StepNr,CurState,Options,ActionName,_Act,State2,ErrorRes),
    !,
    debug_println(20,execute(StepNr,ActionName)),
    %print(step(StepNr,Act,State2)),nl,
    S1 is StepNr+1,
    (nonvar(ErrorRes) -> Result=ErrorRes, ExecutedSteps=StepNr, NewState=CurState
     ; execute_model_steps(S1,State2,MaxNrSteps,Options,NewState,ExecutedSteps,Result)).
execute_model_steps(Steps,CurState,_,_Options,CurState,Steps,deadlock).

:- use_module(specfile,[specfile_trans_or_partial_trans/7]).
:- use_module(error_manager,[throw_enumeration_warnings_in_current_scope/0, add_internal_error/2, error_occurred_in_error_scope/0]).
:- use_module(tools_meta,[safe_time_out/3]).

% TO DO: use same optimisation as in -execute in prob_cli.pl (get_possible_next_operation_for_execute)
cli_trans_aux(StepNr,CurState,Options,ActionName,Act,NewState,ErrorRes) :-
    catch_enumeration_warning_exceptions(
        (throw_enumeration_warnings_in_current_scope,
         (member(timeout(MS),Options) ->
           safe_time_out(specfile_trans_or_partial_trans(CurState,ActionName,Act,NewState,_TransInfo,Residue,_),MS,TR)
           ; specfile_trans_or_partial_trans(CurState,ActionName,Act,NewState,_TransInfo,Residue,_) % no time-out !
          ),
           (TR==time_out
            -> add_error(execute,'Timeout occured during execute after step: ',StepNr),ErrorRes=time_out
            ; error_occurred_in_error_scope ->
                 (member(continue_after_errors,Options) -> true
                   ; add_error(execute,'Error occured during execute after step: ',StepNr),ErrorRes=error)
            ; true)
         ),
             (add_error(virtual_time_out_execute,'Virtual TIME-OUT occured during execute after step: ',StepNr),
              ActionName = '*** VIRTUAL_TIME_OUT ***', Act=ActionName,
              ErrorRes=time_out)
              ),
    (Residue=[] -> true
     ; add_internal_error('Residue during execute after step: ',StepNr:Residue),
       (nonvar(ErrorRes) -> true ; ErrorRes=internal_error)).

% -------------------------------------

% computes a unsat core
:- use_module(tools,[start_ms_timer/1, stop_ms_walltimer_with_msg/2]).
get_unsat_core_with_fixed_conjuncts(Pred,FixedPreds,CoreOut) :-
    typecheck_pred_for_unsat_core(Pred,TypedPred),
    maplist(typecheck_pred_for_unsat_core,FixedPreds,TypedFixedPreds),
    conjunct_predicates(TypedFixedPreds,Conj),
    format('Computing an unsat core~n',[]), start_ms_timer(Timer),
    unsat_chr_core_with_fixed_conjuncts_auto_time_limit(TypedPred,Conj,2000,Core),
    stop_ms_walltimer_with_msg(Timer,'Computed unsat core'),
    translate_bexpression(Core,CoreOut).

% computes the unsat core of minimal size:
get_minimum_unsat_core_with_fixed_conjuncts(Pred,FixedPreds,CoreOut) :-
    typecheck_pred_for_unsat_core(Pred,TypedPred),
    maplist(typecheck_pred_for_unsat_core,FixedPreds,TypedFixedPreds),
    conjunct_predicates(TypedFixedPreds,Conj),
    format('Computing the minimal unsat core~n',[]),
    minimum_unsat_core_with_fixed_conjuncts(TypedPred,Conj,Core),
    translate_bexpression(Core,CoreOut).

typecheck_pred_for_unsat_core(PIn,POut) :-
    evaluate_formula_typecheck('#PREDICATE',PIn,TExpr,[]),
    (get_texpr_expr(TExpr,exists(_,POut)) -> true
      ; format(user_error,'Predicate for unsat core is not existentially quantified~n',[]),
         POut = TExpr).

/**
Access information about the current version of the ProB core.

#### called by:
* ProB 2.0: GetInternalRepresentationPrettyPrintCommand
*/
:- use_module(specfile,[get_internal_representation/1]).
get_pretty_print(PP) :-
    get_internal_representation(PPC),
    atom_codes(PP,PPC).

get_pretty_print_unicode(PP) :-
    set_unicode_mode,
    call_cleanup(get_pretty_print(PP),unset_unicode_mode).

:- use_module(symbolic_model_checker(predicate_handling)).
get_primed_predicate(Pred,PrimedPredOut) :-
    evaluate_formula_typecheck('#PREDICATE',Pred,POutT,[]),
    prime_predicate(POutT,PrimedPred),
    translate_bexpression(PrimedPred,PrimedPredOut).

:- use_module(weakest_preconditions).
:- use_module(preferences, [call_with_preference/3]).
get_weakest_precondition(OpName,Pred,WPOut) :- % TODO: call with preference can be removed once ProB 2 reads ASTs instead of reparsing
    call_with_preference(get_weakest_precondition_aux(OpName,Pred,WPOut),translate_ids_to_parseable_format,true).
get_weakest_precondition_aux(OpName,Pred,WPOut) :-
    b_get_machine_operation(OpName,_Results,_Parameters,OpBody),
    evaluate_formula_typecheck('#PREDICATE',Pred,POutT,[]),
    weakest_precondition(OpBody,POutT,WP),
    translate_bexpression(WP,WPOut).

:- use_module(before_after_predicates,[before_after_predicate_for_operation/2]).
before_after_predicate(OpName,PredicateOut) :-
    call_with_preference(before_after_predicate_aux(OpName,PredicateOut),translate_ids_to_parseable_format,true).
before_after_predicate_aux(OpName,PredicateOut) :-
    before_after_predicate_for_operation(OpName,Predicate),
    translate_bexpression(Predicate,PredicateOut).

/** Synthesis Commands:
*
#### b_synthesis:start_synthesis_from_ui/13 called by:
* ProB 2.0: BSynthesisCommand
*/
start_synthesis_from_ui_(SynthesisMode,AdaptMachineCode,SolverTimeOut,Library,DoNotUseConstants,Solver,ConsiderIfVarNames,Operation,SynthesisType,Positive,Negative,NewMachineAtom,Distinguishing) :-
  start_synthesis_from_ui(SynthesisMode,AdaptMachineCode,SolverTimeOut,Library,DoNotUseConstants,Solver,ConsiderIfVarNames,Operation,SynthesisType,Positive,Negative,NewMachineAtom,Distinguishing).

/*
#### b_synthesis:start_synthesis_single_operation_from_ui/11 called by:
* BSynthesis: StartSynthesisCommand
*/
start_synthesis_single_operation_from_ui_(SolverTimeOut,Operations,Library,DoNotUseConstants,Solver,Operation,action,Positive,Negative,CacheOperationTuple,Distinguishing) :-
  start_synthesis_single_operation_from_ui(SolverTimeOut,Operations,Library,DoNotUseConstants,Solver,Operation,action,Positive,Negative,CacheOperationTuple,Distinguishing).

/*
#### b_synthesis:reset_synthesis_context/0 called by:
* ProB 2.0: ResetBSynthesisCommand
*/
reset_synthesis_context_ :- reset_synthesis_context.

/*
#### synthesis_util:get_invariant_violating_vars_from_examples/3 called by:
* BSynthesis: GetViolatingVarsFromExamples
*/
get_invariant_violating_vars_from_examples_(Positive,Negative,ViolatingVarNames) :-
  get_invariant_violating_vars_from_examples(Positive,Negative,ViolatingVarNames).

/*
#### synthesis_util:get_valid_and_invalid_equality_predicates_for_operation/6 called by:
* BSynthesis: VisualizeOperationCommand
*/
get_valid_and_invalid_equality_predicates_for_operation_(OperationName,ValidAmount,InvalidAmount,ValidPrettyEqualityTuples,InvalidPrettyEqualities,IgnoredIDs) :-
  get_valid_and_invalid_equality_predicates_for_operation(OperationName,ValidAmount,InvalidAmount,ValidPrettyEqualityTuples,InvalidPrettyEqualities,IgnoredIDs).

/*
#### synthesis_util:get_valid_and_invalid_equality_predicates_for_invariants/4 called by:
* BSynthesis: VisualizeInvariantsCommand
*/
get_valid_and_invalid_equality_predicates_for_invariants_(ValidAmount,InvalidAmount,ValidPrettyEqualities,InvalidPrettyEqualities) :-
  get_valid_and_invalid_equality_predicates_for_invariants(ValidAmount,InvalidAmount,ValidPrettyEqualities,InvalidPrettyEqualities).

/*
#### synthesis_util:adapt_machine_code_for_operations/2 called by:
* BSynthesis: AdaptMachineCodeForOperationsCommand
*/
adapt_machine_code_for_operations_(Operations,NewMachineAtom) :-
  adapt_machine_code_for_operations(Operations,NewMachineAtom).

/*
#### predicate_data_generator:generate_synthesis_data_from_predicate_untyped_/5 called by:
* BSynthesisDataGenerator: SynthesisDataFromPredicateCommand
*/
generate_synthesis_data_from_predicate_untyped_(MachinePath, AugmentRecords, SolverTimeoutMs, UntypedPredAst, AugmentedSetOfData) :-
    generate_synthesis_data_from_predicate_untyped(MachinePath, AugmentRecords, SolverTimeoutMs, UntypedPredAst, AugmentedSetOfData).

/*
#### operation_data_generator:generate_operation_data_from_machine_path_/4 called by:
* BSynthesisDataGenerator: SynthesisDataFromOperationCommand
*/
generate_operation_data_from_machine_path_(MachinePath, AugmentRecords, SolverTimeoutMs, OperationData) :-
    generate_operation_data_from_machine_path(MachinePath, AugmentRecords, SolverTimeoutMs, OperationData).

/*
#### operation_data_generator:generate_data_from_machine_operation_/6 called by:
* BSynthesisDataGenerator: SynthesisDataFromOperationCommand
*/
generate_data_from_machine_operation_(OperationName, Augmentations, SolverTimeoutMs, AbsoluteFilePath, MachineName, OperationData) :-
    generate_data_from_machine_operation(OperationName, Augmentations, SolverTimeoutMs, AbsoluteFilePath, MachineName, OperationData).

/*
#### called by:
* ProB 2.0: GetMachineOperationNamesCommand
*/
:- use_module(probcspsrc(haskell_csp),[channel/2]).

csp_channel_or_start('start_cspm_MAIN').
csp_channel_or_start('start_cspm'). % TO DO: do we also need to support print channel ?
csp_channel_or_start(Name) :- channel(Name,_).

get_machine_operation_names(MachineOperationNames) :- b_or_z_mode,!,
    findall(MachineOperationName,b_is_operation_name(MachineOperationName),MachineOperationNames).
get_machine_operation_names(ChannelNames) :- csp_mode,!,
   findall(Name, csp_channel_or_start(Name),ChannelNames).
get_machine_operation_names([]).

:- use_module(probsrc(bsyntaxtree), [get_texpr_id/2]).
:- use_module(probsrc(bmachine),[b_get_operation_non_det_modifies/2, b_get_operation_normalized_read_write_info/3]).
% get list of operation info terms of the form
% operation_info(Name,ResultNames,ParameterNames,TopLevel,OType)
% where ResultNames and ParameterNames are list of atomic names
% TopeLevel is true if the operation is a top-level operation for animation/model checking
% OType is classic, csp or eventb_operation     %was eventb_operation(ChangeSet,ParaValues,Operation)
get_machine_operation_infos(MachineOperationInfos) :- b_or_z_mode,!,
  % TODO: ensure that we do not need to consider preference(show_eventb_any_arguments,EVENTB)
  % TODO: examine what happens for CSP||B
    findall(operation_info(Name,ResultNames,ParameterNames,TopLevel,OTypeF,Read,Modified,NonDetModifies),
            (b_get_machine_operation(Name,Results,RealParameters,_RealBody,OType,_OpPos),
             maplist(get_texpr_id,Results,ResultNames),
             maplist(get_texpr_id,RealParameters,ParameterNames),
             functor(OType,OTypeF,_),
             % TO DO: obtain machine file or machine name
             (b_top_level_operation(Name) -> TopLevel = true ; TopLevel=false),
             (b_get_operation_normalized_read_write_info(Name,OpRead,Modified)
              -> exclude(op_id,OpRead,Read)
                % exclude query operations called in expressions (allow_operation_calls_in_expr);
                %ProB2 Java cannot deal with it and raises exception for non-atomic identifiers
              ; (Read,Modified)=(unknown,unknown)),
             (b_get_operation_non_det_modifies(Name,NonDetModifies) -> true ; NonDetModifies=unknown)
             ),
            MachineOperationInfos).
get_machine_operation_infos(ChannelInfos) :- csp_mode,!,
   findall(operation_info(Name,[],[],true,csp,[],[],[]), csp_channel_or_start(Name),ChannelInfos).
get_machine_operation_infos([]).

op_id(op(_)).


:- use_module(bmachine,[b_filenumber/4, get_machine_identifiers/2]).

get_all_machine_identifiers(SortedAllIDs) :-
    findall(ID,get_machine_identifiers(_,ID),LIds), append(LIds,AllIDs),
    sort(AllIDs,SortedAllIDs).

% get a list of all machine files: TO DO: extend for CSP, ...
get_machine_files(Files) :-
    findall(b_file(Name,Extension,Filename),b_filenumber(Name,Extension,_,Filename), Files).

% a predicate to obtain possible identifier completions for current machine:
:- use_module(tools_matching,[fuzzy_match_codes_lower_case/2, codes_to_lower_case/2, get_current_keywords/1]).
get_possible_completions(ID,Options,Completions) :-
    get_match_ids(Options,SortedAllIDs),
    atom_codes(ID,IDCodes0),
    (member(lower_case,Options) -> codes_to_lower_case(IDCodes0,IDCodes), Opt=lower_case
     ; Opt=case_sensitive, IDCodes = IDCodes0),
    findall(Target,(member(Target,SortedAllIDs),atom_codes(Target,TargetCodes),
                   match(Opt,IDCodes,TargetCodes)),Completions).

match(case_sensitive,Pattern,TargetCodes) :- !, prefix(TargetCodes,Pattern).
match(lower_case,Pattern,TargetCodes) :- codes_to_lower_case(TargetCodes,TC2), prefix(TC2,Pattern).

get_possible_fuzzy_matches(ID,FuzzyMatches) :-
    get_match_ids([keywords],SortedAllIDs),
    atom_codes(ID,IDCodes),
    findall(Target,(member(Target,SortedAllIDs),atom_codes(Target,TargetCodes),
                   fuzzy_match_codes_lower_case(IDCodes,TargetCodes)),FuzzyMatches).

:- use_module(library(ordsets),[ord_union/3]).
get_match_ids(Options,Ids) :- member(keywords,Options),!,
    get_all_machine_identifiers(SortedAllIDs),
    get_current_keywords(Keywords),
    ord_union(SortedAllIDs,Keywords,Ids).
get_match_ids(_,SortedAllIDs) :- get_all_machine_identifiers(SortedAllIDs).
% -----------------------

% preliminary predicate for obtaining values as raw Prolog terms
% sample call:  prob2_interface:evaluate_expression_prolog(0,boolean_true(pos(13,1,6,5,6,7)),R).
evaluate_expression_prolog(StateId,RawAST,NValuePrologTerm) :-
   evaluate_expression_prolog(StateId,RawAST,_,NValuePrologTerm).
evaluate_expression_prolog(StateId,RawAST,TypedAST,NValuePrologTerm) :-
   get_eval_scope(Scope),
   b_type_expression(RawAST,Scope,_Type,TypedAST,Errors),
   Errors=[],
   get_state(StateId, State),
   evaluate_expression(State,TypedAST,force,_SRes,NValuePrologTerm).

:- use_module(bmachine_static_checks, [extended_static_check_machine/0]).
:- mode prob2_extended_static_check(-Problems).
prob2_extended_static_check(Problems) :-
    extended_static_check_machine,
    get_all_errors_with_span_info_and_reset(Problems).

:- use_module(wdsrc(well_def_analyser), [analyse_wd_for_machine/3]).
:- mode prob2_check_well_definedness(-NrDischarged,-NrTotal).
prob2_check_well_definedness(NrDischarged,NrTotal) :-
    analyse_wd_for_machine(NrDischarged,NrTotal,_).
