:- module(tools_io,[safe_open_file/4,
                    intelligent_open/3,
                    safe_intelligent_open_file/3]).

:- use_module(module_information).
:- module_info(group,infrastructure).
:- module_info(description,'This module contains file io helper predicates.').

:- use_module(error_manager).

safe_open_file(Filename,Mode,Stream,Options) :-
    %print(opening(Filename)),nl,
    on_exception(error(A,B),open(Filename,Mode,Stream,Options),
       (add_error(open_file,'Exception occurred opening file: ',error(A,B)),fail)).


safe_intelligent_open_file(Filename,Mode,Stream) :-
    %print(opening(Filename)),nl,
    on_exception(error(A,B),intelligent_open(Filename,Mode,Stream),
       (add_error(open_file,'Exception occurred opening file: ',error(A,B)),fail)).

intelligent_open(user_error,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_error.
intelligent_open(stderr,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_error.
intelligent_open(user_output,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_output.
intelligent_open(stdout,Mode,Stream) :- (Mode=append ; Mode=write), !, Stream=user_output.
intelligent_open(File,Mode,Stream) :- open(File,Mode,Stream).
