
% Below is the code for the CASE statement
% CASE is now translated to LET + IF-THEN-ELSE in b_ast_cleanup
% The code is kept as reference on how the CASE statement should work:

b_execute_statement2(case(Expression,Cases,Else), _, LocalState, InState, OutState,WF,case(Nr,Path),OR) :- !,
 % According to the AtelierB handbook the cases must be distinct one-by-one; Schneider's book does not seem to require that
 % ProB will at the moment allow multiple entries in which case CASE is non-deterministic
 % Atelier-B will actually only allow literal integers, booleans, enumerated set values as cases !
   b_compute_expression(Expression,LocalState,InState,EValue,WF),
   b_compute_cases(Cases,LocalState,InState,CompCases,AllEitherValues,WF),
   ground_det_wait_flag(WF),
   ( nth1(Nr, CompCases, comp_case_or(ListOfValues, Body)),
     member(EitherValue,ListOfValues),
     equal_object_wf(EValue,EitherValue,case_statement,WF),
     ground_det_wait_flag(WF),
     b_execute_inner_statement(Body,LocalState,InState,OutState,WF,Path,OR)
   ; b_not_member_of_value_lists(AllEitherValues,EValue,WF),
     ground_det_wait_flag(WF),
     % print_message(else_case(Body2)),
     Nr=else,
     b_execute_inner_statement(Else,LocalState,InState,OutState,WF,Path,OR)
    ).


b_compute_cases(Cases,LS,S,CompCases,Values,WF) :-
    b_compute_cases2(Cases,LS,S,CompCases,Values,[],WF).
b_compute_cases2([],_,_,[],AllValues,AllValues,_).
b_compute_cases2([CaseOr|Crest],LS,S,[comp_case_or(Values,Subst)|CCrest],AllValues,VR,WF) :-
    get_texpr_expr(CaseOr,case_or(Exprs,Subst)),
    b_compute_cases3(Exprs,LS,S,Values,AllValues,VI,WF),
    b_compute_cases2(Crest,LS,S,CCrest,VI,VR,WF).
b_compute_cases3([],_,_,[],Values,Values,_).
b_compute_cases3([Expr|Erest],LS,S,[Value|Vrest],[Value|ValuesI],Values,WF) :-
    b_compute_expression(Expr,LS,S,Value,WF),
    b_compute_cases3(Erest,LS,S,Vrest,ValuesI,Values,WF).

b_not_member_of_value_lists([],_,_WF).
b_not_member_of_value_lists([EitherValue|Rest],Value,WF) :-
    not_equal_object_wf(Value,EitherValue,WF),
    b_not_member_of_value_lists(Rest,Value,WF).


/* unused:
state_satisfies_negation_of_invariant(const_and_vars(ID,S)) :- !,
  add_error(state_satisfies_negation_of_invariant,'State not expanded: ',const_and_vars(ID,S)).
state_satisfies_negation_of_invariant(State) :- State \== root, State \= concrete_constants(_),
   %print_message(informational,checking_negation_of_invariant),
   b_get_invariant_from_machine(Invariant),
   set_error_context(checking_negation_of_invariant(State)),
   b_not_test_boolean_expression_for_ground_state(Invariant,[],State ),
   clear_error_context,
   debug_println(9,invariant_violated).

% check each conjunct individually; avoid setting up many choice points for big invariants
b_not_test_boolean_expression_for_ground_state(b(Pred,_,Infos),LS,S) :-
    b_not_test_for_ground_aux(Pred,Infos,LS,S).
b_not_test_for_ground_aux(conjunct(LHS,RHS),_,LocalState,State) :- !,
    (b_not_test_boolean_expression_for_ground_state(LHS,LocalState,State) -> true
     ; b_not_test_boolean_expression_for_ground_state(RHS,LocalState,State) -> true).
b_not_test_for_ground_aux(Pred,Infos,LS,S) :-
    (b_not_test_boolean_expression_wf(b(Pred,pred,Infos),LS,S)
      -> true %,print('INV FALSE: '),translate:print_bexpr(b(Pred,pred,Infos)),nl,translate:print_bstate(S),nl
     ).
*/


/* ----------------------------------*/
/*  b_det_test_boolean_expression    */
/* ----------------------------------*/
% a tool to check for non-deterministic predicates
%b_det_test_boolean_expression(b(Expr,_,Infos),LS,S,WF,Ai,Ao) :- !,
% b_det_test_boolean_expression2(Expr,Infos,LS,S,WF,Ai,Ao).
%
%b_det_test_boolean_expression2(conjunct(LHS,RHS),_,LocalState,State,WF,Ai,Ao) :- !,
%%%   print('check lhs of conjunct: '), translate:print_bexpr(LHS),nl,
%        b_det_test_boolean_expression(LHS,LocalState,State,WF,Ai,Aii),
%%%   print('check rhs of conjunct: '), translate:print_bexpr(RHS),nl,
%        b_det_test_boolean_expression(RHS,LocalState,State,WF,Aii,Ao).
%b_det_test_boolean_expression2(Pred,Infos,LS,S,WF,Ai,Ao) :-
%   debug:watch_det(500,b_interpreter:b_test_boolean_expression(b(Pred,pred,Infos),LS,S,WF,Ai,Ao)).



/* call currently commented out above:
% union
b_compute_union_expression(SV1,Arg2,_Info,LocalState,State,Value,WF) :-
    %nonvar(SV1), SV1 \= closure(_,_,_), SV1 \= global_set(_),
    %Info=[add_element_to_set|_], % use member or remove completely ??
    Arg2 = b(set_extension([SingleEl]),_,_),
    !,
    %print(add_el(SV1,SingleEl)),nl,
    b_compute_expression(SingleEl,LocalState,State,SEl,WF),
    %print(value(SEl)),nl,
    %add_element(SEl,SV1,Value).
    kernel_mappings:must_succ_kernel_call(kernel_objects:add_element(SEl,SV1,Value),(SEl,SV1),WF,union(Arg1,Arg1)).
    % equal_object_wf(SV2,[SEl],WF), kernel_mappings:must_succ_kernel_call(kernel_objects:union_wf(SV1,SV2,Value,WF),(SV1,SEl),WF,union(Arg1,Arg1)).
b_compute_union_expression(SV1,Arg2,_Info,LocalState,State,Value,WF) :-
    b_compute_expression(Arg2,LocalState,State,SV2,WF),
    %print(union(SV1,SV2)),nl,
    %binary_kernel_call(kernel_objects,union_wf,SV1,SV2,Value,WF,union(Arg1,Arg1),Info)
    kernel_mappings:must_succ_kernel_call(kernel_objects:union_wf(SV1,SV2,Value,WF),(SV1,SV2),WF,union(Arg1,Arg1)).
*/



/* not used:
b_generate_closure_from_conjunct(Parameters,Condition1,Condition2,LocalState,State,Result,[CClosurePred1,CClosurePred2],WF) :-
    split_names_and_types(Parameters,Names,Types),
    b_compile(Condition1,Names,LocalState,State,CClosurePred1,WF),
    b_compile(Condition2,Names,LocalState,State,CClosurePred2,WF),
    conjunct_predicates([CClosurePred1,CClosurePred2],ClosurePred),
    construct_closure(Names,Types,ClosurePred,Result).
*/



%  %  now done by create_exists_detect_tautology Detected
%b_not_test_exists([OnePara],Condition,_Infos,LocalState,State,WF) :-
% % This case arises often when performing CBC deadlock checking with parameters  c = Expression
%  % TO DO: generalise ??
%   is_an_equality(Condition,LHS,RHS), get_texpr_id(OnePara,ID),
%   (get_texpr_id(LHS,ID)
%    -> Other=RHS
%    ; get_texpr_id(RHS,ID) -> Other=LHS),
%    debug_println(9,equality_in_not_exists(ID,Other)),trace,tautology exists equality over
%    \+ occurs_in_expr(ID,Other),  % OCCURS CHECK
%    !,   %  not(#ID.(ID=Other)) is equivalent to false
%    (always_well_defined_or_disprover_mode(Other) ->
%       fail  % equality will always be true
%       ; % compute Other, in case it is not well-defined
%         % print(computing_other(Other)),nl,
%         b_compute_expression(Other,LocalState,State,_IgnoredResult,WF)
%    ).
%is_an_equality(b(equal(A,B),_,_),A,B).

/*
b_compute_expressions_symbolic([],[], _, _, [],_WF).
b_compute_expressions_symbolic([EXPRsHd|EXPRsTl],[SYMBOLIC|TS],LocalState,InState,[ValHd|ValTl],WF) :-
   (SYMBOLIC==symbolic -> EXPRsHd=b(E,T,I), EXPR = b(E,T,[prob_annotation('SYMBOLIC')|I])
     ; EXPR = EXPRsHd),
   ~~pp_cll(b_compute_expression(EXPR,LocalState,InState,ValHd,WF)),
   b_compute_expressions_symbolic(EXPRsTl,TS,LocalState,InState,ValTl,WF).
*/

% ----------------------
%% FOR PARTIAL EVALUATOR:
% ----------------------

:- unfold(b_test_boolean_expression_for_ground_state(E,_,_),nonvar(E)).
:- unfold(b_test_boolean_expression_for_ground_state2(E,_,_,_),nonvar(E)).
:- unfold(b_test_boolean_expression_wf(E,_,_),nonvar(E)).
:- unfold(b_test_boolean_expression(E,_,_,_),nonvar(E)).
:- unfold(b_test_boolean_expression(E,_,_,_,_,_),nonvar(E)).
:- unfold(b_test_boolean_expression2(_Expr,_Infos,_LS,_S,_WF,_Ai,_Ao)).
:- unfold(b_test_atomic_boolean_expression2(_,_,_,_,_)).
:- unfold(b_test_implication(_CLHS,_RHS,_LocalState,_State,_WF,_Ai,_Ao)).
:- unfold(block_test_implication(_LHS,_RHS,_LocalState,_State,_WF,_WF2)).
:- unfold(b_test_equivalence(_LHS,_RHS,_LocalState,_State,_WF,_Ai,_Ao)).
:- unfold(b_check_test_equivalence(_PredRes,_LHS,_RHS,_LS,_S,_CWF,_Aii,_Ao)).
:- unfold(b_enum_test_equivalence(_LHS,_RHS,_LS,_S,_WF,_WF2,_)).
:- unfold(b_test_disjunction(_LHS,_RHS,_LS,_S,_WF,_WF2,_Ai)).
:- unfold(b_not_test_conjunction(_LHS,_RHS,_LS,_S,_WF,_WF2,_Ai)).
:- unfold(b_compute_expression(E, _, _, _, _),nonvar(E)).
:- unfold(b_compute_expression2(E,_,_,_, _, _, _),nonvar(E)).
:- execute(special_binary_predicate(_,_,_,_,_,_)).
:- execute(preferences:preference(_,_)).
:- execute(preferences:get_computed_preference(_,_)).
:- execute(debug_println(_,_),debug:debug_mode(off)).
:- execute(pge_algo:do_not_evaluate_guard,(preferences:get_preference(pge,PGE),member(PGE,[enabled,full,enabled2,full2]))).
:- execute(abstract_constant(B, _),nonvar(B)).
:- execute(concrete_constant(B, _),nonvar(B)).
:- execute(animation_minor_mode(_)).
:- execute(get_texpr_type(_B, _)).
:- execute(get_texpr_expr(_B, _)).
%:- execute(get_texpr_expr(B, T), (nonvar(B) ; nonvar(T))).
:- execute(always_well_defined(A),ground(A)).
:- execute(always_well_defined_or_disprover_mode(A),ground(A)).
:- execute(create_texpr(_,_,_,_)).
:- execute(construct_member_closure(ID,_Type,_ClosureSetExpression,_Result),ground(ID)). % could in principle be always executed
:- execute(is_set_type(T, _), nonvar(T)).
:- execute(get_set_type(T, _), nonvar(T)).
:- execute(b_interpreter:is_set_type(T, _), nonvar(T)).
:- execute(binary_function(X,M,P),(nonvar(X);ground(M:P))).
:- execute(lookup_value_in_store_and_global_sets(Id,LS,S,_Val,_Span),    (builtin_tools:is_list_skeleton(LS),builtin_tools:is_list_skeleton(S),ground(Id))).
:- unfold(add_typed_var_to_localstate(_Var,_Val,_Type,_InState,_OutState)).
:- execute(kernel_objects:basic_type(_X,TY),nonvar(TY)).
:- execute(add_var_to_localstate(_Var,_Val,_InState,_OutState)).
%:- execute(lookup_value(Id,LS,S,R),(builtin_tools:is_list_skeleton(LS),builtin_tools:is_list_skeleton(S),ground(Id))).
:- execute(kernel_mappings:binary_boolean_operator(B,_,_), nonvar(B)). % does not seem to work
:- execute(kernel_mappings:binary_boolean_operator(_B,_,_)).
:- execute(empty_avl(_)).
:- execute(b_interpreter_check:register_predicate(Pred,_Infos,_PredTruthVar,_Reused,Ai,_Ao), (ground(Pred),nonvar(Ai))).
:- execute(b_interpreter:register_predicate(Pred,_Infos,_PredTruthVar,_Reused,Ai,_Ao), (ground(Pred),nonvar(Ai))).
:- execute(register_predicate(Pred,_Infos,_PredTruthVar,_Reused,Ai,_Ao), (ground(Pred),nonvar(Ai))).
:- execute(b_interpreter_check:register_predicate(_Pred,_Infos,_PredTruthVar,_Reused,_Ai,_Ao)).
:- execute(b_interpreter:register_predicate(_Pred,_Infos,_PredTruthVar,_Reused,_Ai,_Ao)).
:- execute(b_interpreter:init_wait_flags(_)).
:- execute(clone_wait_flags_from1(_WF,_InnerWF)).
:- execute(b_interpreter:b_get_invariant_from_machine(_)).
:- execute(same_texpr(A,B),(ground(A),ground(B))).
%% :- execute(b_interpreter:ground_wait_flags(_)). %% not safe; just for experimenting
:- unfold(b_test_member_expression(_El,_Set,_LocalState,_State,_WF)).
:- unfold(b_test_member_expression2(_S,_T,_I,_El,_LocalState,_State,_WF)).
:- execute(get_priority_of_boolean_expression2(A,_),ground(A)).
:- execute(cst_in_boolean_type(S,_), nonvar(S)).
:- execute(binary_in_boolean_type(S,_), nonvar(S)).
:- execute(unary_in_boolean_type(S,_), nonvar(S)).
:- execute(unary_not_in_boolean_type(S,_), ground(S)).

:- execute(unary_function(S,_,_), nonvar(S)).
:- execute(symbolic_closure_unary_operator(S), nonvar(S)).
:- unfold(b_test_atomic_equal_boolean_expression(_Arg1,_Arg2,Info,LocalState,_State,_WF),
  (ground(Info),nonvar(LocalState))). % TO DO: partial evaluator should itself determine when it is safe to unfold with the cuts
:- unfold(b_compute_card(_Arg1,_Type,_Info,_LocalState,_State,_Card,_WF)).
:- unfold(b_compute_card2(Arg1,_Type,_Info,_LocalState,_State,_Card,_WF),nonvar(Arg1)).
:- execute(kernel_objects:max_cardinality(ST,_MaxCard), ground(ST)).
:- unfold(b_compute_max(Ex,_,_Info,_LocalState,_State,_Max,_WF),nonvar(Ex)).
:- unfold(b_compute_min(Ex,_,_Info,_LocalState,_State,_Max,_WF),nonvar(Ex)).
:- unfold(b_compute_expressions(L,_,_,_,_),nonvar(L)).
:- unfold(b_not_test_boolean_expression(P,_LocalState,_State,_WF),nonvar(P)).
:- unfold(b_not_test_boolean_expression(P,_LocalState,_State,_WF,_,_),nonvar(P)).
:- unfold(b_not_test_boolean_expression2(P,_,_LocalState,_State,_WF,_,_),nonvar(P)).
:- unfold(b_not_test_atomic_boolean_expression2(P,_Info,_LS,_S,_WF,_Ai,_Ao),nonvar(P)).
:- unfold(b_test_notmember_expression(Set,_TElem,_TSet,_LocalState,_State,_WF),ground(Set)).
:- unfold(check_invariant_with_time_out(_,_)).
:- unfold(check_invariant_predicate_with_time_out(_,_,_)).
:- unfold(b_compute_arith_expression(Arg2,_LS,_State,_CLPFD_Expr2,_WF),nonvar(Arg2)).
:- unfold(b_compute_arith_expression2(Arg2,_I,_LS,_State,_CLPFD_Expr2,_WF),nonvar(Arg2)).
% annotations for statements:
:- execute(try_get_op_name(_FullOperation,_Name)).
:- execute(b_top_level_operation(_Name)).
:- unfold(b_execute_top_level_operation_update_wf(_Name,_FullOp,_InState,_Updates,_Path,_WF)).
:- unfold(b_execute_operation2(_Name,_FullOperation,_InState,_Updates,_,_,_,_,_Path,_WF,_)).
:- execute(lookup_cached_transitions(_Name,_InState,_Solutions)). % WARNING: will only work if cache disabled !!!! TO DO : fix
:- execute(b_get_machine_operation_for_animation(_,_,_,_,_,_)).
:- unfold(b_execute_operation3(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_)).
:- execute(empty_state(_)).
:- execute(create_full_operation(_Name,_Results,ParaValues,_ResultValues,_FullOperation), builtin_tools:is_list_skeleton(ParaValues)).
%:- unfold(check_additional_guard(_LocalState,_State,_WF)).
:- execute(check_additional_guard(_,_LocalState,_State,_WF), \+ b_machine_temp_predicate(_)).
:- unfold(b_execute_top_level_statement(Body,_LocalState,_NewInState,_OutState,_WF,_Path,_OR),nonvar(Body)).
:- unfold(b_execute_top_level_statement2(_Stmt,_Body,_LocalState,_NewInState,_OutState,_WF,_Path,_OR)).
:- unfold(b_execute_statement(Body,_LocalState,_NewInState,_OutState,_WF,_Path),nonvar(Body)).
:- unfold(b_execute_statement(Body,_LocalState,_NewInState,_OutState,_WF,_Path,_OR),nonvar(Body)).
:- unfold(b_execute_statement2(Stmt,_Infos,_LocalState,_NewInState,_OutState,_WF,_Path,_OR),nonvar(Stmt)).
:- unfold(b_execute_sequence(First,_Rest,_LocalState,_InState,_OutUpdates,_WF,_Path,_OR),nonvar(First)).
:- unfold(b_execute_statement_when_state_set(_Stmt,_LS,_State,_USoFar,_Res,_WF,_Path,_OR)).
:- unfold(check_state_skeleton_bound(A,_),nonvar(A)).
:- unfold(check_state_skeleton_bound_aux(A,_,_),nonvar(A)).
:- unfold(b_execute_statement_when_state_set2(Stmt,_LS,_State,_USoFar,_Res,_WF,_Path,_Bound,_OR),nonvar(Stmt)).
:- unfold(b_execute_while(_COND,_STMT,INV,_VARIANT,_Modvars,_VarVal,_LocalState,_InState,_OutState,_WF),nonvar(INV)).
:- unfold(b_execute_while_loop(_COND,_STMT,INV,_VARIANT,_,_,_LocalState,_InState,_OutState,_WF),nonvar(INV)). % TO DO - abstract LocalState !
:- unfold(b_check_execute_while_loop1(_INVRes,COND,_STMT,_INV,_VARIANT,_,_,_LocalState,_InState,_OutState,_WF),nonvar(COND)).
:- unfold(b_check_execute_while_loop2(_CondRes,_COND,STMT,_INV,_VARIANT,_,_,_LocalState,_InState,_OutState,_WF),nonvar(STMT)).
:- memo(b_execute_while_body(A,B,C,D,E,F,LocalState,InState,G,WF),
          (generalize_call(list(term(bind,[atomic,dynamic])),LocalState,S1),
          generalize_call(list(term(bind,[atomic,dynamic])),InState,S2)),
          b_execute_while_body(A,B,C,D,E,F,S1,S2,G,WF)).
:- unfold(b_execute_while1(COND,_STMT,_INV,_VARIANT,_PrevVariantValue,_VarVal,_LS,_InState,_OutState,_WF),nonvar(COND)).
:- unfold(b_execute_while_when_state_set(COND,_STMT,_INV,_VARIANT,_PrevVariantValue,_LocalState,_State,_Res,_WF,_),nonvar(COND)).
:- execute(get_while_reads_write_info_and_filter_state(Info,LS,S,_,_,_ModVars),
  (ground(Info),builtin_tools:is_list_skeleton(LS), % we should check if it is a bound state
   builtin_tools:is_list_skeleton(S))).
:- execute(get_texpr_id(IDE,_),nonvar(IDE)).
:- execute(def_get_texpr_id(IDE,_),nonvar(IDE)).
:- unfold(get_results(Results,_,_,_,_ReportError,_,_WF),nonvar(Results)).
:- unfold(get_results1(Results,OutState,_ResultValues,_NewOutState,_ReportError,_,_WF),(Results==[] ; (nonvar(Results),nonvar(OutState)))).
:- unfold(get_results2(Finished,_Results,_RV,_NRV,_OutState2,_ResultValues,_NewOutState,_ReportError,_,_WF),nonvar(Finished)).
%:- execute(bmachine:b_machine_temp_predicate(_)).
:- unfold(split_names_and_types(Ids,_,_),nonvar(Ids)).
:- execute(b_machine_temp_predicate(_)).
:- unfold(set_up_typed_localstate(Ids,_Vals,_TVs,_In,_Out,_),nonvar(Ids)).
%:- execute(l_expand_and_normalise_values(V,_,_),V==[]). % Replaced by unfold
:- unfold(store:l_expand_and_normalise_values(H,_,_),nonvar(H)).
:- execute(set_up_undefined_localstate(V,_,_),ground(V)).
:- execute(b_tighter_enumerate_values(Vals,_WF),Vals==[]).
:- execute(nth1(A,L,_),(nonvar(A) ; builtin_tools:is_list_skeleton(L))).
:- unfold(b_execute_inner_statement(_,_,_,_,_,_,_)).
:- unfold(b_execute_inner_statement(_Body,_LocalState,_InState,_OutState,_WF,_Path,_,_OR)). % TO DO: ensure block decl. generated !
:- unfold(b_execute_statements_in_parallel(S,_,_,_,_,_,_),nonvar(S)).
:- unfold(b_execute_else_list(L,_LocalState,_InState,_OutState,_WF,_LWF,_OR),nonvar(L)).
:- unfold(b_execute_else_list2(_PredRes,Body,_Rest,_LocalState,_InState,_OutState,_InnerWF,_LWF,_OR),nonvar(Body)).
:- unfold(b_compute_assign_expressions_when(_EXPRsTl,_LocalState,_InState,_ValTl,_WF,_OR)).
:- unfold(b_compute_expressions_when2(_EXPRsTl,_LocalState,_InState,_ValTl,_WF,_LWF)).

:- unfold(b_compute_expression_when2(_EXPRsHd,_LocalState,_InState,_ValHd,_WF,_)).
:- unfold(get_assign_expr_priority_wf(OR,_,_),nonvar(OR)).
:- unfold(b_assign_values_or_functions(L,_,_LS,_S,_NRest,_WF,_OR),nonvar(L)).
:- unfold(b_assign_value_or_function(Expr,_Value,_LS,_S,_Id,_New,_WF),nonvar(Expr)).
:- unfold(store_values(IDs, _VALs, _OutState),nonvar(IDs)).
:- unfold(combine_updates(OutState1,OutState2,_OutState),(nonvar(OutState1) ; OutState2==[])).
%:- unfold(combine_updates_aux(OutState1,_OutState2,_OutState),nonvar(OutState1)).

:- unfold(b_try_check_boolean_expression(Pred,_LS,_InState,_WF,_PredRes),nonvar(Pred)).
:- unfold(b_try_check_boolean_expression_c(Pred,_LS,_InState,_WF,_PredRes),nonvar(Pred)).
:- unfold(b_try_check_boolean_expression_c2(Pred,_LS,_InState,_WF,_PredRes),nonvar(Pred)).
:- unfold(b_try_check_boolean_expression_lwf(Pred,_LS,_S,_WF,_PredRes,_LWF),nonvar(Pred)).
:- unfold(b_execute_let_definitions(Defs,_NewLocalState,_InState,_WF),nonvar(Defs)).
:- execute(b_compiler:b_compile(Pred,_,LocalState,InState,_CPRED,_WF),(ground(Pred), b_interpreter:is_bound_state(LocalState), b_interpreter:is_bound_state(InState))).
 :- unfold(b_try_check_boolean_expression_lwf_c(_Pred,_LocalState,_InState,_WF,_PredRes,_LWF)).
:- execute(create_any_bindings(A,_,_),ground(A)).
:- execute(copy_wf_start(_,_)).
:- unfold(convert_list_of_expressions_into_set(_List,_Set)).
:- unfold(convert_list_of_expressions_into_set_wf(_List,_Set,_Type,_WF)).
:- execute(is_extension_function(F,_,_),ground(F)).
:- execute(exclude(F,L,_),(ground(F),F=unused_variable(_),b_interpreter:is_bound_state(L))).
:- execute(is_a_conjunct(_C,_,_)).

:- public is_bound_state/1.
is_bound_state(X) :- var(X),!,fail.
is_bound_state([]).
is_bound_state([H|T]) :- nonvar(H), H=bind(ID,_), ground(ID),
  is_bound_state(T).

:- forward_block(b_compute_expression_when2(?,?,?,?,?,?)) .
:- forward_block(b_execute_inner_statement(?,?,?,?,?,?,?,?)) .
:- forward_block(b_execute_while(?,?,?,?,?,?,?,?,?,?)) .
:- forward_block(b_execute_while_loop(?,?,?,?,?,?,?,?,?,?)) .
:- forward_block(b_check_execute_while_loop1(-,?,?,?,?,?,?,?,?,?,?)).
:- forward_block(get_results1(?,?,?,?,?,?,?)).
:- forward_block(get_results2(?,?,?,?,?,?,?,?,?,?)).
:- forward_block(combine_updates(?,?,?)) .
%:- forward_block(combine_updates_aux(?,?,?)) .
:- forward_block(b_execute_statement_when_state_set2(?,?,?,?,?,?,?,?,?)) .
:- forward_block(check_state_skeleton_bound_aux(?,?,?)) .
:- forward_block(b_check_execute_while_loop2(?,?,?,?,?,?,?,?,?,?,?)) .

:- forward_block(b_enum_test_equivalence(?,?,?,?,?,?,?)) .
:- forward_block(b_test_disjunction(?,?,?,?,?,?,?)) .
:- forward_block(block_test_implication(?,?,?,?,?,?)) .

:- execute(pe_print(_)).
:- public pe_print/1.
pe_print(X) :- print('*** PE INFO: '), print(X),nl.


% ---- for b_interpreter_check.pl:

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(partially_evaluate, true)).
:- use_module(partial_evaluator(pe)).
:- unfold(b_check_boolean_expression(_LHS,_LocalState,_State,_CWF,_PredRes)).
:- unfold(b_check_boolean_expression(_LHS,_LocalState,_State,_CWF,_PredRes,_Ai,_Aii)).
:- unfold(b_check_boolean_expression0(_WDE,_WDV,_Expr,_LS,_S,_WF,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression1(_EXPR,_LS,_S,_WFD,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression2(_Expr,_Infos,_LS,_S,_,_Res,_Ai,_Ao)).
:- unfold(negate_equal_false(A,B,_,_),
      ((ground(A); A \= b(boolean_false,boolean,_)),(ground(B); B \= b(boolean_false,boolean,_)))).
:- unfold(b_check_boolean_expression3_pos(_,_Info,_LocalState,_State,_WFD,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression3_neg(_,_Info,_LocalState,_State,_WFD,_Res,_Ai,_Ao)).
:- unfold(b_check_boolean_expression4(_Pred,_Infos,_LocalState,_State,_WFD,_Res)).
:- unfold(register_predicate_wfd(WFD,_Pred,Infos,_Res,_Reused,_Ai,_Ao),(WFD=wfwd(_,A,B),(A==B;A\=B,ground(Infos)))).
:- unfold(register_predicate_aux(Pred,_PredTruthVar,_Reused,_Ai,_Ao),nonvar(Pred)).
:- unfold(check_pred_truth_var(PredTruthVar),nonvar(PredTruthVar)).
:- execute(get_texpr_type(_,_)).
:- execute(always_well_defined(A),ground(A)).
:- unfold(b_wd_compute_expression(_Expr,_LocalState,_State,_Value,_)).
:- unfold(b_compute_expression_delay(_WDE,_WDV, _Expr,_LocalState,_State,_Value,_WF)).
:- execute(instantiate_to_any_value(A,_E,_WF),ground(A)).
:- unfold(imply_test_boolean_expression(_PredRes1,_PredRes2, _RHS,_LocalState,_State,_WF,_)).
:- unfold(equiv_bidrectional_test_boolean_expression(_PredRes,_PredRes2,_LHS,_RHS,_LS,_S,_WF)).
:- unfold(equiv_test_boolean_expression(_PredRes,_PredRes2,_RHS,_LocalState,_State,_WF,_)).
:- unfold(arithmetic_op(Pred,_Op,_LHS,_RHS),nonvar(Pred)).
:- unfold(check_arithmetic_operator(Op,_LHValue,_RHValue,_Res),nonvar(Op)).
:- execute(get_priority_of_boolean_expression(A,_),ground(A)).
:- execute(get_priority_of_boolean_expression2(A,_),ground(A)).
:- execute(normalise_disjunct(E,_,_),ground(E)).
:- execute(normalise_disjunct2(E,_Info,_,_),ground(E)).
:- execute(construct_norm_disjunct(E,_R),ground(E)).
:- execute(construct_norm_disjunct2(E,_R,_I),ground(E)).
:- execute(empty_avl(_)).
:- unfold(norm_pred(A,_R),nonvar(A)).
:- execute(norm_pred_check(A,_R),nonvar(A)). % Dangerous
:- unfold(norm_expr(A,_R),nonvar(A)).
:- unfold(norm_expr2(A,_R),nonvar(A)).
:- unfold(add_predicate(_Pred,_Var,_AVL,_NewAVL)).
:- unfold(reuse_predicate(_Pred,_Var,_AVL)).
:- execute(avl_fetch(Pred,AVL,_Var), (ground(AVL),(AVL==empty ; ground(Pred)))).
:- execute(avl_store(Pred,AVL,_Var,_NewAVL), (ground(AVL),(AVL==empty ; ground(Pred)))).

:- forward_block(b_test_disjunction(?,?,?,?,?,?,?)) .
:- forward_block(b_enum_test_equivalence(?,?,?,?,?,?)) .
:- forward_block(block_test_implication(?,?,?,?,?,?)) .
:- forward_block(b_compute_expression_delay(?,?,?,?,?,?,?)) .
:- endif.


:- if(environ(use_pe_cache, true)).

 % just use a previously generate pe_cache file defining
  %  state_satisfies_invariant_aux and
  %  b_execute_top_level_operation_update(Name,FullOp,InState,Updates,Path)
:- include(pe_cache).
reset_partial_evaluator. % nothing to be done

:- elif(\+ environ(partially_evaluate, true)).

reset_partial_evaluator. % nothing to be done

state_satisfies_invariant_aux(ID,State) :-
   check_invariant_with_time_out(ID,State).  % version without partial evaluation

:- else.
/* partial evaluation enabled */
:- use_module(partial_evaluator(pe)).
% ANNOTATIONS moved to unused_code/

reset_partial_evaluator :- clear_specialized_code.
state_satisfies_invariant_aux(ID,State) :-
    MemoCode=generalize_call(list(term(bind,[atomic,dynamic])),State,StateSkel),
    fast_partially_evaluate_and_load(check_invariant_with_time_out(PID,StateSkel),
                                     ResCall,MemoCode,[ground/StateSkel]),
    StateSkel=State, PID=ID,
    set_error_context(checking_invariant),
    call(ResCall),
    clear_error_context.

%abstract_state([],[]).
%abstract_state([bind(ID,_)|T],[bind(ID,_)|AT]) :- abstract_state(T,AT).
:- endif.
reset_partial_evaluator
:- if( environ(use_pe_cache, true)).
% the specialized code is in pe_cache.pl
:- elif(\+ environ(partially_evaluate, true)).

b_execute_top_level_operation_update(Name,FullOperation,InState,Updates,Path) :-
    b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF),
    ground_wait_flags(WF).
    %store:store_updates_and_normalise(Updates,InState,NewState).

:- else.
/* partial evaluation enabled */

:- unfold(pe_execute_top_level_operation_update(_Name,_FullOperation,_InState,_Updates,_Path)).
pe_execute_top_level_operation_update(Name,FullOperation,InState,Updates,Path) :-
    b_execute_top_level_operation_update_wf(Name,FullOperation,InState,Updates,Path,WF),
    ground_wait_flags(WF).

b_execute_top_level_operation_update(Name,FullOp,InState,Updates,Path) :-
    %print(compile_op(Name,FullOp)),nl,
    % THIS SHOULD BE DONE BY AN ANNOTATION: start_specialization(p(static,dynamic,...)) rather than by hand
    MemoCode=generalize_call(list(term(bind,[atomic,dynamic])),InState,StateSkel),
    fast_partially_evaluate_and_load(pe_execute_top_level_operation_update(Name,FullOp,StateSkel,Updates,Path),
                                     ResCall,MemoCode,[ground/StateSkel]),
    StateSkel=InState, call(ResCall).
:- endif.

%%%%%%%%%%%%%%%%
OLD CODE

b_enumerate_exists_aux_non_ground(ESWF,WaitVars,_Parameters,LocalWF,WF) :-
           % print(b_test_exists_enumerating(ESWF,_WaitVars,_Parameters,LocalWF)),nl,
        (nonvar(_ESWF)
          -> % old version, before RealWaitVars with used_ids_defined_by_equality was used
             kernel_waitflags:ground_wait_flag_to_enumeration(LocalWF),
             when(ground(WaitVars),b_enum_exists_with_cut(LocalWF,WF))
          ; %print(copying_waitflag_store(_Parameters,LocalWF)),nl, print(to_wf(WF)),nl, print(waitvars(WaitVars)),nl,
            kernel_waitflags:copy_waitflag_store(LocalWF,WF)
            % , portray_waitflags(WF),nl
        ),
         % print(waiting_for_ground(WaitVars)),nl, portray_waitflags(LocalWF),nl,
              % TO DO: use kernel_tools
        true. % when(ground(WaitVars),(print(wv(WaitVars)),nl,b_enum_exists_with_cut(LocalWF,WF))).

b_enum_exists_with_cut(LocalWF,WF) :-
     kernel_waitflags:get_idle_wait_flag(b_enumerate_exists,WF,LWF),
     % allow all other co-routines waiting on WaitVars to complete
     b_enum_exists_with_cut_aux(LWF,LocalWF).
:- block b_enum_exists_with_cut_aux(-,?).
b_enum_exists_with_cut_aux(_LWF,LocalWF) :-
     %  print(b_enum_exists_with_cut(LocalWF)),nl,
     ground_wait_flags_for_exists(LocalWF).

ground_wait_flag_to_enumeration(WF) :-
   last_finite_priority(P),
   ground_wait_flag_to(WF,P).
