% (c) 2009-2019 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(units_tools, [contains_top/1,
                        check_state_add_error/2,
                        insert_spacer_and_join/2,
                        is_unit_list/1,
                        append_to_tcltk_replay_sequence/1,
                        reset_tcltk_replay_sequence/0,
                        replay_tcltk_sequence/0,
                        state_was_updated/2,
                        extract_value_list/2,
                        convert_to_atom/2
                       ]).

:- use_module(library(lists)).

%:- use_module(library(tcltk)).  % will be included automatically by prob_tcltk.pl; we do not want probcli to depend on Tcl/Tk

:- use_module(probsrc(tools)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(state_space)).
%:- use_module(probsrc(coverage_tools_annotations),['$NOT_COVERED'/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,plugin_units).
:- module_info(description,'Units Plugin: Analysis tools used in several places.').


:- use_module(units).

:- dynamic tcltk_replay_sequence/1.

% TO DO: remove this callback to Tcl/Tk
%:- use_module(library(tcltk),[tcl_eval/3]).
to_tcltk(Command) :-
    on_exception(_E, (user:cur_tcl_process(Tcl), tcltk:tcl_eval(Tcl, Command, _)), true).

append_to_tcltk_replay_sequence(X) :-
    retract(tcltk_replay_sequence(S1)),
    assert(tcltk_replay_sequence([X|S1])).

reset_tcltk_replay_sequence :-
    retractall(tcltk_replay_sequence(_)),
    assert(tcltk_replay_sequence([])).

replay_tcltk_sequence :-
    to_tcltk('procInsertHistoryOptionsState'),
    retract(tcltk_replay_sequence(Seq)),
    reverse(Seq,Seq2),
    replay_tcltk_sequence(Seq2).
replay_tcltk_sequence([]).
% in tcltk mode
%replay_tcltk_sequence([Op|T]) :-
%    state_space:current_options(Options),
%    nth0(N,Options,(_,_,Op,_)),
%    ajoin(['procPerformOptionByIOption', ' ', N],Command),
%    to_tcltk(Command), !,
%    replay_tcltk_sequence(T),
%    '$NOT_COVERED'('GUI code not used in automatic tests').
% in cli mode
replay_tcltk_sequence([Op|T]) :-
    current_state_id(CurID),
    visited_expression(CurID,CurState),
    user:compute_all_transitions(CurID,CurState),
    user:tcltk_perform_action(Op,_NewID),
    replay_tcltk_sequence(T).
insert_spacer_and_join(List,Output) :-
    atomic(List) -> Output=List ; ajoin_with_sep(List,' ',Output).

:- assert_must_succeed(contains_top(type(top([])))).
:- assert_must_succeed(contains_top(type(top([_])))).
:- assert_must_succeed(contains_top(type(type2(top([]),_)))).
:- assert_must_succeed(contains_top(type(type2(top([_]),_)))).
:- assert_must_fail(contains_top(type(_))).
contains_top(X) :- var(X), !, fail.
contains_top(X) :- X=top(_).
contains_top([H|T]) :-
    !, (contains_top(H) ; contains_top(T)).
contains_top(Pred) :-
    !, Pred =.. [Head|Tail],
    Tail = [_|_],
    (contains_top(Head) ; contains_top(Tail)).

check_state_add_error(ID, concrete_constants(State)) :-
    !, check_state_add_error(ID,State).
check_state_add_error(_ID,[]).
check_state_add_error(ID,[bind(Var,Type)|T]) :-
    contains_top(Type), !,
    ajoin(['Multiple Units inferred for variable ', Var, '. This is a type error.'], ErrTerm),
    units:units_error_context(Context),
    store_state_error(ID,abort_error('Multiple Units inferred',
        'State contains Error',ErrTerm,Context),_),
    check_state_add_error(ID,T).
check_state_add_error(ID,[bind(Var,Type)|T]) :-
    \+(ground(Type)), units:static_result(_,ListOfVars),
    member(Var,ListOfVars), !,
    ajoin(['Variable ', Var, ' is involved in constraint but still unset'], ErrTerm),
    units:units_error_context(Context),
    store_state_error(ID,abort_error('Constraint variable unset',
        'Unsolved Constraint',ErrTerm,Context),_),
    check_state_add_error(ID,T).
check_state_add_error(ID, [_H|T]) :-
    check_state_add_error(ID,T).

is_unit_list(X) :- var(X), !, fail.
is_unit_list([X]) :- var(X), !, fail.
is_unit_list([]).
is_unit_list([[_,_,_]|T]) :-
    is_unit_list(T).

:- assert_must_succeed(state_was_updated([bind(a,_A)],[bind(a,a)])).
:- assert_must_fail(state_was_updated([bind(a,_A)],[bind(a,_B)])).
:- assert_must_fail(state_was_updated([bind(a,top([]))],[bind(a,top([]))])).
:- assert_must_fail(state_was_updated([bind(a,top([]))],[bind(a,top([[[0,s,1]]]))])).
:- assert_must_fail(state_was_updated([bind(a,top([]))],[bind(a,top([[[0,s,1]],[[0,m,1]]]))])).
:- assert_must_succeed(state_was_updated([bind(a,top([])),bind(b,1)],[bind(a,top([[[0,s,1]]])),bind(b,2)])).
:- assert_must_succeed(state_was_updated([bind(a,A),bind(b,_B)],[bind(a,A),bind(b,b)])).
:- assert_must_succeed(state_was_updated([bind(a,A),bind(b,integer(_B))],[bind(a,A),bind(b,integer([[0,m,1],[0,s,-1]]))])).
:- assert_must_succeed(state_was_updated([bind(a,5),bind(b,integer(_B))],[bind(a,5),bind(b,integer([[0,m,1],[0,s,-1]]))])).
:- assert_must_succeed(state_was_updated([bind(a,integer(_A))],[bind(a,integer([[0,m,1],[0,s,-1]]))])).
:- assert_must_succeed(state_was_updated([bind(v,integer([[0,s,1]])),bind(w,integer(_)),bind(x,integer([[0,m,1]])),bind(y,integer([[0,m,-1]]))],
                                         [bind(v,integer(top([]))),bind(w,integer(_)),bind(x,integer([[0,m,1]])),bind(y,integer([[0,m,-1]]))])).
state_was_updated([bind(ID,Val)|T1],[bind(ID,Val2)|T2]) :-
    (\+ground(Val)), (\+ground(Val2)),
    !, state_was_updated(T1,T2).
state_was_updated([bind(ID,Val)|_T1],[bind(ID,Val2)|_T2]) :-
    (\+ground(Val)), ground(Val2).
state_was_updated([bind(ID,Val)|T1],[bind(ID,Val2)|T2]) :-
    Val == Val2,
    !, state_was_updated(T1,T2).
state_was_updated([bind(ID,Val)|T1],[bind(ID,Val2)|T2]) :-
    contains_top(Val), contains_top(Val2), !,
    state_was_updated(T1,T2).
state_was_updated([bind(ID,Val)|_T1],[bind(ID,Val2)|_T2]) :-
    Val \== Val2.

extract_value_list([],[]).
extract_value_list([bind(_,X)|T],[X|ET]) :- extract_value_list(T,ET).

% allow units to be specified as both "strings" or as atoms
convert_to_atom(Atom,Res) :- atom(Atom),!,Res=Atom.
convert_to_atom(Codes,Res) :- atom_codes(Res,Codes).
